{ MineSweeper solver.
  This program solves the MineSweeper game.
  After the user indicated the program which is the MineSweeper Window,
  the program moves alone the mouse cursor to click on cells and
  solve the game.
  In  tEx063.pp  a simple strategie was used, here the strategie will be
  more sophisticated, using probabilities to decide on which cell to click.

  Here are some rules.
  """"""""""""""""""""
  Simple Rule one :
  A cell (nX,nY) has no bomb for sure, if a neighbour cell indicate a number equal
  to the number of its neighbour having a Flag. i.e.
  if aabBO[i,j] = 0   for (i,j) a neighbour of (nX,nY)
  It is simpler to say that if a cell indicate as many bombs as it has
  neighbour with a flag, then all other neighbour has no bombs.
  But the other way has the advantage of taking only decisions about
  unvisited cells.

  Simple (nX,nY) Rule two :
  A cell has a bomb for sure, if a neighbour cell idicate as many bombs as
  the number of unvisited cells minus those having a Flag. i.e.
  if aabBO[i,j] = aabTS[i,j]  for (i,j) a neighbour of (nX,nY)
  It is simpler to say that if a cell indicate as many bombs as it has
  neighbour with a flag plus unvisited neighbours, then all
  unvisited neighbour have a bombs.

  Many other special rules can be stated. But they will be explored
  systematically, another way.
  Here are some drawings giving examples of other rules.
  x = a visited cell
  a digit = number of neighbour cells having a bomb - those already Flagged
  . = a cell, of any type
  o = has no bombs for sure after deduction

  x . . o    x . . o     x x . o o     x x x
  x 1 1 o    . 2 1 o     x 1 . 1 o     x 1 . o    etc.
  x . . o    x . . o     x x . o o     x . 1 o
                                         o o o

  If no cell can be found without a bomb, using the simple rules, then
  for each cell, the program will see what are the deductions if a bomb is
  put into the cell. If it leads to a contradiction, then for sure this
  cell has no bombs.
  The contradiction will be that  aabBO[i,j] < 0  for a cell (i,j), i.e.
  more flags has to be set than there are free neighbour cells.
  --------------------------------------------}
program tEx064;

// To tell FreePascal that the application is a Graphical User Interface
// i.e. is running in windows, not in text mode.
{$APPTYPE GUI}

Uses

  wApplication0
 ,wWindow0
 ,wDialog0
 ,wWindowBase0
 ,wDialogCreatorLow2
 ,wDialogCreator2
 ,wVariableNumLink2
// ,wUtiles0 // for the AccelTableMSGCreate function
 ,wGlobals2
 ,wDebug0
 ,Windows;

type
  Pinteger = ^integer;
  PMyWindow = ^TMyWindow;
  TMyWindow = object(TWindow)
  //"""""""""""""""""""""""""
  hWinSec       : integer;
  // Handle of the other window

  fGetWinSecHandle : boolean;
  // To get the window handle of another window

  pDialogActivate : PDialog;

  colGray : TypeColor;
  // Gray color of the cels.

  fSolveFast : boolean;
  // To solve faster and not see the mouse mouvments.

  ptTopLeft : POINT;
  // Position in pixels in the client coordinate of the top left corner
  // of the top left mine square.

  ptRightBottomMargin : POINT;
  // Size of the right and Bottom margin of the Minesweeper game.
  // Used to compute the number of cells in the game.

  nXDelta, nYDelta : integer;
  // number of pixels between two mine squares.

  nXNumber, nYNumber : integer;
  // number of mine squares on a line and on a column

  fQuitSolve : boolean;
  // Set to TRUE to quit the solver

  fBombExplosed : boolean;
  // A bomb has been clicked on. => Lost.

  abXFree,abYFree : array[0..2500] of byte;
  // List of all cells which have not been visited.

  nNbFreeCells : integer;
  // Number of not visited cells.

  nNbBombsTot : integer;
  // Total number of bombs. This must be given by the user,
  // in the custom mode.

  nNbBombs : integer;
  // number of unresolved bombs, i.e.
  // number of cells having a bomb, but no flag on it.

  aabBObis, // copy
  aabBO : array[0..51,0..51] of integer;
  // The content of the MinerSweep
  // Row and column 0 are used to simplify computation, but
  // First mine square Row and Column begins with index 1.
  // if aabBO[i,j] < 9, then it contains the number of neighbour not visited
  // (i.e. marked cteToSee) which have a bomb.

  aabTSbis, // copy
  aabTS : array[0..51,0..51] of integer;
  // number of neighbour cels Too See, i.e.
  // not having been explored.

  // To do backtracking, for searching cells having low probability of
  // having a bomb.
  apnModif : array[0..2500] of Pinteger;
  // list of modified variables which has to be set back to their original value

  anModifValue : array[0..2500] of integer;
  // List of the values which have been modifies.

  nIndModif : integer;
  // Index of the two previous arrays.
  // Above it, indexes have not been defined.

  nNbFreeCellsBis : integer;
  // during backtracking, number of unvisited cells.

  nNbBombsBis : integer;
  // during backtracking, number of left bombs.

  nNbFreeCellsInfo : integer;
  // Number of free cells having neighbour containing information.
  // Set by sorting, such that
  // (abXFree,abYFree)[1..nNbFreeCellsInfo] are those interresting
  // To test if they contain bombs or not

  vProbProduct : Float;
  // product of (1-prob)  where prob was the probability that the cell
  // contain a bomb. In most cases, prob = 0, but at starting and
  // sometimes, decisions must be made with a certain probability.

  constructor Init(strWinTitlePrm:ShortString);
  destructor Done; virtual;

  procedure ToggleSolveFast;
  // Option, to set the speed

  procedure GetSecWindowHandle;
  // To get the handle of the minesweeper window

  procedure GetSecWindowPosition;
  procedure SimulateLeftClick;
  procedure SimulateRightClick;

  procedure MoveMouseTo(nX2,nY2:integer);
  // Moving the mouse cursor to the position nXP, nYP, slowly

  procedure ClickOnNewGame;
  // Click on the new game icone to begin a new minesweeper game

  procedure InitCellsData;
  // Initialisation to set that no cells have been visited, to see their content,
  // and other array initialisation

  procedure GetSquarePositions;

  function GetMineContent(hDCS:HDC; nXP,nYP:integer):integer;
  // nXP, nYP = screen coordinate of the upper left of a mine square.
  // return the content of it.
  // = cteToSee if it has not been explored, 0 if empty,
  // = cteFlag  if a flag is in
  // = cteBoom  if a bomb is in, and thus the game finished.

  procedure GetMineSweeperContent(wFlagsP:word32);
  // Get the content of the complete minesweeper window
  // wFlagsP =1 => show the content in the debug window.

  procedure UpdateMineSweeperContent;
  // The content of the mineSweeper is supposed to have been set right
  // before the last click, from it, update the information about
  // the content.

  procedure PutFlags;
  // Put Flags on cells where it is sure that a bomb existe.

  procedure StartingExplore;
  // When we beging, start with some special cells.

  function SaftyExplore:integer;
  // Click on cells which have no bombs for sure

  procedure SaveValue(pnValue:Pinteger);
  // Save in  apnModif[nIndModif]  and anModifValue[nIndModif]
  // the pointer  pnValue,  and the content of pnValue^

  function SetFlagTo(nXP,nYP:integer):integer;
  // Try to set a  cteFlag  or  cteNoBomb to a cell, to see the
  // logical consequeces, and find contradictions. Thus finding cells
  // having no bombs, or having bombs for sure.

  function SaftyExploreMore:integer;
  // Click on cells which have no bombs for sure
  // But the searching of these cells is more sophisticated than SaftyExplore

  procedure SortFreeCells;
  // Sort the abXFree,abYFree arrays such that cells having few unvisited
  // neighbour come first.

  function ProbabilityExplore:integer;
  // Explore the cells to find which has the lower probabilities of
  // having a bomb. Take into account the number of bombs.
  // In some case, this probability will be zero, and thus
  // it is sure that a cell without bomb will have been found.

  procedure SolveMineSweeper;
  // Go and click to solve the minesweeper

  procedure GetStatistiquePlay;
  // Plays many Minesweeper games to get statistics on it.

  procedure AskAcceptDlgQuit(poDlg:PWindowBase0; wCtrlIDP:word32; var wRep:word32); virtual;
  // To set to NIL the pointer on the dialog box

  function WMCommand(wPar:WPARAM; lPar:LPARAM):LRESULT; virtual;
  // Handles the menu messages

  function WindowProc(hWindow:HWND; wMsg:UINT; wPar:WPARAM; lPar:LPARAM):LRESULT; virtual;
  end; // TMyWindow = object(TWindow)

const
  cteFlag   = 20;  // A cell with a bomb has been marked.
  cteNoBomb = 40;  // A cell without a bomb, but not explored
  cteToSee  = 50;  // unexplored cells
  cteBoom   = 99;  // A bomb has been found

var
  oApp        : TApplication;
  poMyMainWin : PMyWindow;

// var nNotUsed : integer; // only to avoid an error message while compliling in DevPascal 1.9

constructor TMyWindow.Init(strWinTitlePrm:ShortString);
//=====================================================
begin
inherited Init(strWinTitlePrm);

Int32Field(  1001, ptTopLeft.x);
Int32Field(  1002, ptTopLeft.y);
Int32Field(  1003, ptRightBottomMargin.x);
Int32Field(  1004, ptRightBottomMargin.y);
IntegerField(1005, nXDelta);
IntegerField(1006, nYDelta);
IntegerField(1007, nXNumber);
IntegerField(1008, nYNumber);
BooleanField(1009, fSolveFast);
IntegerField(1010, nNbBombsTot);

hWinSec:=0;
fGetWinSecHandle:=FALSE;
pDialogActivate:=NIL;
fSolveFast :=FALSE;

// Usually these values are fixed.
// They can be changed by the user, or computed by the program
ptTopLeft.x  :=14;
ptTopLeft.y  :=56;
ptRightBottomMargin.x:= 8; // 2 pixels less, as a security
ptRightBottomMargin.y:= 8;
nXDelta  :=16;
nYDelta  :=16;
nXNumber :=8;  // This can change, depending on the size of the window
nYNumber :=8;  // This can change, depending on the size of the window

// This line is all what is needed to have a status line.
new(poStatusLine, Init(@Self));
end;  // TMyWindow.Init

destructor TMyWindow.Done;
//=========================
begin
inherited Done;
end;  // TMyWindow.Done

procedure TMyWindow.ToggleSolveFast;
//==================================
// Option, to set the speed
var
  wCheck      // to set a check mark in a menu.
              : word;
begin
// Get the handle on the menu
// hMenuWin:=GetMenu(hWindow);  already done

// Toggle from fast solving to slow solwing, showing the mouse movements
fSolveFast:=not fSolveFast;

// change the menu check mark
if fSolveFast then wCheck:=MF_CHECKED
else               wCheck:=MF_UNCHECKED;
CheckMenuItem(hMenuWin, 351, MF_BYCOMMAND or wCheck);
end;  // TMyWindow.ToggleSolveFast

procedure TMyWindow.GetSecWindowHandle;
//=====================================
// Get the handle of the window which is being activated
begin
// By clicking on another window, normally the MineSweeper window,
// Gets the handle of that window.
hWinSec := 0;

// These two values indicate that when the main window is desactivated,
// The new active window is the wanted one.
pDialogActivate:=PDialog(DialogCreateMSG(20199, @Self));
fGetWinSecHandle:=TRUE;

SetActiveWindow(hWin);
SetFocus(hWin);
end;  // TMyWindow.GetSecWindowHandle

procedure TMyWindow.GetSecWindowPosition;
//=======================================
// To get the position and size of the Second window, i.e. the window
// which was clicked before
var rRec : RECT;
begin
if (hWinSec = 0) then begin
   DialogExecMSG(20299, @Self);
   exit;
   end;

// Get the position and size of the Minesweeper window
GetWindowRect(hWinSec, rRec);
DBW(262144, 'Screen Position =' + ITS5(rRec.Left) + ITS5(rRec.Top));
DBW(262144, 'Window Size X,Y =' + ITS5(rRec.Right - rRec.Left)
                           + ITS5(rRec.Bottom - rRec.Top));
GetClientRect(hWinSec, rRec);
DBW(262144, 'Client Size X,Y =' + ITS5(rRec.Right) + ITS5(rRec.Bottom));

// If incoherent values, put standard values
if (nXDelta <=0) then nXDelta:=16;
if (nYDelta <=0) then nYDelta:=16;

// Number of cells
nXNumber:=(rRec.Right  - ptTopLeft.x - ptRightBottomMargin.x) div nXDelta;
nYNumber:=(rRec.Bottom - ptTopLeft.y - ptRightBottomMargin.y) div nYDelta;

// Assuming that we are in non cutom mode, sets the number of bombs.
nNbBombsTot:=0;
if (nXNumber =  8) and (nYNumber =  8) then nNbBombsTot:=10;
if (nXNumber = 16) and (nYNumber = 16) then nNbBombsTot:=40;
if (nXNumber = 30) and (nYNumber = 16) then nNbBombsTot:=99;

DBW(131072, 'Number of Cells X,Y =' + ITS5(nXNumber) + ITS5(nYNumber));

if (nNbBombsTot > 0) then
   DBW(131072, 'Number of bombs     =' + ITS5(nNbBombsTot)
        + '   IF it is WRONG, you must set the right value')
else DBW(1, 'UNKNOWN NUMBER OF BOMBS. You must set the right number from the menu');

end;  // TMyWindow.GetSecWindowPosition

procedure TMyWindow.SimulateLeftClick;
//====================================
// Just to simulate a Left Click on the Minesweeper window.
var
  ptL         : TPoint;
begin
if (hWinSec = 0) then begin
   DialogExecMSG(20299, @Self);
   exit;
   end;

// Get the mouse position, to transmit it as a parameter to the message
GetCursorPos(@ptL);
ScreenToClient(hWinSec, ptL);

SetForegroundWindow(hWinSec);
SendMessage(hWinSec, WM_LBUTTONDOWN, MK_LBUTTON, MakeLong(ptL.x, ptL.y));
SendMessage(hWinSec, WM_LBUTTONUP,            0, MakeLong(ptL.x, ptL.y));

// Set back the main window as the active one.
//???????SetForegroundWindow(hWin);
end;  // TMyWindow.SimulateLeftClick

procedure TMyWindow.SimulateRightClick;
//=====================================
// Just to simulate a Left Click on the Minesweeper window.
var ptL         : TPOINT;
begin
if (hWinSec = 0) then begin
   DialogExecMSG(20299, @Self);
   exit;
   end;

// Get the mouse position, to transmit it as a parameter to the message
GetCursorPos(@ptL);
ScreenToClient(hWinSec, ptL);

SetForegroundWindow(hWinSec);
SendMessage(hWinSec, WM_RBUTTONDOWN, MK_RBUTTON, MakeLong(ptL.x, ptL.y));
SendMessage(hWinSec, WM_RBUTTONUP,            0, MakeLong(ptL.x, ptL.y));

// Set back the main window as the active one.
//???????SetForegroundWindow(hWin);
end;  // TMyWindow.SimulateRightClick

procedure TMyWindow.MoveMouseTo(nX2,nY2:integer);
//===============================================
// Moving the mouse cursor to the position nXP, nYP,
// slowly of fast, depending on  fSolveFast
var
  nSecurity,
  nXStep,nYStep,
  nX1,nY1,
  nX,nY          : int32;
  wTime0      : word32;
  ptL         : TPOINT;
begin
if fSolveFast then begin
   // Solving fast, do not show the mouse cursor movements.
   SetCursorPos(nX2,nY2);
   exit;
   end;

GetCursorPos(@ptL);
nX1:=ptL.x;  nY1:=ptL.y;  // Starting position
nX:=nX1;     nY:=nY1;     // Starting position

if (nX1 = nX2) and (nY1 = nY2) then exit; // nothing to do

// To set the moving direction
if (nX2 > nX1) then nXStep:=1 else if (nX2 < nX1) then nXStep:=-1 else nXStep:=0;
if (nY2 > nY1) then nYStep:=1 else if (nY2 < nY1) then nYStep:=-1 else nYStep:=0;

nSecurity:=2000;
repeat
   // Move the mouse allong a strait line.
   // Must take care if (X1=X2) or if (Y1=Y2)
   if (abs((nY2-nY)*(nX2-nX1)) < abs((nX2-nX)*(nY2-nY1))) then nX:=nX + nXStep
   else if (abs((nY2-nY)*(nX2-nX1)) > abs((nX2-nX)*(nY2-nY1))) then nY:=nY + nYStep
   else if (nX <> nX2) then nX:=nX + nXStep
   else if (nY <> nY2) then nY:=nY + nYStep;
   // else do not move, the mouse cursor is already at destination.

   // Moves the cursor
   SetCursorPos(nX,nY);

   // Wait to let us see the movement
   wTime0:=GetTickCount;
   repeat
      // Stop if the escape key has been pressed
      if (GetAsyncKeyState(VK_ESCAPE) < 0) or
         (GetAsyncKeyState(VK_SHIFT) < 0) then fQuitSolve:=TRUE;
      if (GetAsyncKeyState(VK_F) < 0) then fSolveFast:=TRUE;
      until wTime0 <> GetTickCount;

   dec(nSecurity);
   until ((nX = nX2) and (nY = nY2)) or (nSecurity <= 0);

if (nSecurity <= 0) then
   DBW(1, '***** PROBLEM, X1,Y1,X2,Y2=' + ITS5(nX1) + ITS5(nY1) + ITS5(nX2) + ITS5(nY2));
end;  // TMyWindow.MoveMouseTo

procedure TMyWindow.ClickOnNewGame;
//=================================
// Click on the new game icone to begin a new minesweeper game
begin
if (hWinSec = 0) then begin
   DialogExecMSG(20299, @Self);
   exit;
   end;

// On the game I have, this reset the game.
SendMessage(hWinSec, WM_COMMAND, 510, 0);
end;  // TMyWindow.ClickOnNewGame

procedure TMyWindow.InitCellsData;
//================================
// Initialisation to set that no cells have been visited, to see their content,
// and other array initialisation
var
  nInd, nX,nY       : Integer;
begin
// All cells must be visited
nInd:=0;
for nX:=1 to nXNumber do  for nY:=1 to nYNumber do begin
   inc(nInd);
   abXFree[nInd]:=nX;
   abYFree[nInd]:=nY;
   end;

nNbFreeCells:=nInd;

// Total number of bombs
nNbBombs:=nNbBombsTot;

// initialisation of the aabBO content.
// Just to simplify, have a border of ficive cells
for nX:=0 to nXNumber + 1 do begin
   aabBO[nX,0]          :=cteNoBomb; // Unexplored
   aabBO[nX,nYNumber+1] :=cteNoBomb;
   end;

for nY:=0 to nYNumber + 1 do begin
   aabBO[0,nY]          :=cteNoBomb; // Unexplored
   aabBO[nXNumber+1,nY] :=cteNoBomb;
   end;

for nX:=0 to nXNumber + 1 do
  for nY:=0 to nYNumber + 1 do begin
   aabTS[nX,nY]         := 8; // 8 unexplored cells around
   end;

// At the limits, the number of unexplored neighbour changes
for nX:=1 to nXNumber do begin
   aabTS[nX,         0]:= 3;
   aabTS[nX,         1]:= 5;
   aabTS[nX,nYNumber  ]:= 5;
   aabTS[nX,nYNumber+1]:= 3;
   end;

for nY:=1 to nYNumber do begin
   aabTS[         0,nY]:= 3;
   aabTS[         1,nY]:= 5;
   aabTS[nXNumber  ,nY]:= 5;
   aabTS[nXNumber+1,nY]:= 3;
   end;

aabTS[1,1]:= 3;  aabTS[nXNumber,1]:= 3;  aabTS[1,nYNumber]:= 3;
aabTS[nXNumber,nYNumber]:= 3;
aabTS[0,0]:= 1;                    aabTS[nXNumber+1,0]:= 1;
aabTS[nXNumber+1,nYNumber+1]:= 1;  aabTS[0,nYNumber+1]:= 1;
end;  // TMyWindow.InitCellsData

procedure TMyWindow.GetSquarePositions;
//=====================================
// Get the positions, sizes of all mine squares, and how many they are.
var
  nSecurity,
  nX,nY       : int32;
  wTime0      : word32;
  rRec        : RECT;
  hDCScreen   : HDC;
  colLast,    // Last pixel color
  colPixel    : TypeColor;
  nCountSame // count the number of colors which where the same on a line
              : integer;
  nYBottomL   // bottom of the last mine square
              : integer;
begin
if (hWinSec = 0) then begin
   // A second window must have been selected
   DialogExecMSG(20299, @Self);
   exit;
   end;

// Initialisations
ptTopLeft.x:=0;
ptTopLeft.y :=0;   nYBottomL:=0;
nXDelta :=0;
nYDelta :=0;
nXNumber :=0;
nYNumber :=0;

// To get a starting point
GetWindowRect(hWinSec, rRec);
nX:=rRec.Left;
nY:=rRec.Bottom-20;

// Get a device context of the screen
hDCScreen:=GetDC(0);

// Get first pixel color
colLast:= GetPixel(hDCScreen, nX, nY);

nCountSame:=1;

nSecurity:=2000;
repeat
   nX:=nX + 1;

   // Moves the cursor
   SetCursorPos(nX,nY);

   colPixel:= GetPixel(hDCScreen, nX, nY);

   if (colPixel = colLast) then inc(nCountSame)
   else begin
      DBW(16, 'nb, color=' + ITS5(nCountSame) + ITS5(colLast.bRed)
                          + ITS5(colLast.bGreen) + ITS5(colLast.bBlue));

      if (nCountSame >= 10) then begin
         // a mine square has been passed through
         if (ptTopLeft.x = 0) then ptTopLeft.x:=nX-nCountSame
         else if (nXDelta = 0) then nXDelta:=nX-nCountSame - ptTopLeft.x;

         inc(nXNumber);
         end;

      nCountSame:=1;
      colLast:=colPixel;
      end;

   // Wait to let us see the movement
   wTime0:=GetTickCount;
   repeat until wTime0 <> GetTickCount;

   until (nX >= rRec.Right) or (nSecurity <= 0);

// Loop to count the mine squares vertically
//"""""""""""""""""""""""""""""""""""""""""""
nX:=rRec.Left+20;
nY:=rRec.Bottom;

// Get first pixel color
colLast:= GetPixel(hDCScreen, nX, nY);

nCountSame:=1;

nSecurity:=2000;
repeat
   nY:=nY - 1;

   // Moves the cursor
   SetCursorPos(nX,nY);

   colPixel:= GetPixel(hDCScreen, nX, nY);

   if (colPixel = colLast) then inc(nCountSame)
   else begin
      DBW(16, 'nb, color=' + ITS5(nCountSame) + ITS5(colLast.bRed)
                          + ITS5(colLast.bGreen) + ITS5(colLast.bBlue));

      if (nCountSame >= 10) then begin
         // a mine square has been passed through
         if (nYBottomL = 0) then nYBottomL:=nY
         else if (nYDelta = 0) then nYDelta:=nYBottomL -(nY);

         inc(nYNumber);
         end;

      nCountSame:=1;
      colLast:=colPixel;
      end;

   // Wait to let us see the movement
   wTime0:=GetTickCount;
   repeat until wTime0 <> GetTickCount;

   until (nY <= rRec.Top+80) or (nSecurity <= 0);

ReleaseDC(0, hDCScreen);

ptTopLeft.y:=nYBottomL - nYDelta * (nYNumber-1);

// Moves the cursor to see the left top corner of the left top mine square
SetCursorPos(ptTopLeft.x,ptTopLeft.y);

// Convert to have coordinate in client units.
ScreenToClient(hWinSec, ptTopLeft);

// Determine the margins
GetClientRect(hWinSec, rRec);
ptRightBottomMargin.x:= rRec.Right  - ptTopLeft.x - nXDelta*nXNumber;
ptRightBottomMargin.y:= rRec.Bottom - ptTopLeft.y - nXDelta*nYNumber;

DBW(8, 'nXLeft, nXDelta, nXNumber=' + ITS5(ptTopLeft.x) + ITS5(nXDelta) + ITS5(nXNumber));
DBW(8, 'nYTop, nYDelta, nYNumber =' + ITS5(ptTopLeft.y) + ITS5(nYDelta) + ITS5(nYNumber));
end;  // TMyWindow.GetSquarePositions

function TMyWindow.GetMineContent(hDCS:HDC; nXP,nYP:integer):integer;
//===================================================================
// nXP, nYP = screen coordinate of the upper left of a mine square.
// return the content of it.
// = cteToSee if it has not been explored, 0 if empty,
// = cteFlag  if a flag is in
// = cteBoom  if a bomb is in, and thus the game finished.
var
  colPixel1
 ,colPixel2
            : TypeColor;

begin
// Get some pixels, to see the content of the mine square
colPixel1:= GetPixel(hDCS, nXP, nYP);

//DBW(1, ITS5(nXP) + ITS5(nYP));
//SetPixel(hDCS, nXP, nYP, RGB(0,0,0));

if (colPixel1 <> colGray) then begin
   // It is either an unexplored square, of a desactivated square
   colPixel2:= GetPixel(hDCS,  4 + nXP, 3 + nYP);

   if (colPixel2 <> colGray) then begin
      // A desactivated cell
      GetMineContent:=cteFlag;
      exit;
      end;

   // An unexplored cell
   GetMineContent:=cteToSee;
   exit;
   end;

// Here we must determine the number of bombs around this square,
// i.e. read the number in the square, or return zero.

colPixel1:= GetPixel(hDCS,  7 + nXP,  7 + nYP);

if (colPixel1 = colGray) then begin
   // An empty square
   GetMineContent:=0;
   exit;
   end;

colPixel1:= GetPixel(hDCS,  3 + nXP,  3 + nYP);

if (colPixel1 = colGray) then begin
   // Not a  2,  not a  3, not a 4, not a 5, not a 6
   colPixel2:= GetPixel(hDCS,  4 + nXP,  7 + nYP);

   if (colPixel2 <> colGray) then begin
      // Not a 1
      GetMineContent:=cteBoom;
      exit;
      end;

   GetMineContent:=1;
   exit;
   end;

colPixel2:= GetPixel(hDCS,  4 + nXP,  9 + nYP);
if (colPixel2 = colGray) then begin
   // not a  2, can be 3,4,5,6
//SetPixel(hDCS, 4 + nXP, 9 + nYP, RGB(0,0,0));

   colPixel2:= GetPixel(hDCS,  3 + nXP,  10 + nYP);
   if (colPixel2 = colGray) then begin
      // not a  3
      GetMineContent:=4;
      exit;
      end;

   colPixel2:= GetPixel(hDCS,  3 + nXP,  6 + nYP);
   if (colPixel2 = colGray) then begin
      // not a  3
      GetMineContent:=3;
      exit;
      end;

   colPixel2:= GetPixel(hDCS,  2 + nXP,  9 + nYP);
   if (colPixel2 = colGray) then begin
      // not a  6
      GetMineContent:=5;
      exit;
      end;

   GetMineContent:=6;
   exit;
   end;

GetMineContent:=2;
end;  // TMyWindow.GetMineContent

procedure TMyWindow.GetMineSweeperContent(wFlagsP:word32);
//========================================================
// Get the content of the complete minesweeper window
// wFlagsP =1 => show the content in the debug window.
var
  ptTLScreenL : POINT; // top left in screen coordinate of the upper left mine square
  nContent,   // Content of a mine square
  nX,nY,
  ii,jj       : integer;
  hDCScreen   : HDC;
  strS        : ShortString;
begin
if (hWinSec = 0) then begin
   DialogExecMSG(20299, @Self);
   exit;
   end;

if (nXDelta = 0) then GetSquarePositions;
if (wFlagsP = 1) then DBW(32, '--------------------------------------------');

for nX:=0 to nXNumber + 1 do
  for nY:=0 to nYNumber + 1 do begin
   aabBO[nX,nY] := 0;
   aabTS[nX,nY] := 0; // Unexplored
   end;

// initialisation of the aabBO content.
for nX:=0 to nXNumber + 1 do begin
   aabBO[nX,0]          :=cteNoBomb; // Unexplored
   aabBO[nX,nYNumber+1] :=cteNoBomb;
   end;

for nY:=0 to nYNumber + 1 do begin
   aabBO[0,nY]          :=cteNoBomb; // Unexplored
   aabBO[nXNumber+1,nY] :=cteNoBomb;
   end;

{for nX:=1 to nXNumber do begin
   aabBO[nX,1]        :=0; // Unexplored
   aabBO[nX,nYNumber] :=0;
   end;
}
// Get a device context of the screen
hDCScreen:=GetDC(0);

ptTLScreenL:=ptTopLeft;
ClientToScreen(hWinSec, ptTLScreenL);

// Reference Gray color
colGray:=GetPixel(hDCScreen, 1+ptTLScreenL.x, 1+ptTLScreenL.y);

// Loop on all mine squares to see their content
nNbBombs:=nNbBombsTot;
nNbFreeCells:=0;
nY:=1;
repeat
   nX:=1;

   // Loop on a row
   repeat
      // Get the content of the cell (nX, nY)
      nContent:=GetMineContent(hDCScreen, ptTLScreenL.x + (nX-1)*nXDelta,
                                          ptTLScreenL.y + (nY-1)*nYDelta);

      // To count the neighbour cell to be still explored
      if (nContent = cteToSee) then begin
         for ii:=nX-1 to nX+1 do  for jj:=nY-1 to nY+1 do
           aabTS[ii,jj]:=aabTS[ii,jj] + 1;

         aabTS[nX,nY]:=aabTS[nX,nY] - 1;

         // One more unvisited cell
         inc(nNbFreeCells);
         abXFree[nNbFreeCells]:=nX;
         abYFree[nNbFreeCells]:=nY;
         end
      else if (nContent = cteBoom) then fBombExplosed:=TRUE; // Lost

      // To count the neighbour cell having a marked bomb
      if (nContent = cteFlag) then begin
         for ii:=nX-1 to nX+1 do  for jj:=nY-1 to nY+1 do
           aabBO[ii,jj]:=aabBO[ii,jj] - 1;

         aabBO[nX,nY]:=cteFlag; // aabBO[nX,nY] + 1;

         // One less unvisited cell having a bomb
         dec(nNbBombs);
         end;

      // Content of this cell
      aabBO[nX,nY]:=aabBO[nX,nY] + nContent;

      inc(nX);
      until (nX > nXNumber);

   inc(nY);
   until (nY > nYNumber);

ReleaseDC(0, hDCScreen);

if (wFlagsP <> 1) then exit; // no info in the debug window

// To show what has been found in the squares
for nY:=1 to nYNumber do begin
   strS:='|';
   for nX:=1 to nXNumber do begin
      if (aabBO[nX,nY] >= cteBoom-9) then // A bomb
        strS:=strS + '*'
      else if (aabBO[nX,nY] = cteToSee) then // An untouched square
        strS:=strS + '.'
      else if (aabBO[nX,nY] = cteToSee-1) then // An untouched square
        strS:=strS + ','
      else if (aabBO[nX,nY] >= cteToSee-2) then // An untouched square
        strS:=strS + ':'
      else if (aabBO[nX,nY] >= cteToSee-9) then // An untouched square
        strS:=strS + ';'
      else if (aabBO[nX,nY] >= cteFlag-9) then // A flgs is in
        strS:=strS + '&'
      else
        strS:=strS + IntToStr(aabBO[nX,nY],1);

      strS:=strS + IntToStr(aabTS[nX,nY],1) + ' ';

      end; // for nX

   DBW(64, strS + '|');
   end; // for nY
DBW(64, 'Bombs left =' + ITS5(nNbBombs) + '  Unvisited=' + ITS5(nNbFreeCells));
end;  // TMyWindow.GetMineSweeperContent

procedure TMyWindow.UpdateMineSweeperContent;
//===========================================
// The content of the mineSweeper is supposed to have been set right
// before the last click, from it, update the information about
// the content.
var
  ptTLScreenL : POINT; // top left in screen coordinate of the upper left mine square
  nContent,   // Content of a mine square
  nn,
  nX,nY,
  ii,jj       : integer;
  hDCScreen   : HDC;
//  strS        : ShortString;
begin
if (hWinSec = 0) then begin
   DialogExecMSG(20299, @Self);
   exit;
   end;

// Get a device context of the screen
hDCScreen:=GetDC(0);

ptTLScreenL:=ptTopLeft;
ClientToScreen(hWinSec, ptTLScreenL);

// Reference Gray color
colGray:=GetPixel(hDCScreen, 1+ptTLScreenL.x, 1+ptTLScreenL.y);

// Loop on all unvisited cells to see their content
nn:=0;
repeat
   inc(nn);
   nX:=abXFree[nn];
   nY:=abYFree[nn];

      // Get the content of the cell (nX, nY)
      nContent:=GetMineContent(hDCScreen, ptTLScreenL.x + (nX-1)*nXDelta,
                                          ptTLScreenL.y + (nY-1)*nYDelta);

      // To count the neighbour cell to be still explored
      if (nContent <> cteToSee) then begin
         // Content of this cell
         aabBO[nX,nY]:=aabBO[nX,nY] + nContent - cteToSee;

         if (nContent = cteBoom) then fBombExplosed:=TRUE; // Lost

         // Unvisited (=cteToSee) neighbours
         aabTS[nX,nY]:=aabTS[nX,nY] + 1;

         for ii:=nX-1 to nX+1 do  for jj:=nY-1 to nY+1 do
           aabTS[ii,jj]:=aabTS[ii,jj] - 1;

         // Copy the last cell of the array to the one which has been visited
         abXFree[nn]:=abXFree[nNbFreeCells];
         abYFree[nn]:=abYFree[nNbFreeCells];

         // One less unvisited cell
         dec(nNbFreeCells);

         dec(nn); // this cell is still unvisited
         end;

   until (nn >= nNbFreeCells) or fBombExplosed;

ReleaseDC(0, hDCScreen);

{
// To show what has been found in the squares
DBW(256, '------------------------------------------------------');
for nY:=1 to nYNumber do begin
   strS:='|';
   for nX:=1 to nXNumber do begin
      if (aabBO[nX,nY] >= cteBoom-9) then // A bomb
        strS:=strS + '*'
      else if (aabBO[nX,nY] >= cteToSee-9) then // An untouched square
        strS:=strS + '.'
      else if (aabBO[nX,nY] >= cteFlag-9) then // A flgs is in
        strS:=strS + ';'
      else
        strS:=strS + IntToStr(aabBO[nX,nY],1);

      strS:=strS + IntToStr(aabTS[nX,nY],1) + ' ';
      end; // for nX

   DBW(256, strS + '|');
   end; // for nY
DBW(256, 'Bombs left =' + ITS5(nNbBombs) + '  Unvisited=' + ITS5(nNbFreeCells));
}
end;  // TMyWindow.UpdateMineSweeperContent

procedure TMyWindow.PutFlags;
//===========================
// Put Flags on cells where it is sure that a bomb existe.
var
  ptTLScreenL : POINT; // top left in screen coordinate of the upper left mine square
  nn,
  nX,nY,
  ii,jj
              : integer;
begin
if (hWinSec = 0) then begin
   DialogExecMSG(20299, @Self);
   exit;
   end;

ptTLScreenL:=ptTopLeft;
ClientToScreen(hWinSec, ptTLScreenL);

// Loop on all unexplored cells, to see if they contain a bomb and
// thus a flag must be set.
nn:=0;
repeat
   inc(nn);
   nX:=abXFree[nn];
   nY:=abYFree[nn];

   if (aabTS[nX,nY] < 8) then begin
      // Loop on all neighbour to see if the cell has a bomb
      ii:=nX-2;
      repeat
         inc(ii);
         jj:=nY-2;
         repeat
            inc(jj);
            if (ii=nX) and (jj=nY) then inc(jj); // Skip the (nX,nY) cell

            if (aabBO[ii,jj] = aabTS[ii,jj]) then begin
               // the (nX,nY) cell has a bomb, mark it
//DBW(128, 'SetFlag' + ITS5(nX) + ITS5(nY));
               MoveMouseTo((nX-1)*nXDelta + ptTLScreenL.x+4,
                           (nY-1)*nYDelta + ptTLScreenL.y+4);
               SimulateRightClick;

               // Mark that neighbour has one more neighbour with a bomb
               // and one less unvisited neighbour
               aabTS[nX,nY]:=aabTS[nX,nY] + 1;
               for ii:=nX-1 to nX+1 do  for jj:=nY-1 to nY+1 do begin
                   aabBO[ii,jj]:=aabBO[ii,jj] - 1;
                   aabTS[ii,jj]:=aabTS[ii,jj] - 1;
                   end; // for

               aabBO[nX,nY]:=cteFlag;
               dec(nNbBombs);

               // Copy the last cell of the array to the one which has been visited
               abXFree[nn]:=abXFree[nNbFreeCells];
               abYFree[nn]:=abYFree[nNbFreeCells];

               // One less unvisited cell
               dec(nNbFreeCells);

               dec(nn); // this cell is still unvisited

               // no more searching necessary, quit the repeat loops
               ii:=nX+1;
               jj:=nY+1;
               end; // if
            until (jj > nY);
         until (ii > nX);
      end; // if (aabTS[nX,nY] < 8) then begin
   until (nn >= nNbFreeCells) or fQuitSolve;

// Nothing to update, it is already up to date.
//UpdateMineSweeperContent;
end;  // TMyWindow.PutFlags

procedure TMyWindow.StartingExplore;
//==================================
// When we beging, start with some special cells.
var
  ptTLScreenL : POINT; // top left in screen coordinate of the upper left mine square
  nX,nY,      // where to click
  nDX,nDY     // to go on cells neighbour of the previous
              : integer;
  fQuitRepeat : boolean;
begin
if (aabBO[1,1] <> cteToSee) then  exit;  // not an empty field

ptTLScreenL:=ptTopLeft;
ClientToScreen(hWinSec, ptTLScreenL);
ptTLScreenL.x:=ptTLScreenL.x+4 -nXDelta;
ptTLScreenL.y:=ptTLScreenL.y+4 -nYDelta;

// First cell to explore  (1,1)
nX:=1;   nY:=1;
nDX:=2;  nDY:=2;

// Try to find a good starting corner
fQuitRepeat:=FALSE;
repeat
   MoveMouseTo(nX*nXDelta + ptTLScreenL.x, nY*nYDelta + ptTLScreenL.y);
   SimulateLeftClick;
   UpdateMineSweeperContent;

   if (aabBO[nX,nY] = 0) then exit; // good start.
   if fBombExplosed then exit;

   if (aabBO[nX,nY] = 3) then begin
      // Click near the previous cell  (1,3)
      nY:=nY + nDY;
      MoveMouseTo(nX*nXDelta + ptTLScreenL.x, nY*nYDelta + ptTLScreenL.y);
      SimulateLeftClick;
      UpdateMineSweeperContent;
      if (aabBO[nX,nY] <= 2) then exit; // good second try
      if fBombExplosed then exit;

      if (aabBO[nX,nY] = 5) then begin
         // Click again near the previous cell  (1,5)
         nY:=nY + nDY;
         MoveMouseTo(nX*nXDelta + ptTLScreenL.x, nY*nYDelta + ptTLScreenL.y);
         SimulateLeftClick;
         UpdateMineSweeperContent;
         if (aabBO[nX,nY] <= 2) then exit; // good second try
         if fBombExplosed then exit;

         if (aabBO[nX,nY] = 5) then begin
            // Click again near the previous cell  (1,7)
            nY:=nY + nDY;
            MoveMouseTo(nX*nXDelta + ptTLScreenL.x, nY*nYDelta + ptTLScreenL.y);
            SimulateLeftClick;
            UpdateMineSweeperContent;
            if (aabBO[nX,nY] <= 2) then exit; // good second try
            if fBombExplosed then exit;
            end; // (1,7)
         end; // (1,5)

      // Click near the previous cell  (3,1)
      nX:=nX + nDX;
      nY:=nY - 3*nDY;
      MoveMouseTo(nX*nXDelta + ptTLScreenL.x, nY*nYDelta + ptTLScreenL.y);
      SimulateLeftClick;
      UpdateMineSweeperContent;
      if (aabBO[nX,nY] <= 2) then exit; // good second try
      if fBombExplosed then exit;

      if (aabBO[nX,nY] = 5) then begin
         // Click again near the previous cell  (5,1)
         nX:=nX + nDX;
         MoveMouseTo(nX*nXDelta + ptTLScreenL.x, nY*nYDelta + ptTLScreenL.y);
         SimulateLeftClick;
         UpdateMineSweeperContent;
         if (aabBO[nX,nY] <= 2) then exit; // good second try
         if fBombExplosed then exit;

         if (aabBO[nX,nY] = 5) then begin
            // Click again near the previous cell  (7,1)
            nX:=nX + nDX;
            MoveMouseTo(nX*nXDelta + ptTLScreenL.x, nY*nYDelta + ptTLScreenL.y);
            SimulateLeftClick;
            UpdateMineSweeperContent;
            if (aabBO[nX,nY] <= 2) then exit; // good second try
            if fBombExplosed then exit;
            end; // (7,1)
         end; // (5,1)
      end; // (3,1)

   // Try in another corner
   if (nY = 1) then begin
      nY:=nYNumber;
      nDY:=-2;
      end
   else if (nX = 1) then begin
      nX:=nXNumber;
      nDX:=-2;
      nY:=1;
      nDY:= 2;
      end
   else begin
      nX:=nX + nDX;
      MoveMouseTo(nX*nXDelta + ptTLScreenL.x, nY*nYDelta + ptTLScreenL.y);
      SimulateLeftClick;
      UpdateMineSweeperContent;

      fQuitRepeat:=TRUE;
      end;

   until fQuitRepeat;

// Here other tries could be tempted.

end;  // TMyWindow.StartingExplore

function TMyWindow.SaftyExplore:integer;
//======================================
// Click on cells which have no bombs for sure
// Return the number of cells which have been clicked.
var
  ptTLScreenL : POINT; // top left in screen coordinate of the upper left mine square
  nCount,     // Count how many cells have been clicked
  nn,
  nX,nY,
  ii,jj       : integer;
begin
if (hWinSec = 0) then begin
   DialogExecMSG(20299, @Self);
   exit;
   end;

ptTLScreenL:=ptTopLeft;
ClientToScreen(hWinSec, ptTLScreenL);

nCount:=0;

// Loop on all unexplored cells, to see if they contain a bomb and
// thus a flag must be set.
nn:=0;
repeat
   inc(nn);
   nX:=abXFree[nn];
   nY:=abYFree[nn];

   if (aabTS[nX,nY] < 8) then begin
      // Loop on all neighbour to see if the cell has no bomb for sure
      ii:=nX-2;
      repeat
         inc(ii);
         jj:=nY-2;
         repeat
            inc(jj);
            if (ii=nX) and (jj=nY) then inc(jj); // Skip the (nX,nY) cell

            if (aabBO[ii,jj] = 0) then begin
               // neighbour of this cell have no bombs for sure
               // Particulary the cell (nX, nY) has no bomb
//DBW(128, 'Explore' + ITS5(nX) + ITS5(nY));

               MoveMouseTo((nX-1)*nXDelta + ptTLScreenL.x+4,
                           (nY-1)*nYDelta + ptTLScreenL.y+4);
               SimulateLeftClick;
               inc(nCount);

               // no more searching necessary, quit the repeat loops
               ii:=nX+1;
               jj:=nY+1;
               end; // if
            until (jj > nY);
         until (ii > nX);
      end; // if (aabTS[nX,nY] < 8) then begin
   until (nn >= nNbFreeCells) or fQuitSolve;

// The minesweeper window changed, update the information we have about it.
if (nCount > 0) then UpdateMineSweeperContent;
SaftyExplore:=nCount;
end;  // TMyWindow.SaftyExplore

procedure TMyWindow.SaveValue(pnValue:Pinteger);
//==============================================
// Save in  apnModif[nIndModif]  and anModifValue[nIndModif]
// the pointer  pnValue,  and the content of pnValue^
begin
inc(nIndModif);
apnModif[nIndModif]:=pnValue;
anModifValue[nIndModif]:=pnValue^;
end;  // TMyWindow.SaveValue

function TMyWindow.SetFlagTo(nXP,nYP:integer):integer;
//====================================================
label BeginWithNoBomb;
var
  nX,nY,
  nX0,nY0,
  ii,jj,
  nn
              : integer;
  anXmem,anYmem // to memorise which cells have been modified
              : array[1..100] of integer;
  nInd0,
  nInd1,
  nInd2       : integer; // number of modified cells

begin
SetFlagTo:=100; // Default return value, indicating no contradiction found

// First cell, being marked having a bomb
anXmem[1]:=abs(nXP);    anYmem[1]:=nYP;
nInd1:=0;
nInd2:=1;

//DBW(1, 'Start With :' + ITS5(nXP) + ITS5(nYP) + '*******************');

if (nIndModif >= 0) then begin
   SaveValue(@aabBObis[abs(nXP),nYP]);
   dec(nNbFreeCellsBis);
   if (nXP >= 0) then dec(nNbBombsBis);
   end;

if (nXP < 0) then begin
   // Code to say to begin with a cell marked with no bomb
   nXP:=-nXP;
   aabBObis[nXP,nYP]:=cteNoBomb;
   goto BeginWithNoBomb;
   end
else aabBObis[nXP,nYP]:=cteFlag;

repeat
   // Treat all cells being marked having a bomb
   //""""""""""""""""""""""""""""""""""""""""""""
   nInd0:=nInd1;
   nInd1:=nInd2;
   for nn:=nInd0+1 to nInd1 do begin
      nX0:=anXmem[nn];
      nY0:=anYmem[nn];

      for nX:=nX0-1 to nX0+1 do  for nY:=nY0-1 to nY0+1 do
       if (nX <> nX0) or (nY <> nY0) then begin // Skip the cell (nX0,nY0)
         if (aabBObis[nX,nY] = 0) then begin
            // Contradiction found, the cell (nXP, nYP) cannot have a bomb
//DBW(8*1024, '>< BO=0 ' + ITS5(nX) + ITS5(nY) + ITS5(aabBObis[nX,nY]) + ITS5(aabTSbis[nX,nY]));
            SetFlagTo:=-1;
            exit;
            end;

         // Neighbour has one less cell with a bomb
         // and one less unvisited cell
         if (nIndModif >= 0) then begin
            SaveValue(@aabBObis[nX,nY]);
            SaveValue(@aabTSbis[nX,nY]);
            end;
         aabBObis[nX,nY]:=aabBObis[nX,nY] - 1;
         aabTSbis[nX,nY]:=aabTSbis[nX,nY] - 1;

         if (aabBObis[nX,nY] = 0) then begin
            // All unvisited neighbour of (nX,nY) have no bombs
            // Memorise them to treat them later
            for ii:=nX-1 to nX+1 do for jj:=nY-1 to nY+1 do
             if ((ii <> nX) or (jj <> nY)) and
                (aabBObis[ii,jj] >= cteToSee-9) then begin
               if (nInd2 > 100) then exit; // search too long
               inc(nInd2);
               anXmem[nInd2]:=ii; // Marked as having NO bomb
               anYmem[nInd2]:=jj;

               if (nIndModif >= 0) then begin
                  SaveValue(@aabBObis[ii,jj]);
                  dec(nNbFreeCellsBis);
                  end;
               aabBObis[ii,jj]:=cteNoBomb;
//DBW(16*1024, 'marked. has NO bomb' + ITS5(ii) + ITS5(jj));
               end; // for .. if
            end; // if (aabBObis[nX,nY] = 0) then begin
         end; // for nX:=...
      end; // for nn:=...

BeginWithNoBomb:;
   // Treat all cells having been marked without bombs
   //"""""""""""""""""""""""""""""""""""""""""""""""""
   nInd0:=nInd1;
   nInd1:=nInd2;
   for nn:=nInd0+1 to nInd1 do begin
      nX0:=anXmem[nn];
      nY0:=anYmem[nn];

      for nX:=nX0-1 to nX0+1 do  for nY:=nY0-1 to nY0+1 do
       if (nX <> nX0) or (nY <> nY0) then begin
         if (aabBObis[nX,nY] = aabTSbis[nX,nY]) then begin
            // Contradiction...
//DBW(8*1024, '>< BO=TS ' + ITS5(nX) + ITS5(nY) + ITS5(aabBObis[nX,nY]) + ITS5(aabTSbis[nX,nY]));
            SetFlagTo:=-1;
            exit;
            end;

         if (nIndModif >= 0) then begin
            SaveValue(@aabTSbis[nX,nY]);
            end;
         aabTSbis[nX,nY]:=aabTSbis[nX,nY] -1;

         if (aabBObis[nX,nY] = aabTSbis[nX,nY]) then begin
            // all unvisited neighbour cell have a bomb.
            // Memorise them to treat them later
            for ii:=nX-1 to nX+1 do for jj:=nY-1 to nY+1 do
             if ((ii <> nX) or (jj <> nY)) and
                (aabBObis[ii,jj] >= cteToSee-9) then begin
               if (nInd2 > 100) then exit; // search too long
               inc(nInd2);
               anXmem[nInd2]:=ii; // Marked as having a bomb
               anYmem[nInd2]:=jj;

               if (nIndModif >= 0) then begin
                  SaveValue(@aabBObis[ii,jj]);
                  dec(nNbFreeCellsBis);
                  if (nXP >= 0) then dec(nNbBombsBis);
                  end;
               aabBObis[ii,jj]:=cteFlag;
//DBW(16*1024, 'marked. has bomb' + ITS5(ii) + ITS5(jj));
               end;  // for .. if
            end; // if (aabBObis[nX,nY] = aabTSbis[nX,nY])
         end; // for nX:=...
      end; // for nn:=...

   until (nInd2 = nInd1);

SetFlagTo:=nInd2;
end;  // TMyWindow.SetFlagTo

function TMyWindow.SaftyExploreMore:integer;
//==========================================
// Click on cells which have no bombs for sure
// But the searching of these cells is more sophisticated than SaftyExplore
var
  ptTLScreenL : POINT; // top left in screen coordinate of the upper left mine square
  nn,         // To get at random an unvisited cell
  nCount,
  ii,jj,
  nXL,nYL     : integer;
begin
nCount:=0; // If no more bombs has been found

ptTLScreenL:=ptTopLeft;
ClientToScreen(hWinSec, ptTLScreenL);

// Try to find cells having NO bomb for sure.
nn:=0;
repeat
   inc(nn);
   nXL:=abXFree[nn];
   nYL:=abYFree[nn];

   if (aabTS[nXL,nYL] <> 8) then begin
      aabBObis:=aabBO;
      aabTSbis:=aabTS;
      // Try to see if there is a contradiction setting a flag there
      if (SetFlagTo(nXL,nYL) = -1) then begin
//         DBW(1024, 'cell has no bomb nXL, nYL=' + ITS5(nXL) + ITS5(nYL));

         // Click on this cell
         MoveMouseTo((nXL-1)*nXDelta + ptTLScreenL.x+4,
                     (nYL-1)*nYDelta + ptTLScreenL.y+4);
         SimulateLeftClick;

         nCount:=1; // Indicate that a bomb has been found
         end;
      end;
   until (nn >= nNbFreeCells);

// If nothing found, Try to find cells having a bomb for sure.
nn:=0;
if (nCount = 0) then repeat
   inc(nn);
   nXL:=abXFree[nn];
   nYL:=abYFree[nn];

   if (aabTS[nXL,nYL] <> 8) then begin
      aabBObis:=aabBO;
      aabTSbis:=aabTS;
      // Try to see if there is a contradiction setting a cteNoBomb there
      if (SetFlagTo(-nXL,nYL) = -1) then begin
//         DBW(2048, 'cell has A bomb nXL, nYL=' + ITS5(nXL) + ITS5(nYL));

         // Click on this cell
         MoveMouseTo((nXL-1)*nXDelta + ptTLScreenL.x+4,
                     (nYL-1)*nYDelta + ptTLScreenL.y+4);
         SimulateRightClick;

               // Mark that neighbour has one more neighbour with a bomb
               // and one less unvisited neighbour
               aabTS[nXL,nYL]:=aabTS[nXL,nYL] + 1;
               for ii:=nXL-1 to nXL+1 do  for jj:=nYL-1 to nYL+1 do begin
                   aabBO[ii,jj]:=aabBO[ii,jj] - 1;
                   aabTS[ii,jj]:=aabTS[ii,jj] - 1;
                   end; // for

               aabBO[nXL,nYL]:=cteFlag;
               dec(nNbBombs);

               // Copy the last cell of the array to the one which has been visited
               abXFree[nn]:=abXFree[nNbFreeCells];
               abYFree[nn]:=abYFree[nNbFreeCells];

               // One less unvisited cell
               dec(nNbFreeCells);

               dec(nn); // this cell is still unvisited

         nCount:=1; // Indicate that a bomb has been found
         end;
      end;
   until (nn >= nNbFreeCells) or fQuitSolve;

if (nCount > 0) then UpdateMineSweeperContent;
SaftyExploreMore:=nCount;
end;  // TMyWindow.SaftyExploreMore

procedure TMyWindow.SortFreeCells{(nFirstP,nLastP,nPivotP:integer)};
//================================================================
// Sort the abXFree,abYFree arrays such that cells having few unvisited
// neighbour come first.
var
  nTemp,
  nL,nR       : integer;
  fFoundL     : boolean;

begin
// First sorting, separat cells having a neighbour containing information
// from cells having no neighbour containing information
nL:=1;
nR:=nNbFreeCells;

repeat
   dec(nL);
   repeat
      fFoundL:=FALSE;
      inc(nL);

      // Search a cell having no neighbour containing information, i.e.
      // having only unvisited neighbours
      if (nL > nNbFreeCells) then fFoundL:=TRUE
      else if (abXFree[nL] > 1) and (abXFree[nL] < nXNumber) then begin
         // Not the first neither the last line
         if (abYFree[nL] > 1) and (abYFree[nL] < nYNumber) then begin
            // Not the first neither the last column
            if (aabTS[abXFree[nL],abYFree[nL]] =
                8 + aabBO[abXFree[nL],abYFree[nL]] - cteToSee) then fFoundL:=TRUE;
            end
         else begin // In the first or last column
            if (aabTS[abXFree[nL],abYFree[nL]] =
                5 + aabBO[abXFree[nL],abYFree[nL]] - cteToSee) then fFoundL:=TRUE;
            end;
         end
      else begin
         // In the first or last line
         if (abYFree[nL] > 1) and (abYFree[nL] < nYNumber) then begin
            // Not the first neither the last column
            if (aabTS[abXFree[nL],abYFree[nL]] =
                5 + aabBO[abXFree[nL],abYFree[nL]] - cteToSee) then fFoundL:=TRUE;
            end
         else begin // In the corners
            if (aabTS[abXFree[nL],abYFree[nL]] =
                3 + aabBO[abXFree[nL],abYFree[nL]] - cteToSee) then fFoundL:=TRUE;
            end;
         end;
      until fFoundL;

   inc(nR);
   repeat
      fFoundL:=TRUE;
      dec(nR);

      // Search a cell having a neighbour containing information
      if (nR < 1) then fFoundL:=TRUE
      else if (abXFree[nR] > 1) and (abXFree[nR] < nXNumber) then begin
         // Not the first neither the last line
         if (abYFree[nR] > 1) and (abYFree[nR] < nYNumber) then begin
            // Not the first neither the last column
            if (aabTS[abXFree[nR],abYFree[nR]] =
                8 + aabBO[abXFree[nR],abYFree[nR]] - cteToSee) then fFoundL:=FALSE;
            end
         else begin // In the first or last column
            if (aabTS[abXFree[nR],abYFree[nR]] =
                5 + aabBO[abXFree[nR],abYFree[nR]] - cteToSee) then fFoundL:=FALSE;
            end;
         end
      else begin
         // In the first or last line
         if (abYFree[nR] > 1) and (abYFree[nR] < nYNumber) then begin
            // Not the first neither the last column
            if (aabTS[abXFree[nR],abYFree[nR]] =
                5 + aabBO[abXFree[nR],abYFree[nR]] - cteToSee) then fFoundL:=FALSE;
            end
         else begin // In the corners
            if (aabTS[abXFree[nR],abYFree[nR]] =
                3 + aabBO[abXFree[nL],abYFree[nL]] - cteToSee) then fFoundL:=FALSE;
            end;
         end;
      until fFoundL;

   if (nL < nR) then begin
      // Exange the index, to have first, cells having neighbour containing information
      nTemp       :=abXFree[nR];
      abXFree[nR] :=abXFree[nL];
      abXFree[nL] :=nTemp;
      nTemp       :=abYFree[nR];
      abYFree[nR] :=abYFree[nL];
      abYFree[nL] :=nTemp;
      inc(nL);
      dec(nR);
      end;

   until (nL > nR);

// Number of interresting cells. No deduction can be maide on the others
// without taking the number of bombs into account.
nNbFreeCellsInfo:=nR;

//DBW(1, 'nbFree=' + ITS5(nNbFreeCellsInfo) + ITS5(nNbFreeCells));

// Indexes from  1  to  nR  are from cells having neighbour containing information
{for nL:=1 to nR do begin
   DBW(1, 'info:' + ITS5(abXFree[nL]) + ITS5(abYFree[nL])
                  + ITS5(aabTS[abXFree[nL],abYFree[nL]]));
   end;
}
end;  // TMyWindow.SortFreeCells

function TMyWindow.ProbabilityExplore:integer;
//============================================
// Explore the cells to find which has the lower probabilities of
// having a bomb. Take into account the number of bombs.
// In some case, this probability will be zero, and thus
// it is sure that a cell without bomb will have been found.
var
  ptTLScreenL : POINT; // top left in screen coordinate of the upper left mine square
  nCount,
  nn,         // To get at random an unvisited cell
  nRep,       // Result from SetFlag
  ii,
  nXL,nYL     : integer;

  anLimit : array[1..200] of integer;
  // All points during the backtracking where a decision has been made.

  nIndLimit : integer;
  // Index in the  anLimit  array

  avNbConfigWI : array[0..500] of Float;  // WI = Without Info.
  // Number of configuration having  k  bombs among the cells having no Info.
  // A cell has no Info, if it has no visited neighbour without bombs.

  avNbWithFlag : array[1..2500] of Float;
  // For each cell number of configuration found with the cell
  // having a bomb

  vTotalConfig : Float;
  // Total number of possible configurations found

  vNbConf : Float;
  // Number of configurations found, during one trial.

  nNbCellWI : integer;
  // Number of cells witout information.

  nNbConfigSum : int32;
  // Number of time a config has been found for cells with info.

  nSetFlag    : integer;
  // Determine it a  cteFlag  or a  cteNoBomb  is set to a cell
  // 1 => cteFlag,  0 => cteNoBomb

  vProbability : Float;
  // To give the probability information of being a cell with bomb

begin
SortFreeCells;


if (nNbFreeCellsInfo = 0) {or (nNbFreeCells >= nXNumber*nYNumber-1)}  then begin
   vProbability:=9999;
   ptTLScreenL:=ptTopLeft;
   ClientToScreen(hWinSec, ptTLScreenL);

   // No Free Cells having information to decide where to click.
   if (aabBO[1,1] >= cteToSee-9) then
      MoveMouseTo(0*nXDelta + ptTLScreenL.x+4, 0*nYDelta + ptTLScreenL.y+4)
   else if (aabBO[1,nYNumber] >= cteToSee-9) then
      MoveMouseTo(0*nXDelta + ptTLScreenL.x+4, (nYNumber-1)*nYDelta + ptTLScreenL.y+4)
   else if (aabBO[nXNumber,1] >= cteToSee-9) then
      MoveMouseTo((nXNumber-1)*nXDelta + ptTLScreenL.x+4, 0*nYDelta + ptTLScreenL.y+4)
   else if (aabBO[nXNumber,nYNumber] >= cteToSee-9) then
      MoveMouseTo((nXNumber-1)*nXDelta + ptTLScreenL.x+4, (nYNumber-1)*nYDelta + ptTLScreenL.y+4)
   else if (aabBO[1,3] >= cteToSee-9) then
      MoveMouseTo(0*nXDelta + ptTLScreenL.x+4, 2*nYDelta + ptTLScreenL.y+4)
   else if (aabBO[3,1] >= cteToSee-9) then
      MoveMouseTo(2*nXDelta + ptTLScreenL.x+4, 0*nYDelta + ptTLScreenL.y+4)
   else
      MoveMouseTo((abXFree[1]-1)*nXDelta + ptTLScreenL.x+4, (abYFree[1]-1)*nYDelta + ptTLScreenL.y+4);

   SimulateLeftClick;

   if (nNbFreeCells > 0) then vProbability:=nNbBombs / Float(nNbFreeCells);
   // else not normal

//DBW(128, 'Tryed on' + ITS5(abXFree[1]) + ITS5(abYFree[1]));
DBW(128, 'With bomb probability =' + FTS(vProbability) + ITS5(nNbFreeCells));

   vProbProduct:=vProbProduct * (1.0 - vProbability);

   UpdateMineSweeperContent;
   ProbabilityExplore:=1;
   exit;
   end;

// If there are only few cells without neighbour containing information,
// take them into account, since the number of bombs can make a difference.
// "5" is arbitrary.
if (nNbFreeCells - nNbFreeCellsInfo <= 8) or
   (nNbBombs <= 4) then nNbFreeCellsInfo:=nNbFreeCells;

aabBObis:=aabBO;
aabTSbis:=aabTS;

// Number of cells without information.
nNbCellWI:=nNbFreeCells - nNbFreeCellsInfo;

// Initialisation of the number of configuration of cells without info,
// having  nn  bombs among them.
nNbBombsBis:=nNbBombs;
if (nNbBombsBis > nNbCellWI) then nNbBombsBis:=nNbCellWI;
if (nNbBombsBis > 500) then nNbBombsBis:=500;  // Just to avoid a crash in stupid cases
avNbConfigWI[0]:=1.0;  // only one config having no bombs at all.
for nn:=1 to nNbBombsBis do
   avNbConfigWI[nn]:=avNbConfigWI[nn-1]*(nNbCellWI - nn) / nn;

// Initialisation of the counter of how many configuration exist with
// bombs, for each cell
for nn:=1 to nNbFreeCellsInfo do begin
   avNbWithFlag[nn]:=0;
   end;
vTotalConfig:=1e-9; // to avoid dividing by 0

// Will be used to limit the number of tries to find configurations.
// Usefull if too many configurations are possible
nNbConfigSum:=0;

// Start with the first unvisited cell of the list.
nXL:=abXFree[1];
nYL:=abYFree[1];

// At the begining, we know the real number of free cells and free bombs.
nNbFreeCellsBis:=nNbFreeCellsInfo;
nNbBombsBis:=nNbBombs;

// Starting of the backtracking
nSetFlag:=1;
nIndModif:=0;
nIndLimit:=1;
anLimit[nIndLimit]:=nIndModif+1; // Mark the first level
// Begin backtracking loop
//"""""""""""""""""""""""""
repeat
//DBW(1, '-------------------------');
//DBW(1, 'SetFlag:' + ITS5(nSetFlag) + ITS5(nXL) + ITS5(nYL)
//                  + ITS5(nIndLimit) + ITS5(nIndModif));
//DBW(1, 'nbBomb + free:' + ITS5(nNbBombsBis) + ITS5(nNbFreeCellsBis));

   SaveValue(@nSetFlag);  // to memorise if we set cteFlag or cteNoBomb at starting
   SaveValue(@nXL);
   SaveValue(@nYL);
   SaveValue(@nNbFreeCellsBis);
   SaveValue(@nNbBombsBis);
   if (nSetFlag=1) then nRep:=SetFlagTo( nXL,nYL)
   else                 nRep:=SetFlagTo(-nXL,nYL);

   // if (no more free cells) and (nRep >= 0) and (nbBombs=0) then begin
   if (nNbFreeCellsBis <= 0) and (nRep >= 0) and
      ((nNbBombsBis = 0) or (nNbFreeCellsInfo < nNbFreeCells)) then begin
      // More possible configurations of bombs has been found.
      // Number of bombs left among the cells without info.
      vNbConf:=avNbConfigWI[nNbBombsBis];
      inc(nNbConfigSum);

      // Memorise the results
      vTotalConfig:=vTotalConfig + vNbConf;

//DBW(64, 'Total config=' + ITS5(nNbConfigSum) + ITS5(nNbBombsBis));
      for ii:=1 to nNbFreeCellsInfo do begin
         if (aabBObis[abXFree[ii],abYFree[ii]] <= cteFlag) then
            avNbWithFlag[ii]:=avNbWithFlag[ii] + vNbConf;
         end;
     end;
    // if (nRep = -1) then Contradiction found, thus this case is not taken into account

//if (nRep = -1) or (nNbBombsBis < 0) then
//  DBW(1, '>< nRep, bomb=' + ITS5(nRep) + ITS5(nNbBombsBis));


   if (nRep >= 0) and (nNbFreeCellsBis > 0) then begin
      // Take the next free cell
      ii:=0;
      repeat
         inc(ii);
         until (ii > nNbFreeCellsInfo) or
               (aabBObis[abXFree[ii],abYFree[ii]] >= cteToSee-9);

      if (ii > nNbFreeCellsInfo) then begin
         // Error, this should have been taken into account
         nRep:=-1;
DBW(1, 'ERROR ***** ii=' + ITS5(ii) + ITS5(nNbFreeCellsInfo));
         end
      else begin
         // Give a state to the next unvisited cell
         inc(nIndLimit);   // Memorise the state of the modifications
         anLimit[nIndLimit]:=nIndModif+1;
         nXL:=abXFree[ii];
         nYL:=abYFree[ii];
         nSetFlag:=ord(nNbBombsBis > 0); // If no more bombs, only cells without bombs are possible
         // Continue the exploration setting this cell to have a bomb (= a Flag)
         // or no bombs, if no more bombs left
         end;
      end
   else nRep:=-1;  // indicate that has to come back in the backtracking

   if (nRep < 0) then begin
      // Contradiction found, or finished
      // go back in the backtracking
      //"""""""""""""""""""""""""""""
      // Set back their original values to aabBObis and aabTObis
      // Also to the number of bombs and free cells
//DBW(1, 'Backtracking nIndModif=:' + ITS5(nIndModif) + ITS5(nIndLimit));
      repeat
         while (nIndModif > anLimit[nIndLimit]) do begin
            apnModif[nIndModif]^:=anModifValue[nIndModif];
            dec(nIndModif);
            end;
//DBW(4*1024, 'After BT, bomb, free=' + ITS5(nNbBombsBis) + ITS5(nNbFreeCellsBis));
//DBW(4*1024, '     setFlag=' + ITS5(anModifValue[nIndModif]) + ITS5(nIndModif) + ITS5(nIndLimit));

         nSetFlag:=anModifValue[nIndModif];
         dec(nIndModif);
         if (nSetFlag = 0) then begin
            // Has to do more Backtracking
            dec(nIndLimit);
            end;
         until (nSetFlag <> 0) or (nIndLimit <= 0);

//for ii:=1 to nNbFreeCellsInfo do
//  DBW(2048, 'cells=' + ITS5(abXFree[ii]) + ITS5(abYFree[ii])
//                  + ITS5(aabBObis[abXFree[ii],abYFree[ii]]));


      if (nSetFlag = 1) then begin
         // Try know setting this cell to no bomb.
         nSetFlag:=0;
         nXL:=anModifValue[nIndModif+2];
         nYL:=anModifValue[nIndModif+3];
//            inc(nIndLimit);   // Memorise the state of the modifications
//            anLimit[nIndLimit]:=nIndModif+1;  done already
         end
      else if (nIndLimit = 0) then nIndLimit:=-1; // quit the loop
      end;
   until (nIndLimit < 0) or (nNbConfigSum > 80000) or
         (GetAsyncKeyState(VK_SHIFT) < 0);

// Here we have in  anNbWithFlag
// the number of configurations with bombs
// and in .. the total number of possible configurations

DBW(32, 'Total config=' + ITS5(nNbConfigSum));
for ii:=1 to nNbFreeCellsInfo do
   DBW(32, 'flags on :' + ITS5(abXFree[ii]) + ITS5(abYFree[ii]) +
                         FTS(avNbWithFlag[ii] / vTotalConfig));

// Click on cells without bomb for sure
nCount:=0;
ptTLScreenL:=ptTopLeft;  ClientToScreen(hWinSec, ptTLScreenL);
for ii:=1 to nNbFreeCellsInfo do begin
    if (avNbWithFlag[ii] = 0) then begin
       // This cell has no bomb for sure
       MoveMouseTo((abXFree[ii]-1)*nXDelta + ptTLScreenL.x+4,
                   (abYFree[ii]-1)*nYDelta + ptTLScreenL.y+4);
       SimulateLeftClick;
       inc(nCount);
       end;
    end;

// If still no cells without bomb has been found,
// go on a cell having a lower probability of having a bomb
if (nCount = 0) then begin
   vProbability:=9999;
   nn:=1;
   for ii:=2 to nNbFreeCellsInfo do
       if (avNbWithFlag[nn] > avNbWithFlag[ii]) then nn:= ii;

  if (nNbFreeCellsInfo < nNbFreeCells) then begin
     // Test if clicking on a cell having no neighbour with information
     // is not better
     if (avNbWithFlag[nn] * nNbFreeCells  > nNbBombs * vTotalConfig) then begin
        // other cells have lower probabilities of having a bomb
        nn:=nNbFreeCellsInfo+1; // Choose on at random
        if (nNbFreeCells > 0) then vProbability:=nNbBombs / Float(nNbFreeCells);
        // else not normal

        // Among the other cells, click on one having as few neighbour as possible
        // and being as near as possible from the other cells.

        // Try do do better here ?!?
        if (aabBO[1,nYNumber] >= cteToSee-2) and
           (aabTS[1,nYNumber] = 3 - (aabBO[1,nYNumber]-cteToSee)) then begin
           // Cells have only flags or not visited neighbours
           nn:=0;
           MoveMouseTo(                       ptTLScreenL.x+4,
                       (nYNumber-1)*nYDelta + ptTLScreenL.y+4);
           end
        else if (aabBO[nXNumber,1] >= cteToSee-2) and
           (aabTS[nXNumber,1] = 3 - (aabBO[nXNumber,1]-cteToSee)) then begin
           nn:=0;
           MoveMouseTo((nXNumber-1)*nXDelta + ptTLScreenL.x+4,
                                              ptTLScreenL.y+4);
           end
        else if (aabBO[nXNumber,nYNumber] >= cteToSee-2) and
           (aabTS[nXNumber,nYNumber] = 3 - (aabBO[nXNumber,nYNumber]-cteToSee)) then begin
           nn:=0;
           MoveMouseTo((nXNumber-1)*nXDelta + ptTLScreenL.x+4,
                       (nYNumber-1)*nYDelta + ptTLScreenL.y+4);
           end

        else begin
           // Try on the top line
           nXL:=1;
           repeat
              inc(nXL);
              if (aabBO[nXL,1] >= cteToSee-4) and
                 (aabTS[nXL,1] = 5 - (aabBO[nXL,1]-cteToSee)) then begin
                 nn:=0;
                 MoveMouseTo((nXL-1)*nXDelta + ptTLScreenL.x+4,
                                             + ptTLScreenL.y+4);
                 nXL:=nXNumber;  // Quit the repeat ... until
                 end;
              until (nXL >= nXNumber-1);

           if (nn > 0) then begin
              // Try on the bottom line
              nXL:=1;
              repeat
                 inc(nXL);
                 if (aabBO[nXL,nYNumber] >= cteToSee-4) and
                    (aabTS[nXL,nYNumber] = 5 - (aabBO[nXL,nYNumber]-cteToSee)) then begin
                    nn:=0;
                    MoveMouseTo((nXL-1)*nXDelta + ptTLScreenL.x+4,
                                (nYNumber-1)*nYDelta + ptTLScreenL.y+4);
                    nXL:=nXNumber; // Quit the repeat ... until
                    end;
                 until (nXL >= nXNumber-1);
              end;
           end; // Better ?!?

        end
     else begin
        if (vTotalConfig >= 1) then vProbability:=avNbWithFlag[nn] / vTotalConfig
        else if (nNbFreeCells > 0) then vProbability:=nNbBombs / Float(nNbFreeCells);
        // else not normal
        end;
     end
  else if (vTotalConfig >= 1) then vProbability:=avNbWithFlag[nn] / vTotalConfig;

  if (nNbFreeCellsInfo = 0) then begin // first click on an unexplored window
     if (nNbFreeCells > 0) then vProbability:=nNbBombs / Float(nNbFreeCells);
     // else not normal or no more free cells
     end;

DBW(4, 'Tryed on' + ITS5(abXFree[nn]) + ITS5(abYFree[nn]) +
                  '  nbExplored=' + ITS5(nXNumber*nYNumber - nNbFreeCells));
DBW(4, 'With bomb probability =' + FTS(vProbability) + ITS5(nNbConfigSum));

//   fSolveFast:=FALSE;

   if (nn > 0) then MoveMouseTo((abXFree[nn]-1)*nXDelta + ptTLScreenL.x+4,
                                (abYFree[nn]-1)*nYDelta + ptTLScreenL.y+4);
   SimulateLeftClick;
   inc(nCount);

   vProbProduct:=vProbProduct * (1.0 - vProbability);
   end;

if (nCount > 0) then UpdateMineSweeperContent;
ProbabilityExplore:=nCount;
end;  // TMyWindow.ProbabilityExplore

procedure TMyWindow.SolveMineSweeper;
//===================================
// Go and click to solve the minesweeper
// This will be better done in  tEx064.pp
var
  ptTLScreenL : POINT; // top left in screen coordinate of the upper left mine square
  nSecurity,
  nn,         // To get at random an unvisited cell
  nXL,nYL     : integer;
begin
if (hWinSec = 0) then begin
   DialogExecMSG(20299, @Self);
   exit;
   end;

// Screen position of the cells
ptTLScreenL:=ptTopLeft;
ClientToScreen(hWinSec, ptTLScreenL);
SetCursorPos(ptTLScreenL.x+50, ptTLScreenL.y-40);

// A first click can be usefull
MoveMouseTo(ptTLScreenL.x-4, ptTLScreenL.y-4);
SimulateLeftClick;

{
// First cell at random
            nn:=1+Random(nNbFreeCells);
            nXL:=abXFree[nn];
            nYL:=abYFree[nn];

            MoveMouseTo((nXL-1)*nXDelta + ptTLScreenL.x+4, (nYL-1)*nYDelta + ptTLScreenL.y+4);
            SimulateLeftClick;
}
// Get the cells content.
GetMineSweeperContent(0);
fQuitSolve:=FALSE;
fBombExplosed:=FALSE;

// Product of the probabilities of clicking on a bomb.
vProbProduct:=1.0;

// If field empty, start exploring.
if (nNbFreeCells = nXNumber*nYNumber) then begin
   StartingExplore;
   if fBombExplosed then exit;
   end;

// Loop to solve
nSecurity:=500;
repeat
   // Try to put flags on cells having bombs
   PutFlags;   if fQuitSolve then exit;

   // Try to desactivate bombs. If none is found, choose one cell at random
   if (SaftyExplore = 0) and (not fQuitSolve) then begin
      // Try a more sophisticated way to find bombs
      if (nNbFreeCells <= 0) then exit; // finished
DBW(8, 'Explore more');
      if (SaftyExploreMore = 0) and (not fQuitSolve) then begin
         if (nNbFreeCells <= 0) then exit; // finished
DBW(8, 'Probability Explore');

         // Try solving using probabilities
         if (ProbabilityExplore = 0) and (not fQuitSolve) then begin
            if (nNbFreeCells <= 0) then exit; // finished
DBW(1, 'No cells found without a bomb');
exit;

            nn:=1+Random(nNbFreeCells);
            nXL:=abXFree[nn];
            nYL:=abYFree[nn];

            MoveMouseTo((nXL-1)*nXDelta + ptTLScreenL.x+4, (nYL-1)*nYDelta + ptTLScreenL.y+4);
            SimulateLeftClick;
            UpdateMineSweeperContent;
DBW(512, 'Tried at random :' + ITS5(nXL) + ITS5(nYL));
            end; // if (ProbabilityExplore = 0)
         end; // if (SaftyExploreMore = 0) then begin
      end; // if (SaftyExplore = 0) then begin

   dec(nSecurity);
   until fBombExplosed or (nNbFreeCells <= 0) or (nSecurity <= 0) or
         fQuitSolve;
end;  // TMyWindow.SolveMineSweeper

procedure TMyWindow.GetStatistiquePlay;
//=====================================
// Plays many Minesweeper games to get statistics on it.
var
  nNbGames,  // number of played games
  nNbWin     // number of times we win
             : int32;
  ch         : char;
  vMeanProb  : Float;
//  nLost1     : int32;

begin
nNbGames :=0;
nNbWin   :=0;
vMeanProb:=0;
//nLost1   :=0;
repeat
//   fSolveFast:=TRUE;
   ClickOnNewGame;
   SolveMineSweeper;

   if not fQuitSolve then begin
      inc(nNbGames);
      vMeanProb:=vMeanProb + vProbProduct;
{
      if (nNbFreeCells >= nXNumber * nYNumber -2) then begin
         inc(nLost1);
DBW(1, 'nbFree=' + ITS5(nNbFreeCells) + FTS(nLost1 / Float(nNbGames)));
         end;
}
      if (not fBombExplosed) then begin
          inc(nNbWin);
          ch:='*';
          vMeanProb:=vMeanProb + vProbProduct;
          end
      else begin
         ch:=' ';
         if not fSolveFast then DBWaitSec(2); // wait 2 seconds to let us see
         end;

      DBW(1, 'prob=' + ITS(nNbGames) + ch + FTS(nNbWin / Float(nNbGames)));

      end;

   until {fQuitSolve or} (GetAsyncKeyState(VK_CONTROL) < 0);

SetActiveWindow(hWin);
SetFocus(hWin);
SetForegroundWindow(hWin);
SetActiveWindow(hWin);
SetFocus(hWin);
SetForegroundWindow(hWin);
end;  // TMyWindow.GetStatistiquePlay

function TMyWindow.WMCommand(wPar:WPARAM; lPar:LPARAM):LRESULT;
//=============================================================
// Handles the menu messages
Var wNrMenu     : word;
begin
WMCommand := 0; // Default return value when processing a message

if (lPar <> 0) then begin
   // call the parent WMCommand method.
   WMCommand:=inherited WMCommand(wPar, lPar);
   exit;
   end;

// for the menus,  lPar = 0
//=========================
wNrMenu   := LoWord(wPar);
Case wNrMenu Of
   101 : GetSecWindowHandle;
   102 : GetSecWindowPosition;
   103 : SimulateLeftClick;
   104 : SimulateRightClick;
   107 : GetSquarePositions;
   108 : GetMineSweeperContent(1);
   109 : SolveMineSweeper;

   111 : begin GetMineSweeperContent(0);  PutFlags;  end;
   112 : begin GetMineSweeperContent(0);  SaftyExplore;  end;
   113 : begin GetMineSweeperContent(0);  SaftyExploreMore;  end;
   114 : begin GetMineSweeperContent(0);  ProbabilityExplore; end;

   115 : ClickOnNewGame;
   116 : GetStatistiquePlay;

   // Parameters
   301 : DialogExecMSG(30199, @Self);
   351 : ToggleSolveFast;

   // A help information
   901 : DialogExecMSG(90199, @Self);

   else WMCommand:=inherited WMCommand(wPar, lPar);
   end;  // Case NrMenu Of

SetActiveWindow(hWin);
SetFocus(hWin);
SetForegroundWindow(hWin);
SetActiveWindow(hWin);
SetFocus(hWin);
SetForegroundWindow(hWin);
end;  // TMyWindow.WMCommand

procedure TMyWindow.AskAcceptDlgQuit(poDlg:PWindowBase0; wCtrlIDP:Word32; var wRep:word32);
//=========================================================================================
// notification that the dialog box pointed by  PDialog(lPar)
// is about to quit, accepting its changes (With OK).
// If we refuse to quit, just Set  wRep:=0;
begin
wCtrlIDP:=wCtrlIDP;  wRep:=wRep;  // to avoid a hint
// notification that the dialog box pointed by  pDlg is about to quit.
// Its the best place to check if we accept to quit.
if (poDlg = PWindowBase0(pDialogActivate)) then begin
   // Closing the dialog Box, sets the pointer on it to NIL.
   pDialogActivate:=NIL;
   fGetWinSecHandle:=FALSE;
   end; // if (poDlg = pDialogActivate) then begin
end;  // TMyWindow.AskAcceptDlgQuit

function TMyWindow.WindowProc(hWindow:HWND; wMsg:UINT; wPar:WPARAM; lPar:LPARAM):LRESULT;
//======================================================================================
var
  pppT : Point;

begin
if (wMsg = WM_ACTIVATE) and (LoWord(wPar) = WA_INACTIVE) and
   fGetWinSecHandle and (pDialogActivate <> NIL) and
   (HWND(lPar) <> pDialogActivate^.hWin) then begin
   // Get the handle of the second window, usually the MineSweeper
//   hWinSec:=GetForegroundWindow;  doesn't work
   GetCursorPos(@pppT);
   hWinSec:=WindowFromPoint(pppT);
   fGetWinSecHandle:=FALSE;

   // Closes the dialog Box
   DestroyWindow(pDialogActivate^.hWin);
   pDialogActivate:=NIL;

   // Get the position and size of the Minesweeper, and
   // deduce from it the number of cells
   GetSecWindowPosition;
   end;

// Default treatment of a message. Does nearly nothing
WindowProc := inherited WindowProc(hWindow, wMsg, wPar, lPar);
end;  // TMyWindow.WindowProc

Begin // MAIN ******
// Name under which the main window is registered. Spaces are not allowed
oApp.Init('Example064');

// Creation of the main window
new(poMyMainWin, Init('tEx064. Solving the MineSweeper game.'));

// The App.WindowClass can be modified here, if wanted. (Not usual)
oApp.Run;
oApp.Done;  // The main window object is disposed here.
End.

