{ Example of entering a password in an edit box. And
  example of changing the controls creation order.
  Many other examples are given in  tEx038.pp (and tEx035.pp)

  The controls creation order can be changed, but it is not usual, and
  not very simple.
  The effect of it is that the TAB key will jump in the creation order
  among the controls.

  Use the  wDialogCreatorLOW2  instead of the  wDialogCreator2.
  It will hide many constant, types etc. which are not usefull
  to know for creating Dialog boxes.

  Remark : wDialog0.gglpoActiveDialog  points to the active dialog,
  Modal or modeless.

  See how simple it is to create a dialog box.
  --------------------------------------------}
program tEx068;

// To tell FreePascal that the application is a Graphical User Interface
// i.e. is running in windows, not in text mode.
{$APPTYPE GUI}

Uses
  wApplication0
 ,wWindow0
 ,wDialog0
 ,wControl0
 ,wPopupWindow0
 ,wDialogCreatorLOW2
 ,wDialogCreator2
 ,wVariableNumLink2
 ,wWindowBase0
 ,wGlobals2
 ,wDebug0
 ,Windows;

type
  PMyWindow = ^TMyWindow;
  TMyWindow = object(TWindow)
  //"""""""""""""""""""""""""
  // Used in 6th test, to do range checking
  wNum1       : Word32;
  fLimitIt    : boolean;
  strName     : string[60];
  afConvert   : array[1..3] of boolean;

  constructor Init(strWinTitlePrm:ShortString);
  destructor Done; virtual;

  procedure Test1; virtual;

  procedure Test2; virtual;

  procedure CtrlVarUpdate(poCtrl:PWindowBase0; pVar:pointer); virtual;
  // When a control changed its value and it is linked to a variable and
  // it belongs to a dialog belonging to a window, then
  // the control calls this method with a pointer on itself and
  // a pointer on the associated variable.
  // Either you know the type, or you can get it with
  // pControl(poCtrl)^.rVarLinkInfo.wVarType

  procedure AskAcceptDlgQuit(poDlg:PWindowBase0; wCtrlIDP:word32; var wRep:word32); virtual;
  // If a dialog box is about to quit with return, i.e.
  // accepting its changes, then this dialog parent's method is called
  // poDlg is a pointer on the dialog box.
  // If the value of wRep is set to  0
  // then it is refused to quit the dialog box.
  // For example, modifying  wRep:=0;
  // will reject the possibility to quit the dialog box.
  // Is destinated to be overriden, does nothing by default.

  function WMCommand(wPar:WPARAM; lPar:LPARAM):LRESULT; virtual;
  // Handles the menu messages

//  function WindowProc(hWindow:HWND; wMsg:UINT; wPar:WPARAM; lPar:LPARAM):LRESULT; virtual;
    // Not Used
  end; // TMyWindow = object(TWindow)

constructor TMyWindow.Init(strWinTitlePrm:ShortString);
//=====================================================
var nn : integer;
begin
inherited Init(strWinTitlePrm);

// default values
wNum1       := 0;
fLimitIt    :=FALSE;
strName     := '';
afConvert[1]:=TRUE;
afConvert[2]:=FALSE;
afConvert[3]:=FALSE;

// For the whole program, if these numbers are not used by other
// variables in the program, (the end developper must not make this mistake)
// then these links will remain.
// It will not be necessary to 'ClearField(..', since the linke number are not used again
Word32Field(201, wNum1);
BooleanField(211, fLimitIt);
StrField(301, strName, SizeOf(strName));
for nn:=1 to 3 do BooleanField(310+nn, afConvert[nn]);

// This line is all what is needed to have a status line.
new(poStatusLine, Init(@Self));
end;  // TMyWindow.Init

destructor TMyWindow.Done;
//========================
begin
inherited Done;
end;  // TMyWindow.Done

procedure TMyWindow.Test1;
//========================
begin
// Example of entering a password in an edit box
// and a little change of the control's creation order
DialogExecMSG(11199, @Self);
DBW0(1, 'Password ="' + strName + '"');
end;  // TMyWindow.Test1

procedure TMyWindow.Test2;
//========================
begin
// Example of entering a password in an edit box
// and a little change of the control's creation order
DialogExecMSG(11299, @Self);
DBW0(1, 'Password ="' + strName + '"');
end;  // TMyWindow.Test2

procedure TMyWindow.CtrlVarUpdate(poCtrl:PWindowBase0; pVar:pointer);
//===================================================================
// a variable has changed in a dialog box.
// pVar = pointer on the variable
// poCtrl = pointer on the control linked to this variable
// poCtrl^.poParent = pointer on the dialog box
// PDialog(poCtrl^.poParent)^.wMsgDescriptor =
// = msg which generated the dialog box
begin
pVar:=pVar;  // to avoid a hint.

if (PDialog(poCtrl^.poParent)^.wMsgDescriptor <= 11299) then begin
   // We don't do the range checking here, but only when the the dialog
   // is about to quit.
   // Here we do string conversion
   if (afConvert[2]) then begin
       strName:=UpCase(strName);
       UpdateFromVar;
       end
   else if (afConvert[3]) then begin
       strName:=LowerCase(strName);
       UpdateFromVar;
       end; // if
   end; // if

end;  // TMyWindow.CtrlVarUpdate

procedure TMyWindow.AskAcceptDlgQuit(poDlg:PWindowBase0; wCtrlIDP:Word32; var wRep:word32);
//=========================================================================================
// notification that the dialog box pointed by  PDialog(lPar)
// is about to quit, accepting its changes (With OK).
// If we refuse to quit, just Set  wRep:=0;
var
  poDlgL      : PDialog;
  poCtrl      : PControl;
  nXPosL, nYPosL,   // position and size of the control linked to wNum1
  nWidthL, nHeightL : Int32;
begin
wCtrlIDP:=wCtrlIDP;  // to avoid a hint
// notification that the dialog box pointed by  pDlg is about to quit.
// Its the best place to check if we accept to quit.
if (PDialog(poDlg)^.wMsgDescriptor <= 11299) then begin
   if (fLimitIt) and ((wNum1 < 1) or (wNum1 > 100)) then begin
      wRep:=0; // refuse
      // Then explain why we refused.
      // Show a popup window explaining why we refuse to quit.
      // if positions are (-1, -1) then the mouse cursor position
      // is taken to plac the popup window.
      poDlgL:=PDialog(poDlg);  // we know that it points on a dialog box
      poCtrl:=poDlgL^.GetControlFromVar(wNum1);
      poCtrl^.GetScreenPosSize(nXPosL, nYPosL, nWidthL, nHeightL);

      // Set the focus to the control which needs to be modified to accept to quit.
      // It must be set before we show the popup Window.
      SetFocus(poCtrl^.hWin);

      // If you want that all text is selected.
      // If not, remove this line.
      SendMessage(poCtrl^.hWin, EM_SetSel, 0, -1);

      // Opens the popup window with the text from message 12000 in it.
      PopupWindowMessage(11100, nXPosL, nYPosL + nHeightL + 5, @Self);
      end;
   end; // if (PDialog(poDlg)^.wMsgDescriptor = 11699) then begin
end;  // TMyWindow.AskAcceptDlgQuit

function TMyWindow.WMCommand(wPar:WPARAM; lPar:LPARAM):LRESULT;
//=============================================================
// Handles the menu messages
Var
  wNrMenu     : word;
begin
WMCommand := 0; // Default return value when processing a message

if (lPar <> 0) then begin
   // call the parent WMCommand method.
   WMCommand:=inherited WMCommand(wPar, lPar);
   exit;
   end; // if (lPar <> 0) then begin

// for the menus,  lPar = 0
//=========================
wNrMenu   := LoWord(wPar);
Case wNrMenu Of
   101 : Test1;
   102 : Test2;
   else WMCommand:=inherited WMCommand(wPar, lPar);
   end;  // Case NrMenu Of
end;  // TMyWindow.WMCommand

var
  oApp        : TApplication;
  poMyMainWin : PMyWindow;

Begin // MAIN ******
// Name under which the main window is registered. Spaces are not allowed
oApp.Init('Example068');

// Creation of the main window
new(poMyMainWin, Init('tEx068.  Example of entering a password.'));

// The App.WindowClass can be modified here, if wanted. (Not usual)
oApp.Run;
oApp.Done;  // The main window object is disposed here.
End.

