{ Example : reads voltage from the PCS10 "4 channel signal recorder" from Velleman
  Sets also pin out of the parallel port data, to command external electronic.
  The files FASTTIME32.DLL, K8047D.DLL and K8047E.EXE 
  must be located in the program file folder or in the Windows' SYSTEM32 folder.
  --------------------------------------------}
program tEx070;

// To tell FreePascal that the application is a Graphical User Interface
// i.e. is running in windows, not in text mode.
{$APPTYPE GUI}

Uses
  wApplication0
 ,wWindowText0
 ,wWindow0
 ,wDialog0
 ,wPopupWindow0
 ,wDialogCreatorLOW2
 ,wDialogCreator2
 ,wVariableNumLink2
 ,wWindowBase0
 ,wErrors0
 ,wGlobals2
 ,wDebug0
 ,Windows
 ,Strings
 ,zlportio  // to access the parallel port or any other port
 ,wReadTime // to read time intervals with accuracy.
 ;

const
SND_SYNC = $0000; // play synchronously (default)
SND_ASYNC = $0001; // play asynchronously
SND_NODEFAULT = $0002; // don't use default sound
SND_MEMORY = $0004; // lpszSoundName points to a memory file
SND_LOOP = $0008; // loop the sound until next sndPlaySound
SND_NOSTOP = $0010; // don't stop any currently playing sound
SND_NOWAIT = $00002000; // don't wait if the driver is busy
SND_ALIAS = $00010000; // name is a registry alias
SND_ALIAS_ID = $00110000; // alias is a predefined ID
SND_FILENAME = $00020000; // name is file name
SND_RESOURCE = $00040004; // name is resource name or atom
SND_PURGE = $0040; // purge non-static events for task
SND_APPLICATION = $0080; // look for application specific association

// declaration of the playsound function which is in : winmm.dll
function PlaySound(pszSound: PChar; hmod: HMODULE; fdwSound: DWORD): BOOL;
stdcall; external 'winmm.dll' name {$ifdef unicode}'PlaySoundW' {$else}
'PlaySoundA' {$endif};

const
  id_D0_OFF   = 100;
  id_D0_ON    = 101;
  id_D1_OFF   = 110;
  id_D1_ON    = 111;
  idQUIT      = 999;
  TIMER_ID    = 11;

var bPortState : byte;  // bits stats of the parallel port

  // gain of the channels : 1, 2, 5 or 10.
  glnGain1, glnGain2, glnGain3, glnGain4 : integer;
 
  // limit under which a sound is played to call the user
  glvVoltLimit : Float;

type
  PMyDialog = ^TMyDialog;
  TMyDialog = object(TDialog)
  //"""""""""""""""""""""""""

  vXValue : Float;
  // the value shown in the edit box, must be different if we don't
  // want rounding errors to occure.

  fClearNext : boolean;
  // The next digit clears the avValue[1] and replaces it by the digit

  fAddNext : boolean;
  // The next digit push the stack up and
  //  clears the avValue[1] and replaces it by the digit

  vPeriode : Float;
  // If the periode has been pressed, indicate how many digits are after the
  // period.

  vMemory : Float;
  // A simple memory.

  constructor Init(poParentW:PWindowBase);

  function WMCommand(wPar:WPARAM; lPar:LPARAM):BOOL; virtual;
  // Handles the menu messages

//  function DialogProc(hDialog:HWND; wMsg:UINT; wPar:WPARAM; lPar:LPARAM):LRESULT; virtual;
    // Not used
  // Handles the dialog messages. See comments in  wDialog0.pp
  end; // TMyDialog = object(TDialog)

type
  PMyWindow = ^TMyWindow;
  TMyWindow = object(TWindow)
  //"""""""""""""""""""""""""

  wTimer1ID : word32;
  // ID of the first (and only) timer
  // = 0 when stopped

  wNb_msStart : word32;
  // to set the starting time

  constructor Init(strWinTitlePrm:ShortString);

  destructor Done; virtual;

  procedure WinCreate(strWinClass:ShortString); virtual;
  // When creating the window, we ask to show the dialog box, not the window

  function WMCommand(wPar:WPARAM; lPar:LPARAM):LRESULT; virtual;
  // Handles the menu messages

  procedure WMTimer(wTimerID:WParam); virtual;
  // To handle the timer messages.

  function WindowProc(hWindow:HWND; wMsg:UINT; wPar:WPARAM; lPar:LPARAM):LRESULT; virtual;
  end; // TMyWindow = object(TWindow)

var
  oApp        : TApplication;
  poMyMainWin : PMyWindow;

PROCEDURE StartDevice; stdcall; external 'K8047d.dll';
PROCEDURE StopDevice; stdcall; external 'K8047d.dll';
PROCEDURE ReadData(Buffer:Pointer); stdcall; external 'K8047d.dll';
PROCEDURE SetGain(Channel: Longint ; Gain: Longint); stdcall; external 'K8047d.dll';
PROCEDURE LEDon; stdcall; external 'K8047d.dll';
PROCEDURE LEDoff; stdcall; external 'K8047d.dll';


function myCreateDialogObj(poParent:Pointer):Pointer;
//===================================================
var poObj : PMyDialog;
begin
new(poObj, Init(poParent));
myCreateDialogObj:=poObj;
end;  // myCreateDialogObj

//******************************************************************
// TMyDialog = object(TDialog)
//******************************************************************

constructor TMyDialog.Init(poParentW:PWindowBase);
//================================================
begin
// must be done, even if it seems useless.
inherited Init(poParentW);

vXValue:=0.0;
FloatField(100, vXValue);

fClearNext :=FALSE;
fAddNext   :=FALSE;
vPeriode   :=1; // no digits after the period.
vMemory    :=0;
end;  // TMyDialog.Init

function TMyDialog.WMCommand(wPar:WPARAM; lPar:LPARAM):BOOL;
//==========================================================
// Handles the menu messages
var
  wCtrlID     : word32;
  wSpeed,
  wCpt        : word32;
  dnTick1, dnTick2  // to read the time
              : int64;
  dwTimeNS  // elapsed time in nano seconds
              : dword;
  anDataBuffer: array[0..7] of Int32;
  nn          : integer;
  aanDataBuffers: array[1..100] of array[0..7] of Int32;

begin
if (HiWord(wPar) <> BN_CLICKED) then begin
   WMCommand:=inherited WMCommand(wPar, lPar);
   exit;
   end;
//DBW(1, 'WMCommand=' + DBNotifToStr(HiWord(wPar)) + ITS5(LoWord(wPar)));
wCtrlID:=LoWord(wPar);

//MessageBeep(0);

//DBW(1, 'Button clicked=' + ITS(poBtn^.wCtrlID));

// sets output D0 to 0 [V]
if (wCtrlID = id_D0_OFF) then begin
   DBW(0, 'set D0 to off');
   bPortState:=bPortState and ($FF - $01);
   portwriteb($378, bPortState);
   end;

// sets output D0 to more than 2 [V]
if (wCtrlID = id_D0_ON) then begin
   DBW(0, 'set D0 to on');
   bPortState:=bPortState or $01;
   portwriteb($378, bPortState);
   end;

// sets output D1 to 0 [V]
if (wCtrlID = id_D1_OFF) then begin
   DBW(0, 'set D1 to off');
   bPortState:=bPortState and ($FF - $02);
   portwriteb($378, bPortState);
   end;

// sets output D1 to more than 2 [V]
if (wCtrlID = id_D1_ON) then begin
   DBW(0, 'set D1 to on');
   bPortState:=bPortState or $02;
   portwriteb($378, bPortState);
   end;

// sets Led OFF on the PCS10
if (wCtrlID = 300) then begin
   DBW(0, 'set LED off');
   LEDOff;
end;

// sets Led ON on the PCS10
if (wCtrlID = 301) then begin
   DBW(0, 'set LED on');
   LEDOn;
   end;

// sets the gain to 1 on all channels on the PCS10
if (wCtrlID = 401) then begin
   SetGain(1, 1);   SetGain(2, 1);   SetGain(3, 1);   SetGain(4, 1);
   glnGain1:=1; glnGain2:=1; glnGain3:=1; glnGain4:=1; 
   end;

// sets the gain to 2 on all channels on the PCS10
if (wCtrlID = 402) then begin
   SetGain(1, 2);   SetGain(2, 2);   SetGain(3, 2);   SetGain(4, 2);
   glnGain1:=2; glnGain2:=2; glnGain3:=2; glnGain4:=2;
   end;

// sets the gain to 5 on all channels on the PCS10
if (wCtrlID = 405) then begin
   SetGain(1, 5);   SetGain(2, 5);   SetGain(3, 5);   SetGain(4, 5);
   glnGain1:=5; glnGain2:=5; glnGain3:=5; glnGain4:=5;
   end;

// sets the gain to 10 on all channels on the PCS10
if (wCtrlID = 410) then begin
   SetGain(1, 10);   SetGain(2, 10);   SetGain(3, 10);   SetGain(4, 10);
   glnGain1:=10; glnGain2:=10; glnGain3:=10; glnGain4:=10;
   end;

// Read voltage from the PCS10
if (wCtrlID = 700) then begin
   ReadData(@anDataBuffer);
   DBW(0, ITS(anDataBuffer[0]+256*anDataBuffer[1]) + ' ' + ITS(anDataBuffer[2]) + ' ' + ITS(anDataBuffer[3])
  + ' ' + ITS(anDataBuffer[4]) + ' ' + ITS(anDataBuffer[5]));
   end;

// Read voltage from the PCS10
if (wCtrlID = 710) then begin
   for nn:=1 to 30 do ReadData(@aanDataBuffers[nn]);  // reads the data

   for nn:=1 to 30 do begin
      DBW(0, ITS(aanDataBuffers[nn,0]+256*aanDataBuffers[nn,1]) + ' ' + ITS(aanDataBuffers[nn,2]) + ' ' + ITS(aanDataBuffers[nn,3])
     + ' ' + ITS(aanDataBuffers[nn,4]) + ' ' + ITS(aanDataBuffers[nn,5]));
      end;
   end;

// sets output D0 to more than 2 [V]
if (wCtrlID = 801) then begin
   wCpt:=0;
   wSpeed:=CPUGetSpeed;
   DBW(0, 'Speed = ' + WTS(wSpeed));
   dnTick1:=cpugettick;
   for wCpt:=1 to 100 do begin  end;
   dnTick2:=cpugettick;

   dwTimeNS:=cpugetNS(dnTick1); // get elapsed time in nano seconds

   DBW(0, 'Tick1, Tick2 & diff =' + WTS(dnTick1) + ' & ' + WTS(dnTick2) + WTS(dnTick2 - dnTick1));
   DBW(0, 'Delta time [ns] =' + WTS(dwTimeNS) + '  ' + FTS(_1CPUFrequencyNS));
   end;

// Did we click on the quit button ?
// Close the main window, thus quit the application.
if (wCtrlID = idQUIT) then  PostMessage(poMyMainWin^.hWin, WM_CLOSE, 0, 0);

WMCommand:=inherited WMCommand(wPar, lPar);
end;  // TMyDialog.WMCommand

//******************************************************************
// TMyWindow = object(TWindow)
//******************************************************************

constructor TMyWindow.Init(strWinTitlePrm:ShortString);
//=====================================================
begin
inherited Init(strWinTitlePrm);
wTimer1ID:=0;
bPortState:=0;
glnGain4:=1;
glnGain3:=glnGain4;  // used to avoid a wanring "assigned but never used".
glnGain2:=glnGain3;
glnGain1:=glnGain1;
glvVoltLimit:=1.1; // Volts
FloatField(1001, glvVoltLimit);

StartDevice;  // for the PCS10
end;  // TMyWindow.Init

destructor TMyWindow.Done;
//========================
begin
LEDOff;
StopDevice; // for the PCS10
inherited Done;
end;

procedure TMyWindow.WinCreate(strWinClass:ShortString);
//=====================================================
//var strName : string[50];
Begin
// We don't want to see the main window
//wSizeState := SIZE_MINIMIZED;

inherited WinCreate(strWinClass);

// Here calling :  DialogExecMSG(11199, @Self);  doesn't work,
// But posting a message to open the dialog box at startup is possible.
PostMessage(hWin, WM_COMMAND, 101, 0);

//strName:='zlportio' + char(0);
//DBW(0, 'result= ' + ITS(driverstart(@strName[1])));

end;  // TMyWindow.WinCreate

procedure TMyWindow.WMTimer(wTimerID:WParam);
//===========================================
// To handle the timer messages.
// a measurement is done here.
var
  wNb_ms      : word32;
//  poWrWin     : PWindowText;  // pointer on the window where the measurements are written.
  anDataBuffer: array[0..7] of Int32; // to read the time and the four voltages from the Velleman PCS10
  vVolt1      : Float;  // Voltage ridden
  vAmpere2    : Float;  // Current ridden (voltage divided by 10)

begin
if (wTimerID = TIMER_ID) then begin
   // time in mili-seconds since Windows started minus the time when
   // the start button has been pressed.
   wNb_ms:=(GetTickCount - wNb_msStart) div 1000;

   ReadData(@anDataBuffer);

   vVolt1   := anDataBuffer[2] * 30 / glnGain1 / 256;  // gain set to  30/glnGain1  Volts for an output of 256.
   vAmpere2 := anDataBuffer[3] * 30 / glnGain2 / 2560; // gain set to   3/glnGain2  amperes for an output of 256.
   WriteWin(1, IntToStr(wNb_ms, 6) // + IntToStr(anDataBuffer[0]+256*anDataBuffer[1], 7)  // time from the PCS10
             + FloatToStr(vVolt1,6,2) + FloatToStr(vAmpere2,7,3)
             + FloatToStr(vVolt1*vAmpere2,7,3)
             );  // last field = power [W]

   // save the measurement file every ... minutes.
//   if ((wNb_ms mod 1200) = 0) then begin
//      poWrWin:=GetWriteWinNr(1);
//      poWrWin^.oRichEditFrame.SaveRICH;
//      end;

   // if the Voltage is below a limite, a sound message is played to alert the user.
   if (vVolt1 < glvVoltLimit) then begin
      PlaySound('Warning_Voltage_Bellow_Limit.wav', 0, SND_ASYNC or SND_FILENAME);
      end;

   end; // if (wTimerID = TIMER_ID)
end;  // TMyWindow.WMTimer

function TMyWindow.WMCommand(wPar:WPARAM; lPar:LPARAM):LRESULT;
//=============================================================
// treats the menu commands
var
  poWrWin     : PWindowText;  // pointer on the window where the measurements are written.
  TheLogFont            : TLogFont;
  TheFont               : HFont;

begin
// Execute the dialog box at starting
// And the help window
if (lPar = 0) and (wPar =101) then DialogCreateMSG(11199, @Self);
//if (lPar = 0) and (wPar =121) then DialogCreateMSG(21199, @Self);

// Accelerators act as button click.
// Send the command to the active dialog
case LoWord(wPar) of
  idQUIT : if (gglpoActiveDialog <> NIL) then
               gglpoActiveDialog^.WMCommand(LoWord(wPar), 0);

  111    : if (wTimer1ID = 0) then begin
              // starts the measurements

              // if the measurement output window is already opened, use it and don't ask to open another.
              poWrWin:=GetWriteWinNr(1);

              if (poWrWin = NIL) then begin
                 // Sets the Gain.  5 times for each channels, such that  256 correspond to 6 [V].
                 // SetGain(1, 5);   SetGain(2, 5);   SetGain(3, 5);   SetGain(4, 5);
                 // glnGain1:=5; glnGain2:=5; glnGain3:=5; glnGain4:=5;

                 // first write a first line and ask where to save the file
                 WriteWin(1, ' ');
                 poWrWin:=GetWriteWinNr(1);

                 // sets the font to 'Courier New"
                 with TheLogFont do begin
                    lfHeight         := 18;               // Default logical height of font
                    lfWidth          := 0;                // Default logical average character width
                    lfEscapement     := 0;                // angle of escapement
                    lfOrientation    := 0;                // base-line orientation angle
                    lfWeight         := FW_NORMAL;        // font weight
                    lfItalic         := 0;                // italic attribute flag
                    lfUnderline      := 0;                // underline attribute flag
                    lfStrikeOut      := 0;                // strikeout attribute flag
                    lfCharSet        := DEFAULT_CHARSET;  // character set identifier
                    lfOutPrecision   := OUT_DEFAULT_PRECIS;  // output precision
                    lfClipPrecision  := CLIP_DEFAULT_PRECIS; // clipping precision
                    lfQuality        := DEFAULT_QUALITY;     // output quality
                    lfPitchAndFamily := DEFAULT_PITCH;    // pitch and family
                    Strcopy(lfFaceName,'Courier New');    // pointer to typeface name string
                    end; // with
                    //      TheColor := GetSysColor(COLOR_WINDOWTEXT);
                    TheFont  := CreateFontIndirect(@TheLogFont);
                    SendMessage(poWrWin^.oRichEditFrame.hWin,WM_SETFONT,TheFont,1);

                 // sets the default name the first time, and ask where to save
                 poWrWin^.oRichEditFrame.strFilePath:='./z3Mesasure_.txt';  // name of the data file
                 poWrWin^.oRichEditFrame.SaveRICHas;
                 end;

              // starts the timer, for the measurements
              wTimer1ID:=TIMER_ID;
              SetTimer( hWin, TIMER_ID, // TIMER_ID = a unique timer ID
                        30000,  // timeout in mili-seconds
                        NIL);

              // time when timer started to count
              wNb_msStart:=GetTickCount;

              WMTimer(TIMER_ID);  // start a measurement imdiately
              end;

  112    : if (wTimer1ID <> 0) then begin
              // stops the measurements
              KillTimer(hWin, TIMER_ID);
              wTimer1ID:=0;
              end;

  end; // case

WMCommand:=inherited WMCommand(wPar, lPar);
end;  // TMyWindow.WMCommand

function TMyWindow.WindowProc(hWindow:HWND; wMsg:UINT; wPar:WPARAM; lPar:LPARAM):LRESULT;
//=======================================================================================
// All window messages are send here.  The timer command also
var
  poWrWin     : PWindowText;  // pointer on the window where the measurements are written.

Begin
WindowProc := 0; // Default return value when processing a message

{ if (wMsg <> WM_MouseMove)   and (wMsg <> WM_NCMouseMove) and
   (wMsg <> WM_MENUSELECT )   and (wMsg <> WM_ENTERIDLE ) and
   (wMsg <> WM_SetCursor)   and (wMsg <> WM_NCHittest) and
   (wMsg <> WM_SetFocus) and (wMsg <> WM_KillFocus) and
   (wMsg <> WM_KEYDOWN) and (wMsg <> WM_KEYUP) and
   (wMsg <> WM_Move)        and (wMsg <> WM_NCActivate) then
   DBW0(32, DBMsgToStr(wMsg) + ITS(wPar) + ITS(lPar)); }

if (wMsg = WM_Destroy) then begin
   if (wTimer1ID <> 0) then begin
      // stops the measurements
      KillTimer(hWin, TIMER_ID);
      wTimer1ID:=0;
      end;
 
   // saves the text file where the measurements have been done
   poWrWin:=GetWriteWinNr(1);
   if (poWrWin <> NIL) then poWrWin^.oRichEditFrame.SaveRICH;
   end;

Case wMsg Of
//   WM_Command : WindowProc:=WMCommand(wPar, lPar);  // done by default
   WM_Timer   : WMTimer(wPar);

else begin // Case wMsg Of
      // Default processing of the messages.
      WindowProc:=inherited WindowProc(hWindow, wMsg, wPar, lPar);
      end; // Case wMsg Of
   End; // Case wMsg Of

End;  // TMyWindow.WindowProc

//**********************************************************************
Begin // MAIN ******
// Assign the pointer on the function which will create the dialog.
// It is used by the library. So the library will construct
// a user defined object.
gglfnCreateDialogObj:=@myCreateDialogObj;

// Name under which the main window is registered. Spaces are not allowed
oApp.Init('Example070');



// Creation of the main window
new(poMyMainWin, Init('tEx070_Voltmeter_PCS10_Velleman.pp'));

// The App.WindowClass can be modified here, if wanted. (Not usual)
oApp.Run;
oApp.Done;  // The main window object is disposed here.
End.

