{ Test some sound possibilities

Information to play a self defined sound.
The PlaySound function plays a sound specified by the given filename, resource, or system event. 
(A system event may be associated with a sound in the registry or in the WIN.INI file.)

BOOL PlaySound(
    LPCSTR pszSound,
    HMODULE hmod,	
    DWORD fdwSound	
   );	
Parameters
==========
pszSound
A string that specifies the sound to play. If this parameter is NULL, any currently playing waveform sound is stopped. To stop a non-waveform sound, specify SND_PURGE in the fdwSound parameter.
Three flags in fdwSound (SND_ALIAS, SND_FILENAME, and SND_RESOURCE) determine whether the name is interpreted as an alias for a system event, a filename, or a resource identifier. If none of these flags are specified, PlaySound searches the registry or the WIN.INI file for an association with the specified sound name. If an association is found, the sound event is played. If no association is found in the registry, the name is interpreted as a filename.

hmod
Handle of the executable file that contains the resource to be loaded. This parameter must be NULL unless SND_RESOURCE is specified in fdwSound.

fdwSound
Flags for playing the sound. The following values are defined:

SND_APPLICATION
The sound is played using an application-specific association.

SND_ALIAS
The pszSound parameter is a system-event alias in the registry or the WIN.INI file. Do not use with either SND_FILENAME or SND_RESOURCE.

SND_ALIAS_ID
The pszSound parameter is a predefined sound identifier.

SND_ASYNC
The sound is played asynchronously and PlaySound returns immediately after beginning the sound. To terminate an asynchronously played waveform sound, call PlaySound with pszSound set to NULL.

SND_FILENAME
The pszSound parameter is a filename.

SND_LOOP
The sound plays repeatedly until PlaySound is called again with the pszSound parameter set to NULL. You must also specify the SND_ASYNC flag to indicate an asynchronous sound event.

SND_MEMORY
A sound events file is loaded in RAM. The parameter specified by pszSound must point to an image of a sound in memory.

SND_NODEFAULT
No default sound event is used. If the sound cannot be found, PlaySound returns silently without playing the default sound.

SND_NOSTOP
The specified sound event will yield to another sound event that is already playing. If a sound cannot be played because the resource needed to generate that sound is busy playing another sound, the function immediately returns FALSE without playing the requested sound.
If this flag is not specified, PlaySound attempts to stop the currently playing sound so that the device can be used to play the new sound.

SND_NOWAIT
If the driver is busy, return immediately without playing the sound.

SND_PURGE
Sounds are to be stopped for the calling task. If pszSound is not NULL, all instances of the specified sound are stopped. If pszSound is NULL, all sounds that are playing on behalf of the calling task are stopped.
You must also specify the instance handle to stop SND_RESOURCE events.

SND_RESOURCE
The pszSound parameter is a resource identifier; hmod must identify the instance that contains the resource.

SND_SYNC
Synchronous playback of a sound event. PlaySound returns after the sound event completes.

Return Values
Returns TRUE if successful or FALSE otherwise.

Remarks
The sound specified by pszSound must fit into available physical memory and 
be playable by an installed waveform-audio device driver. PlaySound searches 
the following directories for sound files: the current directory; 
the Windows directory; the Windows system directory; directories listed in
the PATH environment variable; and the list of directories mapped in a network.
For more information about the directory search order, see the documentation for the OpenFile function.

If it cannot find the specified sound, PlaySound uses the default system event sound entry instead. 
If the function can find neither the system default entry nor the default sound, 
it makes no sound and returns FALSE.
  --------------------------------------------}
program tEx072_Sound_Tests;

// To tell FreePascal that the application is a Graphical User Interface
// i.e. is running in windows, not in text mode.
{$APPTYPE GUI}

Uses
  wApplication0
 ,wWindowText0
 ,wWindow0
 ,wDialog0
 ,wPopupWindow0
 ,wDialogCreatorLOW2
 ,wDialogCreator2
 ,wVariableNumLink2
 ,wWindowBase0
 ,wErrors0
 ,wGlobals2
 ,wDebug0
 ,Windows
 ;
// cette partie est une rcupration des constantes de playsound,
// j'ai laiss les commentaires originaux qui peuvent etre utiles
const
SND_SYNC = $0000; // play synchronously (default)
SND_ASYNC = $0001; // play asynchronously
SND_NODEFAULT = $0002; // don't use default sound
SND_MEMORY = $0004; // lpszSoundName points to a memory file
SND_LOOP = $0008; // loop the sound until next sndPlaySound
SND_NOSTOP = $0010; // don't stop any currently playing sound
SND_NOWAIT = $00002000; // don't wait if the driver is busy
SND_ALIAS = $00010000; // name is a registry alias
SND_ALIAS_ID = $00110000; // alias is a predefined ID
SND_FILENAME = $00020000; // name is file name
SND_RESOURCE = $00040004; // name is resource name or atom
SND_PURGE = $0040; // purge non-static events for task
SND_APPLICATION = $0080; // look for application specific association

// declaration de la fonction playsound contenue dans la dll winmm.dll
function PlaySound(pszSound: PChar; hmod: HMODULE; fdwSound: DWORD): BOOL;
stdcall; external 'winmm.dll' name {$ifdef unicode}'PlaySoundW' {$else}
'PlaySoundA' {$endif};

type
  TWavSound = array[0..30000000] of byte;  // to store the sound
  PWavSound = ^TWavSound;
var
  paS     : PWavSound; // where the sound will be stored

type
  PMyDialog = ^TMyDialog;
  TMyDialog = object(TDialog)
  //"""""""""""""""""""""""""
  constructor Init(poParentW:PWindowBase);

  function WMCommand(wPar:WPARAM; lPar:LPARAM):BOOL; virtual;
  // Handles the menu messages

//  function DialogProc(hDialog:HWND; wMsg:UINT; wPar:WPARAM; lPar:LPARAM):LRESULT; virtual;
    // Not used
  // Handles the dialog messages. See comments in  wDialog0.pp
  end; // TMyDialog = object(TDialog)

type
  PMyWindow = ^TMyWindow;
  TMyWindow = object(TWindow)
  //"""""""""""""""""""""""""
  constructor Init(strWinTitlePrm:ShortString);

  destructor Done; virtual;

  procedure WinCreate(strWinClass:ShortString); virtual;
  // When creating the window, we ask to show the dialog box, not the window

  function WMCommand(wPar:WPARAM; lPar:LPARAM):LRESULT; virtual;
  // Handles the menu messages

  function WindowProc(hWindow:HWND; wMsg:UINT; wPar:WPARAM; lPar:LPARAM):LRESULT; virtual;
  end; // TMyWindow = object(TWindow)

var
  oApp        : TApplication;
  poMyMainWin : PMyWindow;

function myCreateDialogObj(poParent:Pointer):Pointer;
//===================================================
var poObj : PMyDialog;
begin
new(poObj, Init(poParent));
myCreateDialogObj:=poObj;
end;  // myCreateDialogObj

//******************************************************************
// TMyDialog = object(TDialog)
//******************************************************************

constructor TMyDialog.Init(poParentW:PWindowBase);
//================================================
begin
// must be done, even if it seems useless.
inherited Init(poParentW);
end;  // TMyDialog.Init

procedure Test1;
//==============
(* creates a .wav sound file at 22050 Hz sampling in memory, in the variable paS^[...]
 Play next this sound.
 in Hex :  for a 8 bit, mono, 22'050 Hz
  52 49 46 46 bb aa dd cc     aabbccdd = length of the data in bytes = file length in bytes - 8 bytes
  57 41 56 45 66 6D 74 20
  10 00 00 00 01 00 01 00
  22 56 00 00 22 56 00 00   // 00 00 56 22  = 22'050 in hexadecimal = $5622
  01 00 08 00 64 61 74 61  ....data
  bb aa dd cc                 aabbccdd = length of the data in bytes = file length in bytes - $90 bytes

  DATA BEGINS...  list of bytes...

  4c 49 53 54 5c 00 00 00  LIST\...
  49 4e 46 4f 49 43 52 44  INFOICRD
  0c 00 00 00 32 30 30 38  ....2008
  2d 30 34 2d 31 34 00 00  -04-14..
  49 45 4e 47 0f 00 00 00  IENG....
  42 65 72 6e 61 72 64 20  Bernard
  47 69 73 69 6e 00 00 01  Gisin...
  49 53 46 54 23 00 00 00  ISFT#...
  53 6f 6e 79 20 53 6f 75  Sony Sou
  6e 64 20 46 6f 72 67 65  nd Forge
  20 41 75 64 69 6f 20 53   Audio S
  74 75 64 69 6f 20 38 2e  tudio 8.
  30 00 00 02              0...
*)

const
  aHead : array[0..43] of byte = ($52, $49, $46, $46  // a 8 bit, mono, ???
          ,   1,   2,   3,   4  // file length in bytes - 8 bytes,  will be rewritten
          , $57, $41, $56, $45, $66, $6D, $74, $20
          , $10, $00, $00, $00, $01, $00, $01, $00
          , $22, $56, $00, $00, $22, $56, $00, $00  // $22 $56 $00 $00  = 22'050 in hexadecimal = $5622
          , $01, $00, $08, $00, $64, $61, $74, $61
          ,   1,   2,   3,   4  // file length in bytes - $90 bytes,  will be rewritten  ($90 = 144 = 100 + 44)
          );
  aFoot : array[0..99] of byte = (
            $4c, $49, $53, $54, $5c, $00, $00, $00
          , $49, $4e, $46, $4f, $49, $43, $52, $44
          , $0c, $00, $00, $00, $32, $30, $30, $38
          , $2d, $30, $34, $2d, $31, $34, $00, $00
          , $49, $45, $4e, $47, $0f, $00, $00, $00
          , $42, $65, $72, $6e, $61, $72, $64, $20
          , $47, $69, $73, $69, $6e, $00, $00, $01
          , $49, $53, $46, $54, $23, $00, $00, $00
          , $53, $6f, $6e, $79, $20, $53, $6f, $75
          , $6e, $64, $20, $46, $6f, $72, $67, $65
          , $20, $41, $75, $64, $69, $6f, $20, $53
          , $74, $75, $64, $69, $6f, $20, $38, $2e
          , $30, $00, $00, $02
          );

var
  nn      : word32;    // position in PWavSound
  nLenData : int32;  // number of bytes coding the sound, without the heading and footing.
  nLen : int32;
  b1, b2, b3, b4 : byte;
  n1, n1Max : int32;
  vT0, vT1  : Float;

begin
// Creates a copy of the wav sound in the paSound variable

// Header
for nn:=0 to 43 do paS^[nn]:=aHead[nn];
nn:=43;

// Sound code
vT1:=0;
vT1:=10;
repeat
  vT0:=vT1;
  vT1:=vT0 + exp(-vT0 * ln(2)/4);

  n1Max:=Round(22050 * (vT1 - vT0));
  for n1:=1 to n1Max do begin inc(nn); paS^[nn]:=$10; end;
  for n1:=1 to n1Max do begin inc(nn); paS^[nn]:=$f0; end;
  until (vT1 - vT0 < 0.02);

// number of data
nLenData:=nn;

// Footer
for n1:=0 to 99 do begin inc(nn); paS^[nn]:=aFoot[n1]; end;

// writing the "file" length
nLen := nLenData + 144 - 8;
b1:= nLen         mod $100;
b2:=(nLen div $100) mod $100;
b3:=(nLen div $10000) mod $100;
b4:=(nLen div $1000000) mod $100;
paS^[4]:=b1;  paS^[5]:=b2;  paS^[6]:=b3;  paS^[7]:=b4;

// writing the data length
nLen := nLenData;
b1:= nLen         mod $100;
b2:=(nLen div $100) mod $100;
b3:=(nLen div $10000) mod $100;
b4:=(nLen div $1000000) mod $100;
paS^[40]:=b1;  paS^[41]:=b2;  paS^[42]:=b3;  paS^[43]:=b4;

// the sound is in memory, plays it.
PlaySound(LPCSTR(paS), 0, SND_MEMORY or SND_ASYNC);  // SND_ASYNC return immediatly   or  SND_NODEFAULT
end;  // Test1;

procedure Test2;
//==============
(* creates a .wav sound file at 22050 Hz sampling in memory, in the variable paS^[...]
 Play next this sound.
 in Hex :  for a 8 bit, mono, 22'050 Hz
  52 49 46 46 bb aa dd cc     aabbccdd = length of the data in bytes = file length in bytes - 8 bytes
  57 41 56 45 66 6D 74 20
  10 00 00 00 01 00 01 00
  22 56 00 00 22 56 00 00   // 00 00 56 22  = 22'050 in hexadecimal = $5622
  01 00 08 00 64 61 74 61  ....data
  bb aa dd cc                 aabbccdd = length of the data in bytes = file length in bytes - $90 bytes

  DATA BEGINS...  list of bytes...

  4c 49 53 54 5c 00 00 00  LIST\...
  49 4e 46 4f 49 43 52 44  INFOICRD
  0c 00 00 00 32 30 30 38  ....2008
  2d 30 34 2d 31 34 00 00  -04-14..
  49 45 4e 47 0f 00 00 00  IENG....
  42 65 72 6e 61 72 64 20  Bernard
  47 69 73 69 6e 00 00 01  Gisin...
  49 53 46 54 23 00 00 00  ISFT#...
  53 6f 6e 79 20 53 6f 75  Sony Sou
  6e 64 20 46 6f 72 67 65  nd Forge
  20 41 75 64 69 6f 20 53   Audio S
  74 75 64 69 6f 20 38 2e  tudio 8.
  30 00 00 02              0...
*)

const
  aHead : array[0..43] of byte = ($52, $49, $46, $46  // a 8 bit, mono, ???
          ,   1,   2,   3,   4  // file length in bytes - 8 bytes,  will be rewritten
          , $57, $41, $56, $45, $66, $6D, $74, $20
          , $10, $00, $00, $00, $01, $00, $01, $00
          , $22, $56, $00, $00, $22, $56, $00, $00  // $22 $56 $00 $00  = 22'050 in hexadecimal = $5622
          , $01, $00, $08, $00, $64, $61, $74, $61
          ,   1,   2,   3,   4  // file length in bytes - $90 bytes,  will be rewritten  ($90 = 144 = 100 + 44)
          );
  aFoot : array[0..99] of byte = (
            $4c, $49, $53, $54, $5c, $00, $00, $00
          , $49, $4e, $46, $4f, $49, $43, $52, $44
          , $0c, $00, $00, $00, $32, $30, $30, $38
          , $2d, $30, $34, $2d, $31, $34, $00, $00
          , $49, $45, $4e, $47, $0f, $00, $00, $00
          , $42, $65, $72, $6e, $61, $72, $64, $20
          , $47, $69, $73, $69, $6e, $00, $00, $01
          , $49, $53, $46, $54, $23, $00, $00, $00
          , $53, $6f, $6e, $79, $20, $53, $6f, $75
          , $6e, $64, $20, $46, $6f, $72, $67, $65
          , $20, $41, $75, $64, $69, $6f, $20, $53
          , $74, $75, $64, $69, $6f, $20, $38, $2e
          , $30, $00, $00, $02
          );

var
  nn      : word32;    // position in PWavSound
  nLenData : int32;  // number of bytes coding the sound, without the heading and footing.
  nLen : int32;
  b1, b2, b3, b4 : byte;
  vT0  : Float;  // time
  vFreq : Float; // frequency

begin
// Creates a copy of the wav sound in the paSound variable

// Header
for nn:=0 to 43 do paS^[nn]:=aHead[nn];
nn:=43;

// Sound code
vT0:=0;      // temps en secondes
vFreq:=440;  // frequency in Hz
repeat
  inc(nn); paS^[nn]:=Round(128+120*sin(2*3.1415928*vFreq*vT0));  // 440 = the sound frequency
  vT0:=vT0 + 1/22050;  // 22050 = the sampling frequency
  vFreq:=440 + 14000*vT0/(30*2);  // change slowly the frequency with the time.
  // if signal = sin(2*pi*w(t)), frequency = derivative w'(t).  Here,  w(t) = 220*t +9000*t^2/(30*2),  w'(t)=220 +9000*t/30
  until (vT0 > 30);  // sets the playing time in seconds

// number of data
nLenData:=nn;

// for b1:=1 to 200 do DBW(1, ITS(word(paS^[5000-1*integer(b1)])));

// Footer
for b1:=0 to 99 do begin inc(nn); paS^[nn]:=aFoot[b1]; end;

// writing the "file" length
nLen := nLenData + 144 - 8;
b1:= nLen         mod $100;
b2:=(nLen div $100) mod $100;
b3:=(nLen div $10000) mod $100;
b4:=(nLen div $1000000) mod $100;
paS^[4]:=b1;  paS^[5]:=b2;  paS^[6]:=b3;  paS^[7]:=b4;

// writing the data length
nLen := nLenData;
b1:= nLen         mod $100;
b2:=(nLen div $100) mod $100;
b3:=(nLen div $10000) mod $100;
b4:=(nLen div $1000000) mod $100;
paS^[40]:=b1;  paS^[41]:=b2;  paS^[42]:=b3;  paS^[43]:=b4;

// the sound is in memory, plays it.
PlaySound(LPCSTR(paS), 0, SND_MEMORY or SND_ASYNC);  // SND_ASYNC return immediatly   or  SND_NODEFAULT
end;  // Test2;

procedure Test100kHz;
//===================
(* creates a .wav sound file at 100000 Hz sampling in memory, in the variable paS^[...]
 Play next this sound.
*)

const
  aHead : array[0..43] of byte = ($52, $49, $46, $46  // a 8 bit, mono, ???
          ,   1,   2,   3,   4  // file length in bytes - 8 bytes,  will be rewritten
          , $57, $41, $56, $45, $66, $6D, $74, $20
          , $10, $00, $00, $00, $01, $00, $01, $00
          , $A0, $86, $01, $00, $A0, $86, $01, $00  // $A0 $86 $01 $00  = 100'000 in hexadecimal = $186A0
          , $01, $00, $08, $00, $64, $61, $74, $61
          ,   1,   2,   3,   4  // file length in bytes - $90 bytes,  will be rewritten  ($90 = 144 = 100 + 44)
          );

  aFoot : array[0..99] of byte = (
            $4c, $49, $53, $54, $5c, $00, $00, $00
          , $49, $4e, $46, $4f, $49, $43, $52, $44
          , $0c, $00, $00, $00, $32, $30, $30, $38
          , $2d, $30, $34, $2d, $31, $34, $00, $00
          , $49, $45, $4e, $47, $0f, $00, $00, $00
          , $42, $65, $72, $6e, $61, $72, $64, $20
          , $47, $69, $73, $69, $6e, $00, $00, $01
          , $49, $53, $46, $54, $23, $00, $00, $00
          , $53, $6f, $6e, $79, $20, $53, $6f, $75
          , $6e, $64, $20, $46, $6f, $72, $67, $65
          , $20, $41, $75, $64, $69, $6f, $20, $53
          , $74, $75, $64, $69, $6f, $20, $38, $2e
          , $30, $00, $00, $02
          );

var
  nn      : word32;    // position in PWavSound
  nLenData : int32;  // number of bytes coding the sound, without the heading and footing.
  nLen : int32;
  b1, b2, b3, b4 : byte;
  vT0  : Float;  // time
  vFreq : Float; // frequency

begin
// Creates a copy of the wav sound in the paSound variable

// Header
for nn:=0 to 43 do paS^[nn]:=aHead[nn];
nn:=43;

// Sound code
vT0:=0;      // temps en secondes
vFreq:=440;  // frequency in Hz
repeat
  inc(nn); paS^[nn]:=Round(128+120*sin(2*3.1415928*vFreq*vT0));  // 440 = the sound frequency
  vT0:=vT0 + 1/100000;  //  / "the sampling frequency"
  vFreq:=440 + 14000*vT0/(30*2);  // change slowly the frequency with the time.
  // if signal = sin(2*pi*w(t)), frequency = derivative w'(t).  Here,  w(t) = 220*t +9000*t^2/(30*2),  w'(t)=220 +9000*t/30
  until (vT0 > 30);  // sets the playing time in seconds

// number of data
nLenData:=nn;

// for b1:=1 to 200 do DBW(1, ITS(word(paS^[5000-1*integer(b1)])));

// Footer
for b1:=0 to 99 do begin inc(nn); paS^[nn]:=aFoot[b1]; end;

// writing the "file" length
nLen := nLenData + 144 - 8;
b1:= nLen         mod $100;
b2:=(nLen div $100) mod $100;
b3:=(nLen div $10000) mod $100;
b4:=(nLen div $1000000) mod $100;
paS^[4]:=b1;  paS^[5]:=b2;  paS^[6]:=b3;  paS^[7]:=b4;

// writing the data length
nLen := nLenData;
b1:= nLen         mod $100;
b2:=(nLen div $100) mod $100;
b3:=(nLen div $10000) mod $100;
b4:=(nLen div $1000000) mod $100;
paS^[40]:=b1;  paS^[41]:=b2;  paS^[42]:=b3;  paS^[43]:=b4;

// the sound is in memory, plays it.
PlaySound(LPCSTR(paS), 0, SND_MEMORY or SND_ASYNC);  // SND_ASYNC return immediatly   or  SND_NODEFAULT
end;  // Test100kHz;

procedure Test200kHz;
//===================
(* creates a .wav sound file at 100000 Hz sampling in memory, in the variable paS^[...]
 Play next this sound.
 It it useless to use a 200 kHz sampling rate, since usual maximum sampling rates of soundcards are limited to 48 kHz.

*)

const
  aHead : array[0..43] of byte = ($52, $49, $46, $46  // a 8 bit, mono, ???
          ,   1,   2,   3,   4  // file length in bytes - 8 bytes,  will be rewritten
          , $57, $41, $56, $45, $66, $6D, $74, $20
          , $10, $00, $00, $00, $01, $00, $01, $00
          , $40, $0D, $03, $00, $40, $0D, $03, $00  // $40 $0D $03 $00  = 100'000 in hexadecimal = $30D40
          , $01, $00, $08, $00, $64, $61, $74, $61
          ,   1,   2,   3,   4  // file length in bytes - $90 bytes,  will be rewritten  ($90 = 144 = 100 + 44)
          );

  aFoot : array[0..99] of byte = (
            $4c, $49, $53, $54, $5c, $00, $00, $00
          , $49, $4e, $46, $4f, $49, $43, $52, $44
          , $0c, $00, $00, $00, $32, $30, $30, $38
          , $2d, $30, $34, $2d, $31, $34, $00, $00
          , $49, $45, $4e, $47, $0f, $00, $00, $00
          , $42, $65, $72, $6e, $61, $72, $64, $20
          , $47, $69, $73, $69, $6e, $00, $00, $01
          , $49, $53, $46, $54, $23, $00, $00, $00
          , $53, $6f, $6e, $79, $20, $53, $6f, $75
          , $6e, $64, $20, $46, $6f, $72, $67, $65
          , $20, $41, $75, $64, $69, $6f, $20, $53
          , $74, $75, $64, $69, $6f, $20, $38, $2e
          , $30, $00, $00, $02
          );

var
  nn      : word32;    // position in PWavSound
  nLenData : int32;  // number of bytes coding the sound, without the heading and footing.
  nLen : int32;
  b1, b2, b3, b4 : byte;
  vT0  : Float;  // time
  vFreq : Float; // frequency

begin
// Creates a copy of the wav sound in the paSound variable

// Header
for nn:=0 to 43 do paS^[nn]:=aHead[nn];
nn:=43;

// Sound code
vT0:=0;      // temps en secondes
vFreq:=440;  // frequency in Hz
repeat
  inc(nn); paS^[nn]:=Round(128+120*sin(2*3.1415928*vFreq*vT0));  // 440 = the sound frequency
  vT0:=vT0 + 1/200000;  //  / "the sampling frequency"
//  vFreq:=220 + 14000*vT0/(30*2);  // change slowly the frequency with the time.
  // if signal = sin(2*pi*w(t)), frequency = derivative w'(t).  Here,  w(t) = 220*t +9000*t^2/(30*2),  w'(t)=220 +9000*t/30
  until (vT0 > 30);  // sets the playing time in seconds

// number of data
nLenData:=nn;

// for b1:=1 to 200 do DBW(1, ITS(word(paS^[5000-1*integer(b1)])));

// Footer
for b1:=0 to 99 do begin inc(nn); paS^[nn]:=aFoot[b1]; end;

// writing the "file" length
nLen := nLenData + 144 - 8;
b1:= nLen         mod $100;
b2:=(nLen div $100) mod $100;
b3:=(nLen div $10000) mod $100;
b4:=(nLen div $1000000) mod $100;
paS^[4]:=b1;  paS^[5]:=b2;  paS^[6]:=b3;  paS^[7]:=b4;

// writing the data length
nLen := nLenData;
b1:= nLen         mod $100;
b2:=(nLen div $100) mod $100;
b3:=(nLen div $10000) mod $100;
b4:=(nLen div $1000000) mod $100;
paS^[40]:=b1;  paS^[41]:=b2;  paS^[42]:=b3;  paS^[43]:=b4;

// the sound is in memory, plays it.
PlaySound(LPCSTR(paS), 0, SND_MEMORY or SND_ASYNC);  // SND_ASYNC return immediatly   or  SND_NODEFAULT
end;  // Test200kHz;

procedure Test48kHz;
//==================
(* creates a .wav sound file at 100000 Hz sampling in memory, in the variable paS^[...]
 Play next this sound.
 Uses the maximum sampling rate of my sound card which is 48 kHz
*)

const
  aHead : array[0..43] of byte = ($52, $49, $46, $46  // a 8 bit, mono, ???
          ,   1,   2,   3,   4  // file length in bytes - 8 bytes,  will be rewritten
          , $57, $41, $56, $45, $66, $6D, $74, $20
          , $10, $00, $00, $00, $01, $00, $01, $00
          , $80, $bb, $00, $00, $80, $bb, $00, $00  // $80 $bb $00 $00  = 48'000 in hexadecimal = $bb80
          , $01, $00, $08, $00, $64, $61, $74, $61
          ,   1,   2,   3,   4  // file length in bytes - $90 bytes,  will be rewritten  ($90 = 144 = 100 + 44)
          );

  aFoot : array[0..99] of byte = (
            $4c, $49, $53, $54, $5c, $00, $00, $00
          , $49, $4e, $46, $4f, $49, $43, $52, $44
          , $0c, $00, $00, $00, $32, $30, $30, $38
          , $2d, $30, $34, $2d, $31, $34, $00, $00
          , $49, $45, $4e, $47, $0f, $00, $00, $00
          , $42, $65, $72, $6e, $61, $72, $64, $20
          , $47, $69, $73, $69, $6e, $00, $00, $01
          , $49, $53, $46, $54, $23, $00, $00, $00
          , $53, $6f, $6e, $79, $20, $53, $6f, $75
          , $6e, $64, $20, $46, $6f, $72, $67, $65
          , $20, $41, $75, $64, $69, $6f, $20, $53
          , $74, $75, $64, $69, $6f, $20, $38, $2e
          , $30, $00, $00, $02
          );

var
  nn      : word32;    // position in PWavSound
  nLenData : int32;  // number of bytes coding the sound, without the heading and footing.
  nLen : int32;
  b1, b2, b3, b4 : byte;
  vT0  : Float;  // time
  vFreq : Float; // frequency

begin
// Creates a copy of the wav sound in the paSound variable

// Header
for nn:=0 to 43 do paS^[nn]:=aHead[nn];
nn:=43;

// Sound code
vT0:=0;      // temps en secondes
vFreq:=440;  // frequency in Hz
repeat
  inc(nn); paS^[nn]:=Round(128+120*sin(2*3.1415928*vFreq*vT0));  // 440 = the sound frequency
  vT0:=vT0 + 1/48000;  //  / "the sampling frequency"
  vFreq:=440 + 14000*vT0/(30*2);  // change slowly the frequency with the time.
  // if signal = sin(2*pi*w(t)), frequency = derivative w'(t).  Here,  w(t) = 220*t +9000*t^2/(30*2),  w'(t)=220 +9000*t/30
  until (vT0 > 30);  // sets the playing time in seconds

// number of data
nLenData:=nn;

// for b1:=1 to 200 do DBW(1, ITS(word(paS^[5000-1*integer(b1)])));

// Footer
for b1:=0 to 99 do begin inc(nn); paS^[nn]:=aFoot[b1]; end;

// writing the "file" length
nLen := nLenData + 144 - 8;
b1:= nLen         mod $100;
b2:=(nLen div $100) mod $100;
b3:=(nLen div $10000) mod $100;
b4:=(nLen div $1000000) mod $100;
paS^[4]:=b1;  paS^[5]:=b2;  paS^[6]:=b3;  paS^[7]:=b4;

// writing the data length
nLen := nLenData;
b1:= nLen         mod $100;
b2:=(nLen div $100) mod $100;
b3:=(nLen div $10000) mod $100;
b4:=(nLen div $1000000) mod $100;
paS^[40]:=b1;  paS^[41]:=b2;  paS^[42]:=b3;  paS^[43]:=b4;

// the sound is in memory, plays it.
PlaySound(LPCSTR(paS), 0, SND_MEMORY or SND_ASYNC);  // SND_ASYNC return immediatly   or  SND_NODEFAULT
end;  // Test48kHz;

function TMyDialog.WMCommand(wPar:WPARAM; lPar:LPARAM):BOOL;
//==========================================================
// Handles the menu messages
var
  wCtrlID     : word32;

begin
if (HiWord(wPar) <> BN_CLICKED) then begin
   WMCommand:=inherited WMCommand(wPar, lPar);
   exit;
   end;
//DBW(1, 'WMCommand=' + DBNotifToStr(HiWord(wPar)) + ITS5(LoWord(wPar)));
wCtrlID:=LoWord(wPar);

//MessageBeep(0);

//DBW(1, 'Button clicked=' + ITS(poBtn^.wCtrlID));

// sets output D0 to 0 [V]
if (wCtrlID = 100) then begin
   Test1;
   end;

// sets output D0 to more than 2 [V]
if (wCtrlID = 101) then begin
   MessageBeep(MB_ICONASTERISK);
   end;

// sets output D1 to 0 [V]
if (wCtrlID = 110) then begin
   MessageBeep(MB_ICONEXCLAMATION);
   end;

// sets output D1 to more than 2 [V]
if (wCtrlID = 111) then begin
   MessageBeep(MB_ICONHAND);
   end;

// sets Led OFF on the PCS10
if (wCtrlID = 300) then begin
   MessageBeep(MB_ICONQUESTION);
end;

// sets Led ON on the PCS10
if (wCtrlID = 301) then begin
   MessageBeep(MB_OK);
   end;

// sets the gain to 1 on all channels on the PCS10
if (wCtrlID = 401) then begin
   MessageBeep(0);
   end;

// sets the gain to 2 on all channels on the PCS10
if (wCtrlID = 402) then begin
   MessageBeep(0);
   end;

// sets the gain to 5 on all channels on the PCS10
if (wCtrlID = 405) then begin
   PlaySound('Self_defined_sound_1.wav', 0, SND_ASYNC or SND_FILENAME);
   end;

// sets the gain to 10 on all channels on the PCS10
if (wCtrlID = 410) then begin
   PlaySound('Self_defined_sound_2.wav', 0, SND_ASYNC or SND_FILENAME);
   end;

// Stops the playing sound
if (wCtrlID = 700) then begin
   PlaySound(NIL, 0, SND_PURGE);
   end;

// Plays a sin sound
if (wCtrlID = 710) then begin
   Test2;
   end;

// Plays a sin sound sampled at 100 kHz
if (wCtrlID = 801) then begin
   Test100kHz;
   end;

// Plays a sin sound sampled at 200 kHz
if (wCtrlID = 802) then begin
   Test200kHz;
   end;

// Plays a sin sound sampled at 200 kHz
if (wCtrlID = 803) then begin
   Test48kHz;
   end;

// Did we click on the quit button ?
// Close the main window, thus quit the application.
if (wCtrlID = 999) then  PostMessage(poMyMainWin^.hWin, WM_CLOSE, 0, 0);

WMCommand:=inherited WMCommand(wPar, lPar);
end;  // TMyDialog.WMCommand

//******************************************************************
// TMyWindow = object(TWindow)
//******************************************************************

constructor TMyWindow.Init(strWinTitlePrm:ShortString);
//=====================================================
begin
new(paS);  // To save the sound into memory.

inherited Init(strWinTitlePrm);
end;  // TMyWindow.Init

destructor TMyWindow.Done;
//========================
begin
inherited Done;

dispose(paS); // free the memory from the data.
paS:=nil;

end;

procedure TMyWindow.WinCreate(strWinClass:ShortString);
//=====================================================
//var strName : string[50];
Begin
// We don't want to see the main window
//wSizeState := SIZE_MINIMIZED;

inherited WinCreate(strWinClass);

// Here calling :  DialogExecMSG(11199, @Self);  doesn't work,
// But posting a message to open the dialog box at startup is possible.
PostMessage(hWin, WM_COMMAND, 101, 0);

//strName:='zlportio' + char(0);
//DBW(0, 'result= ' + ITS(driverstart(@strName[1])));

end;  // TMyWindow.WinCreate

function TMyWindow.WMCommand(wPar:WPARAM; lPar:LPARAM):LRESULT;
//=============================================================
// treats the menu commands
begin
// Execute the dialog box at starting
// And the help window
if (lPar = 0) and (wPar =101) then DialogCreateMSG(11199, @Self);
//if (lPar = 0) and (wPar =121) then DialogCreateMSG(21199, @Self);

// Accelerators act as button click.
// Send the command to the active dialog
case LoWord(wPar) of
  999 : if (gglpoActiveDialog <> NIL) then
               gglpoActiveDialog^.WMCommand(LoWord(wPar), 0);
  end; // case

WMCommand:=inherited WMCommand(wPar, lPar);
end;  // TMyWindow.WMCommand

function TMyWindow.WindowProc(hWindow:HWND; wMsg:UINT; wPar:WPARAM; lPar:LPARAM):LRESULT;
//=======================================================================================
// All window messages are send here.

Begin
WindowProc:=inherited WindowProc(hWindow, wMsg, wPar, lPar);

{ if (wMsg <> WM_MouseMove)   and (wMsg <> WM_NCMouseMove) and
   (wMsg <> WM_MENUSELECT )   and (wMsg <> WM_ENTERIDLE ) and
   (wMsg <> WM_SetCursor)   and (wMsg <> WM_NCHittest) and
   (wMsg <> WM_SetFocus) and (wMsg <> WM_KillFocus) and
   (wMsg <> WM_KEYDOWN) and (wMsg <> WM_KEYUP) and
   (wMsg <> WM_Move)        and (wMsg <> WM_NCActivate) then
   DBW0(32, DBMsgToStr(wMsg) + ITS(wPar) + ITS(lPar)); }

//if (wMsg = WM_Destroy) then begin
//   end;

End;  // TMyWindow.WindowProc

//**********************************************************************
Begin // MAIN ******
// Assign the pointer on the function which will create the dialog.
// It is used by the library. So the library will construct
// a user defined object.
gglfnCreateDialogObj:=@myCreateDialogObj;

// Name under which the main window is registered. Spaces are not allowed
oApp.Init('Example072');

// Creation of the main window
new(poMyMainWin, Init('tEx072_Sound_Tests.pp'));

// The App.WindowClass can be modified here, if wanted. (Not usual)
oApp.Run;
oApp.Done;  // The main window object is disposed here.
End.

