{ Communicate with the arduino microcontroller chip
  through the serial port.  c.f. http://arduino.cc/

  --------------------------------------------}
program tEx079_Serial_Arduino;

// To tell FreePascal that the application is a Graphical User Interface
// i.e. is running in windows, not in text mode.
{$APPTYPE GUI}

Uses
  wApplication0
 ,wWindow0
 ,wToolBar0
 ,wDialog0
 ,wPopupWindow0
 ,wDialogCreatorLOW2
 ,wDialogCreator2
 ,wVariableNumLink2
 ,wWindowBase0
 ,wErrors0
 ,wGlobals2
 ,wDebug0
 ,Windows
// ,zlportio  // to access the parallel port or any other port
 ,wReadTime // to read time intervals with accuracy.
 ;

type
  PMyDialog = ^TMyDialog;
  TMyDialog = object(TDialog)
  //"""""""""""""""""""""""""

  vXValue : Float;
  // the value shown in the edit box, must be different if we don't
  // want rounding errors to occure.

  fClearNext : boolean;
  // The next digit clears the avValue[1] and replaces it by the digit

  fAddNext : boolean;
  // The next digit push the stack up and
  //  clears the avValue[1] and replaces it by the digit

  vPeriode : Float;
  // If the periode has been pressed, indicate how many digits are after the
  // period.

  vMemory : Float;
  // A simple memory.

  constructor Init(poParentW:PWindowBase);

  function WMCommand(wPar:WPARAM; lPar:LPARAM):BOOL; virtual;
  // Handles the menu messages

  function DialogProc(hDialog:HWND; wMsg:UINT; wPar:WPARAM; lPar:LPARAM):LRESULT; virtual;
  // Handles the dialog messages. See comments in  wDialog0.pp
  end; // TMyDialog = object(TDialog)

type
  PMyWindow = ^TMyWindow;
  TMyWindow = object(TWindow)
  //"""""""""""""""""""""""""
  poToolBar : PToolBar;
  // Remember the toolbar, to destroy it when a new one is wanted.

  constructor Init(strWinTitlePrm:ShortString);

  destructor Done; virtual;

  procedure WinCreate(strWinClass:ShortString); virtual;
  // When creating the window, we ask to show the dialog box, not the window

  procedure WMPaint; virtual;
  // the system want the program to repaint the client area.

  function WMCommand(wPar:WPARAM; lPar:LPARAM):LRESULT; virtual;
  // Handles the menu messages

  function WindowProc(hWindow:HWND; wMsg:UINT; wPar:WPARAM; lPar:LPARAM):LRESULT; virtual;
  end; // TMyWindow = object(TWindow)

var
  oApp        : TApplication;
  poMyMainWin : PMyWindow;

  glfStopGenerator : boolean;
  // The Start_Generator procedure loops, until one click in the dialog box to stop it

  glfStopMeasure : boolean;
  // Allow to stop the Measurement

  glvInPeriod     : Float;
  // Period between two pulses from the BUSY pin 11,  to be Displayed in big in the main window

  glwPulseWidth : word32;
  // width in microseconds of the pulse of light generated by the Stroboscope
  // Variable numbre : 1001.  Word32Field(1001, glwPulseWidth);

  glafCheckOut : array[0..7] of boolean;

function myCreateDialogObj(poParent:Pointer):Pointer;
//===================================================
var poObj : PMyDialog;
begin
new(poObj, Init(poParent));
myCreateDialogObj:=poObj;
end;  // myCreateDialogObj

//******************************************************************
// TMyDialog = object(TDialog)
//******************************************************************

procedure Test1;
//==============
var
  hPort     : Handle;
  fOK       : boolean; // if comm states has been set correctly.
  dcbStates : DCB;     // address of device-control block structure
  timeOuts  : COMMTIMEOUTS;  // to define the timeOut
  abData    : array[0..10] of byte;  // data from the serial port
  wRidden   : DWORD; // number of data ridden
  nn        : integer;
  acError   : array[0..1024] of char;  // message d'erreur
  strS      : string;

begin
DBW(1, 'hello');

// opens an access to a serial port.
hPort:=CreateFile(
     'COM7' // LPCTSTR lpFileName,	// pointer to name of the file
    ,GENERIC_READ	or GENERIC_WRITE // DWORD dwDesiredAccess,	// access (read-write) mode
    ,0 // FILE_SHARE_READ or FILE_SHARE_WRITE // DWORD dwShareMode,	// share mode
    ,NIL // LPSECURITY_ATTRIBUTES lpSecurityAttributes,	// pointer to security attributes
    ,OPEN_EXISTING	 // DWORD dwCreationDistribution,	// how to create
    ,FILE_ATTRIBUTE_NORMAL // FILE_ATTRIBUTE_SYSTEM // DWORD dwFlagsAndAttributes,	// file attributes
    ,0 // HANDLE hTemplateFile 	// handle to file with attributes to copy
   );

DBW(1, 'hPort=' + ITS(hPort));

if (hPort = INVALID_HANDLE_VALUE) then begin  // INVALID_HANDLE_VALUE = -1
  DBW(1, 'LastError=' + ITS(GetLastError) + ITS(ERROR_FILE_NOT_FOUND));

FormatMessage(
  FORMAT_MESSAGE_FROM_SYSTEM or FORMAT_MESSAGE_IGNORE_INSERTS,
  nil,
  GetLastError,
  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
  @acError,
  1024,
  nil);

  strS:='';
  nn:=0;
  while ((Byte(acError[nn]) <> 0) and (nn < 255)) do begin
    strS:=strS + acError[nn];
    inc(nn);
    end;

  DBW(1, 'Erreur = ' + strS);
  exit;
  end;

dcbStates.DCBlength:=sizeof(dcbStates);   // sizeof(DCB)

// Sets default values
fOK:=GetCommState(hPort, @dcbStates);

dcbStates.BaudRate:=CBR_9600;            // current baud rate
dcbStates.ByteSize:=8;             // number of bits/byte, 4-8
dcbStates.Parity:=NOPARITY;    // 0-4=no,odd,even,mark,space
dcbStates.StopBits:=ONESTOPBIT;             // 0,1,2 = 1, 1.5, 2


fOK:=SetCommState(
    hPort  // HANDLE hFile,	// handle of communications device
   ,dcbStates
   );

DBW(1, 'OK=' + ITS(int32(fOK)));

timeOuts.ReadIntervalTimeout:=1000;  // en millisecondes
timeOuts.ReadTotalTimeoutMultiplier:=10;  // ??
timeOuts.ReadTotalTimeoutConstant:=1000; // ms
timeOuts.WriteTotalTimeoutMultiplier:=50;  // ms
timeOuts.WriteTotalTimeoutConstant:=50;    // ms

fOK:=SetCommTimeouts(
    hPort
   ,timeOuts	// address of communications time-out structure
   );

DBW(1, 'OK=' + ITS(int32(fOK)));

for nn:=1 to 3 do begin

fOK:=ReadFile(
    hPort
   ,abData  // address of buffer that receives data
   ,2        // DWORD nNumberOfBytesToRead, // number of bytes to read
   ,wRidden // LPDWORD lpNumberOfBytesRead, // address of number of bytes read
   ,nil
   );

DBW(1, 'OK=' + ITS(int32(fOK)));
DBW(1, 'Data=' + ITS(abData[0]) + ITS(abData[1]) + ITS(wRidden));
end;

(*
BOOL WriteFile(

    HANDLE hFile,	// handle to file to write to
    LPCVOID lpBuffer,	// pointer to data to write to file
    DWORD nNumberOfBytesToWrite,	// number of bytes to write
    LPDWORD lpNumberOfBytesWritten,	// pointer to number of bytes written
    LPOVERLAPPED lpOverlapped 	// pointer to structure needed for overlapped I/O
   );
*)

// close the communication
CloseHandle(hPort);
end;

constructor TMyDialog.Init(poParentW:PWindowBase);
//================================================
begin
// must be done, even if it seems useless.
inherited Init(poParentW);

vXValue:=0.0;
FloatField(100, vXValue);

fClearNext :=FALSE;
fAddNext   :=FALSE;
vPeriode   :=1; // no digits after the period.
vMemory    :=0;
glfStopGenerator :=true;  // When starting, the generator is not running
glfStopMeasure   := true; // No measurements are done at the beginning.

//FloatField(900, vFrequency);

//Int32Field(915, nStateERR);

//FloatField(921, vInFrequency);

end;  // TMyDialog.Init

function TMyDialog.WMCommand(wPar:WPARAM; lPar:LPARAM):BOOL;
//==========================================================
// Handles the menu messages
var
  rMessage    : Msg;
  wCtrlID     : word32;
  wSpeed,
  wCpt        : word32;
  dnTick1, dnTick2  // to read the time
              : int64;
  dwTimeNS  // elapsed time in nano seconds
              : dword;
  pDial       : PWindowBase;  // Pointer on a new dialog box

begin
{
if (HiWord(wPar) <> EN_UPDATE) and (HiWord(wPar) <> EN_CHANGE) then
  DBW(0, 'WMCommand=' + DBNotifToStr(HiWord(wPar)) + ITS5(LoWord(wPar)));
}
if (HiWord(wPar) <> BN_CLICKED) then begin
   WMCommand:=inherited WMCommand(wPar, lPar);
   exit;
   end;
wCtrlID:=LoWord(wPar);

// do only accept the Message to stop the Generator if it is not running
if ((not glfStopGenerator) and (wCtrlID < 1010)) then exit;

//MessageBeep(0);

//DBW(1, 'Button clicked=' + ITS(poBtn^.wCtrlID));

// Ground are pins 18 to 25 on a 25 pins connector.

// sets the 8 outputs D0..D7 to 0 [V]  (pin 2 to 9)
if (wCtrlID = 109) then begin
   DBW(0, 'set outputs to off');
   Test1;
   end;

// sets the 8 outputs D0..D7 to OPEN COLLECTOR [V]  (pin 2 to 9)
if (wCtrlID = 119) then begin
   DBW(0, 'set outputs to on');
   end;

// sets output D0 to open collector (pin 2)
if (wCtrlID = 100) then begin
   DBW(0, 'set D0 to off');
   end;

// sets output D0 to open collector (pin 2)
if (wCtrlID = 110) then begin
   DBW(0, 'set D0 to on');
   end;

// sets output D6 to 0 [V] (pin 8)
if (wCtrlID = 106) then begin
   DBW(0, 'set D6 to off');
   end;

// sets output D6 to open collector 2 [V] (pin 8)
if (wCtrlID = 116) then begin
   DBW(0, 'set D6 to on');
   end;

// Test the accuracy timer to measure the time.
if (wCtrlID = 201) then begin
   wCpt:=0;

   // CPU frequency in MHz
   wSpeed:=CPUGetSpeed;
   DBW(0, 'CPU Frequency in MHz = ' + WTS(wSpeed));
   DBW(0, 'CPU time periode in ns = ' + FTS(_1CPUFrequencyNS));

   dnTick1:=cpugettick;  // get number of CPU clock since computer started
   for wCpt:=1 to 10000 do begin wSpeed:=wSpeed; end;  // loose some time

   dnTick2:=cpugettick;  // get number of CPU clock since computer started
   dwTimeNS:=cpugetNS(dnTick1); // get elapsed time in nano seconds


   DBW(0, 'Number of CPU clock =' + WTS(dnTick2) + ' - ' + WTS(dnTick1) + ' = ' +  WTS(dnTick2 - dnTick1));
   DBW(0, 'Delta time [ns] =' + WTS(dwTimeNS-37) + ' ~= ' + FTS((dnTick2 - dnTick1)*_1CPUFrequencyNS));
   end;

// Generate a frequency on D0 (pin 2) of the parallel port.
if (wCtrlID = 301) then Test1;

// Open a new modeless dialog box to control the frequency
if (wCtrlID = 401) then DialogCreateMSG(11299, @Self);

// Open a new modeless dialog box to measure the inputs of the parallel port
if (wCtrlID = 411) then begin
  pDial:=DialogCreateMSG(11399, @Self);
  PostMessage(pDial^.hWin, WM_Command, 1011, 0);  // Inputs_Measure; is started
  end;

// Open a new modeless dialog box to measure the frequency of a signal
if (wCtrlID = 412) then begin
  pDial:=DialogCreateMSG(11499, @Self);
  PostMessage(pDial^.hWin, WM_Command, 1012, 0);  // Frequency_Measure; is started
  end;

// Stop generating pulses
if (wCtrlID = 1010) then begin
  glfStopGenerator:=true;
  DBW(0, 'STOPPING');
  end;

// Stop the measurements
if (wCtrlID = 1019) then begin
  glfStopMeasure:=true;
//  PostMessage(hWin, WM_Close, 0, 0);
  end;

//UpdateFromVar; // update the dialog box

while (PeekMessage(@rMessage, 0, // any messages for this application
                    0, 0,         // Let the dialog box being updated
                    PM_REMOVE))   // remove the message from the queue
                    do begin
  TranslateMessage(rMessage);
//      DBW(0, ITS(rMessage.message) + '  ' + ITS(LoWord(rMessage.wParam)) + '  ' + ITS(HiWord(rMessage.wParam)) + '  ' + ITS(LoWord(rMessage.lParam)) + '  ' + ITS(HiWord(rMessage.lParam)) + '  ' + ITS(rMessage.hwnd));
  DispatchMessage(rMessage);
  end;

// Did we click on the quit button ?
// Close the main window, thus quit the application.
if (wCtrlID = 999) then begin
  // Quit the programme if no measurements are in progress.
  if (glfStopGenerator and glfStopMeasure) then PostMessage(poMyMainWin^.hWin, WM_CLOSE, 0, 0);
  glfStopGenerator:=true;
  glfStopMeasure:=true;
  exit;
  end;

WMCommand:=inherited WMCommand(wPar, lPar);
end;  // TMyDialog.WMCommand

function TMyDialog.DialogProc(hDialog:HWND; wMsg:UINT; wPar:WPARAM; lPar:LPARAM):LRESULT;
//======================================================================================
// Accept the  WM_CLOSE  message, only if no measurements are in action.
begin
{
if (wMsg <> WM_COMMAND) and
   (wMsg <> WM_MOUSEMOVE) and
   (wMsg <> WM_MOVE) and
   (wMsg <> WM_NCCALCSIZE) and
//   (wMsg <> WM_WINDOWPOSCHANGE) and
   (wMsg <> WM_NCHITTEST) and
   (wMsg <> WM_ACTIVATEAPP) and
   (wMsg <> WM_ACTIVATE) and
   (wMsg <> 674) and
   (wMsg <> WM_CTLCOLORBTN) and
   (wMsg <> WM_NCACTIVATE) and
   (wMsg <> WM_GETICON) and
   (wMsg <> WM_SETTEXT) and
   (wMsg <> 70) and  // WM_WINDOWPOSCHANGING
   (wMsg <> WM_INITDIALOG) and
   (wMsg <> WM_NCPAINT) and
   (wMsg <> WM_WINDOWPOSCHANGED) and
   (wMsg <> WM_SIZE) and
   (wMsg <> WM_GETTEXT) and
   (wMsg <> WM_PARENTNOTIFY) and
   (wMsg <> WM_STYLECHANGING) and
   (wMsg <> WM_STYLECHANGED) and
   (wMsg <> WM_SHOWWINDOW) and
   (wMsg <> WM_ERASEBKGND) and
   (wMsg <> WM_CTLCOLORDLG) and
   (wMsg <> WM_PAINT) and
   (wMsg <> 295) and
   (wMsg <> WM_CTLCOLOREDIT) and
   (wMsg <> WM_NCMOUSEMOVE) and
   (wMsg <> WM_MOUSEACTIVATE) and
   (wMsg <> WM_PSD_PAGESETUPDLG) and
   (wMsg <> 296) and
   (wMsg <> 174) and
   (wMsg <> 136) and
   (wMsg <> WM_SYSCOMMAND) and
   (wMsg <> WM_CAPTURECHANGED) and
   (wMsg <> WM_SETCURSOR) then Begin
  DBW(0, 'Proc=' + DBMsgToStr(wMsg) + ITS5(HiWord(wPar)) + ITS5(LoWord(wPar)));
  end;
}
Case wMsg Of
   WM_Close   : if ((not glfStopGenerator) or (not glfStopMeasure)) then begin
                   // Do not procede the Close message if measurements are in progress.
                   // But stop the measurements, which will generate a WM_CLOSE
                   glfStopGenerator:=true;
                   glfStopMeasure:=true;
                   exit;
                   end;
   End; // Case wMsg Of

DialogProc:=inherited DialogProc(hDialog, wMsg, wPar, lPar);
end;  // TMyDialog.DialogProc

//******************************************************************
// TMyWindow = object(TWindow)
//******************************************************************

constructor TMyWindow.Init(strWinTitlePrm:ShortString);
//=====================================================
var nn : integer;
begin
inherited Init(strWinTitlePrm);

glvInPeriod:=0;
glwPulseWidth:=100;
Word32Field(1001, glwPulseWidth);

// The check boxes to set the outputs
for nn:=0 to 7 do begin
  booleanField(120 + nn, glafCheckOut[nn]);
  glafCheckOut[nn]:=false;
  end;

poToolBar:=NIL;
end;  // TMyWindow.Init

destructor TMyWindow.Done;
//========================
begin
glfStopGenerator:=true;
glfStopMeasure:=true;

inherited Done;
end;

procedure TMyWindow.WinCreate(strWinClass:ShortString);
//=====================================================
//var strName : string[50];
Begin
// We don't want to see the main window
//wSizeState := SIZE_MINIMIZED;

inherited WinCreate(strWinClass);

// Here calling :  DialogExecMSG(11199, @Self);  doesn't work,
// But posting a message to open the dialog box at startup is possible.
// Opens also the help window, indicating the key action.
//PostMessage(hWin, WM_COMMAND, 121, 0);
PostMessage(hWin, WM_COMMAND, 101, 0);

//strName:='zlportio' + char(0);
//DBW(0, 'result= ' + ITS(driverstart(@strName[1])));

poToolBar:=PToolBar(ToolBarCreateMSG(10110, @self));
end;  // TMyWindow.WinCreate

procedure TMyWindow.WMPaint;
//==========================
Var
  hDcPaint    : hdc;
  rPS         : paintstruct;
  nLineY      : integer; // vertical position ot the text in the line

// shows some character fonts.
var
  strS        : string[80];
  rLogFont    : TLogFont;
  hFontNew1,  // four style of font, for normal, bold and italic.
  hFontOld    : HFONT;
  astrFont    : array[1..4] of string[30];
  vFrequency  : Float;
  vRPM        : Float; // Round Per Minute

begin
// Get a device context.
// Here we have a facility, during the WMPaint, we must always
// begin with  "BeginPaint", which gives a Device Context to paint, and
// end with "EndPaint", to tell that the painting is done.
hDcPaint:=BeginPaint(hWin, @rPS);
// fonts to be used :
astrFont[1]:='Arial';
astrFont[2]:='Times New Roman';
astrFont[3]:='Courier New';
astrFont[4]:='Comic Sans MS';  // if you have it.

// define a new font.
strS:=astrFont[1]+#0;
with rLogFont do  begin
   lfHeight         := 0;                // Default logical height of font
   lfWidth          := 0;                // Default logical average character width
   lfEscapement     := 0;                // angle of escapement
   lfOrientation    := 0;                // base-line orientation angle
   lfWeight         := FW_NORMAL;        // font weight = 400
   lfItalic         := 0;                // italic attribute flag
   lfUnderline      := 0;                // underline attribute flag
   lfStrikeOut      := 0;                // strikeout attribute flag
   lfCharSet        := DEFAULT_CHARSET;  // character set identifier
   lfOutPrecision   := OUT_DEFAULT_PRECIS;  // output precision
   lfClipPrecision  := CLIP_DEFAULT_PRECIS; // clipping precision
   lfQuality        := DEFAULT_QUALITY;     // output quality
   lfPitchAndFamily := DEFAULT_PITCH;    // pitch and family
   lfFaceName       :=PCHAR(@strS[1]);         // pointer to typeface name string
   end; // with

// creates the new Font.
// CreateFont(...) could also be used.
rLogFont.lfWeight:=700; // bold, could go up to 1000 for very bold
rLogFont.lfHeight:=-36;
hFontNew1 := CreateFontIndirect(@rLogFont);

nLineY:=30;

// If you want to draw text in the selected font.
hFontOld:=SelectObject(hDcPaint, hFontNew1);
strS:='Period = ' + FloatToStr(glvInPeriod,8,3) + ' [ms]';
TextOut(hDcPaint, 10, nLineY, @strS[1], length(strS));

nLineY:=nLineY + 50;
if (glvInPeriod <> 0) then vFrequency:=1000.0 / glvInPeriod else vFrequency:=0;
strS:='Frequency = ' + FloatToStr(vFrequency,8,3) + ' [Hz]';
TextOut(hDcPaint, 10, nLineY, @strS[1], length(strS));

nLineY:=nLineY + 50;
vRPM:=60*vFrequency;
strS:='Round per Minute = ' + FloatToStr(vRPM,9,2);
TextOut(hDcPaint, 10, nLineY, @strS[1], length(strS));

// As always, you must select the original font, the delete
// the created "objects" to free resources.
SelectObject(hDcPaint, hFontOld);
DeleteObject(hFontNew1);

EndPaint(hWin, rPS);
end;  // TMyWindow.WMPaint

function TMyWindow.WMCommand(wPar:WPARAM; lPar:LPARAM):LRESULT;
//=============================================================
// There are no menus treatment, only one dialog box
var
  pDial       : PWindowBase;  // Pointer on a new dialog box

begin
// Execute the dialog box at starting
if (lPar = 0) and (wPar =101) then DialogCreateMSG(11199, @Self);
//if (lPar = 0) and (wPar =121) then DialogCreateMSG(21199, @Self);

// Accelerators act as button click.
// Send the command to the active dialog
case LoWord(wPar) of
  // Dialog box to control the frequency output
  111 : DialogCreateMSG(11299, @Self); // di

  // Open a new modeless dialog box to measure the inputs of the parallel port
  121 : begin
     pDial:=DialogCreateMSG(11399, @Self);
     PostMessage(pDial^.hWin, WM_Command, 1011, 0);  // Inputs_Measure; is started
     end;

  // Open a new modeless dialog box to measure the frequency of a signal
  131 : begin
     pDial:=DialogCreateMSG(11499, @Self);
     PostMessage(pDial^.hWin, WM_Command, 1012, 0);  // Frequency_Measure; is started
     end;

  // Open a new modal dialog box to Sets the outputs.
  151 : begin
     DialogExecMSG(11599, @Self);
     end;

//  999
//          : if (gglpoActiveDialog <> NIL) then
//               gglpoActiveDialog^.WMCommand(LoWord(wPar), 0);

  999 : begin  // We clicked on the quit button.
     // Close the main window, thus quit the application.
     // Quit the programme if no measurements are in progress.
     if (glfStopGenerator and glfStopMeasure) then PostMessage(poMyMainWin^.hWin, WM_CLOSE, 0, 0)
     else begin
        glfStopGenerator:=true;  // stops the generator
        glfStopMeasure:=true;    // stops the measurements
//        PostMessage(poMyMainWin^.hWin, WM_Command, 999, 0); // post again a quit message
        end;
     end;
  end; // case

WMCommand:=inherited WMCommand(wPar, lPar);
end;  // TMyWindow.WMCommand

function TMyWindow.WindowProc(hWindow:HWND; wMsg:UINT; wPar:WPARAM; lPar:LPARAM):LRESULT;
//=======================================================================================
// All window messages are send here.
Begin
WindowProc := 0; // Default return value when processing a message

Case wMsg Of
   WM_Paint   : WMPaint; // must redraw the client area part.
   WM_Command : WindowProc:=WMCommand(wPar, lPar);
   else begin // Case wMsg Of
      // Default processing of the messages.
      WindowProc:=inherited WindowProc(hWindow, wMsg, wPar, lPar);
      end; // Case wMsg Of
   End; // Case wMsg Of

End;  // TMyWindow.WindowProc

//**********************************************************************
Begin // MAIN ******
// Assign the pointer on the function which will create the dialog.
// It is used by the library. So the library will construct
// a user defined object.
gglfnCreateDialogObj:=@myCreateDialogObj;

// Name under which the main window is registered. Spaces are not allowed
oApp.Init('tEx079_Serial_Arduino');

// Creation of the main window
new(poMyMainWin, Init('tEx079_Serial_Arduino'));

// The App.WindowClass can be modified here, if wanted. (Not usual)
oApp.Run;
oApp.Done;  // The main window object is disposed here.
End.
