unit wAutoAxes0;

{ Unit pour le calcul et l'affichage automatique des axes d'un graphique,
  dessin automatique d'une grille sur le graphique.
  Le choix automatique des axes d'un graphique contient beaucoup de
  paramtres, car on peut dsirer plusieurs reprsentations diffrentes
  telles que :
  -	Position des axes, haut, bas, origine, droite, gauche,...
  -	Position des Tics sur les axes
  -	Nombre de Tics majeurs et mineur sur les axes.
  -	Prsence ou absance de Tics et d'chlle sur les axes.
  -	Coincidence des limites du graphique avec les Tics majeurs ou mineur.
  -	Correspondance de l'axe X avec l'axe Y.
  -	Respect d'apparence d'un cercle.
  -	Echelle minimale
  -	Augmentation d'chelle par rapport  celle calcule pour les courbes.
  -	...

  Ceci justifie la cration d'un objet ddi au calcul du choix automatique
  des axes, avec des paramtres par dfaut, facilement ajustables.
-------------------------------------------------------------------------}

// To tell FreePascal that the application is a Graphical User Interface
// i.e. is running in windows, not in text mode.
{$APPTYPE GUI}

interface
{"""""""}

Uses
  wGrBase0
 ,wWindow0
// ,wPrinter0
 ,wErrors0
 ,wWindowBase0
 ,wGlobals2
 ,wDebug0
 ,Strings
 ,Windows;

const
{ Cas particuliers pour la compatibilit avec Quinn-Curtis v7.x --- }
{ if (lFlags = 0) then lFlags:=AxesInterceptX0 or AxesInterceptY0;
  if (lFlags = 2) then lFlags:=AxesInterceptXUp or AxesInterceptYRight;
  if (lFlags = 3) then lFlags:=AxesInterceptXUp; }

{ Si pas contre indiqu, les axes se croisent  gauche et en bas }
flaAxesInterceptXDownYLeft   = 1 shl  2; {= 4}

{ L'axe X croise l'axe Y en 0 si possible, sinon en bas,
  ou en haut si flaAxesIntercepteXUp est donn. }
flaAxesInterceptX0           = 1 shl  3; {= 8}

{ L'axe Y croise l'axe X en 0 si possible. }
flaAxesInterceptY0           = 1 shl  4; {=16}

{ Axes X et Y se croisent en (0,0) si possible. }
flaAxesInterceptX0Y0         = flaAxesInterceptX0 or flaAxesInterceptY0;

{ L'axe X est en haut, ou en 0 si  flaAxesInterceptX0 est donn. }
flaAxesInterceptXUp          = 1 shl  5;

{ L'axe Y est a droite, ou en 0 si flaAxesInterceptY0 est donn. }
flaAxesInterceptYRight       = 1 shl  6;

{ Axes X et Y se croisent en haut  droite. }
flaAxesInterceptXUpYRight    = flaAxesInterceptXUp or flaAxesInterceptYRight;

{ Utilise les donnes pour afficher plusieurs courbes. c.f. AdjustLimits }
flaAxesMultiCurves           = 1 shl  7;

{ Imposes des limites sur les dimensions des axes. c.f. vXMinMin, ... }
{flaAxesWithLimitsValues      = 1 shl  8; c.f. 1 shl 24}

{ Impose que les limites de l'axe X se terminent sur des Tics majeurs }
flaAxesBorderXOnTic          = 1 shl  9;

{ Impose que les limites de l'axe Y se terminent sur des Tics majeurs }
flaAxesBorderYOnTic          = 1 shl 10;

{ Ne met pas de Tics sur l'axe X }
flaAxesNoTicX                = 1 shl 11;

{ Ne met pas de Tics sur l'axe Y }
flaAxesNoTicY                = 1 shl 12;

{ Ne met pas de Label sur l'axe X }
flaAxesWithoutXLabel         = 1 shl 13;

{ Ne met pas de Label sur l'axe Y }
flaAxesWithoutYLabel         = 1 shl 14;

{ Les bornes des axes en X et Y sont les mmes. (isotrop) }
flaAxesSameXY                = 1 shl 15; {=32768}

{ Les limites de l'axe X sont grandis de vXLimitIncrease }
flaAxesNotSmallX             = 1 shl 16; {=65536}

{ Les limites de l'axe Y sont grandis de  vYMinLimitIncrease et vYMaxLimitIncrease }
flaAxesNotSmallY             = 1 shl 17;

{ Pour avoir un axe X symtrique }
flaAxesSymetricXAxes         = 1 shl 18;

{ Pour avoir un axe Y symtrique }
flaAxesSymetricYAxes          = 1 shl 19;

{ Seul le calcul des Axes se fait, pas le dessin, qui peut tre fait
  aprs coup avec DrawX(lFlags) et DrawY(lFlags). }
flaAxesNoDraw                = 1 shl 20;

{ Demande de dessiner un cadre autour du graphique }
flaAxesWithFrame             = 1 shl 21;

{ pour qu'il n'y ait pas de label X sur le cadre.
  Ceci est automatique si flaAxesWithFrame est slectionn }
flaAxesNoXLabelOnFrame       = 1 shl 22;

{ pour qu'il n'y ait pas de label Y sur le cadre.
  Ceci est automatique si flaAxesWithFrame est slectionn }
flaAxesNoYLabelOnFrame       = 1 shl 23;

{ Imposes des limites sur les dimensions des axes. c.f. vXMinMin, ... }
flaAxesWithLimitsValuesX     = 1 shl 24;
flaAxesWithLimitsValuesY     = 1 shl 25;
flaAxesWithLimitsValuesXY    = flaAxesWithLimitsValuesX or
                               flaAxesWithLimitsValuesY;

{c30   = 1 shl 30; // = 1'073'741'824}
{c31   = 1 shl 31; // =-2'147'483'648}

const
flaDrawTicAuto               =  0;
flaDrawTicLeft               =  1;
flaDrawTicDown               =  flaDrawTicLeft;
flaDrawTicRight              =  2;
flaDrawTicUp                 =  flaDrawTicRight;
flaDrawTicCross              =  4; { les Tics croisent l'axe }

const
flaAxesX                     = 1;
flaAxesY                     = 2;

type
PAutoAxes = ^TAutoAxes;
TAutoAxes = object
{================}
  { pointeur sur le graphisme de base utilis pour le dessin }
  pGrBase           : PGraphicBase;

  { Min et Max des axes en X et Y
    l'utilisation du  lFlag = flaAxeMultiCurves
    permet de dire si on veut utiliser ces valeurs }
  vXMinAxe,vXMaxAxe,
  vYMinAxe,vYMaxAxe      : Float;

  fLimitsDefined   : boolean;
  { Read-Only }
  { indique si les limites ci-dessus sont dfinies }

  { Valeurs limites imposes aux Min et Max des axes en X et Y.
    l'utilisation du  lFlag = flaAxesWithLimitsValues
    permet de dire si on veut utiliser ces valeurs. }
  vXMinMin,vXMinMax,  { impose vXMinMin <= vXMinAxe <= vXMinMax }
  vXMaxMin,vXMaxMax,  { impose vXMaxMin <= vXMaxAxe <= vXMaxMax }
  vYMinMin,vYMinMax,  { impose vYMinMin <= vYMinAxe <= vYMinMax }
  vYMaxMin,vYMaxMax   { impose vYMaxMin <= vYMaxAxe <= vYMaxMax }
                   : Float;

  { nCotation-1 = nombre de tics mineurs entre les tics majeurs,
    Si = 0, cotation automatique pour que Tics mineurs = puissance de 10. }
  { Tics majeurs = graduation avec un Label }
  nXCotation        : integer;
  nYCotation        : integer;

  { pour augmenter les limites des axes en X et/ou Y }
  vXMinLimitIncrease,
  vXMaxLimitIncrease,
  vYMinLimitIncrease,
  vYMaxLimitIncrease  : Float;

  { Direction des Tics lors du dessin des axes, c.f. flaDrawTic... }
  nXDirTic,
  nYDirTic         : integer;

  { longueurs des Tics }
  mmXLenTicBig,
  mmXLenTicSml,
  mmYLenTicBig,
  mmYLenTicSml     : Float;

  { Valeurs calcules dans AutoAxes.
    On les mmorise ici, pour dessiner une grille. }
  vXTicBigMin,  { position du Tic majeur minimal }
  vXTicBigMax,  { position du Tic majeur maxmal }
  vXTicSmlMin,  { position du Tic mineur minimal }
  vXTicSmlMax,  { position du Tic mineur maximal }
  vXTicBigDelta,{ distance entre deux Tics majeurs }
  vXTicSmlDelta,{ distance entre deux Tics mineurs }
  vYIntercept   { position Y o l'axe X croise l'axe Y }
                   : Float;
  nXAxePos,     { position de l'axe X, -1=gauche, 0=en 0,  1=droite }
  nXDirTicUsed, { Direction utilise pour les labels sous les Tics }
  nXCotationUsed   : integer; {=nb Tic mineurs entre deux Tic Majeur, -1 }
  fNoXLabelOnFrame : boolean; { indique que l'on ne veut pas de label X
                                sur le cadre }

  vYTicBigMin,  { position du Tic majeur minimal }
  vYTicBigMax,  { position du Tic majeur maxmal }
  vYTicSmlMin,  { position du Tic mineur minimal }
  vYTicSmlMax,  { position du Tic mineur maximal }
  vYTicBigDelta,{ distance entre deux Tics majeurs }
  vYTicSmlDelta,{ distance entre deux Tics mineurs }
  vXIntercept   { position X o l'axe Y croise l'axe  X }
                   : Float;
  nYAxePos,     { position de l'axe Y, -1=en bas,  0=en 0,  1=en haut }
  nYDirTicUsed, { Direction utilise pour les labels sous les Tics }
  nYCotationUsed   : integer; {=nb Tic mineurs entre deux Tic Majeur, -1 }
  fNoYLabelOnFrame : boolean; { indique que l'on ne veut pas de label Y
                                sur le cadre }

  { Donnes pour les lignes du cadre }
  styleFrame       : TStyle;    {=PS_SOLID, PS_DASH, PS_DOT, PS_DASHDOT, c.f. CreatePen }
  pixWidthFrame    : integer;   { 0..infinity }
  mmWidthFrame     : Float;     { paisseur minimale en milimtres }
  colFrame         : TColorRef; { =RGB(0..255, 0..255, 0..255) }

  { - Donnes prives }
  { ----------------- }
{  Private}
  {"""""}

  { - Mthodes publiques }
  { ==================== }
  public
  {""""}

  { - constructeur et destructeur - }
  { ------------------------------- }
  { Init, en donnant le pointeur sur le graphisme de base associ }
  constructor Init(pGrBase1:PGraphicBase);
  destructor  Done;

  procedure ClearLimits; virtual;
  { Indique que on n'utilise pas vXMinAxe,vXMaxAxe,vYMinAxe,vYMaxAxe }

  procedure AdjustLimits(nbPts:integer; var avX,avY:array of Float); virtual;
  {=========================================================================}
  { Pour dfinir le min et max des axes en X et Y  partir d'une courbe avX,avY.
    Permet de dessiner plusieurs graphes sur le mme graphique:
    INPUT :
    nbPts, avX,avY = une courbe

    OUTPUT :
    Dpent de  fLimitsDefined.
    (Re)dfini les limites du graphique.

    Exemple d' UTILISATION :
    Pour 3 graph  dessiner sur le mme graphique
    ClearLimits;
    AdjustLimits(nbPts1, avXCurve1, avYCurve1);
    AdjustLimits(nbPts2, avXCurve2, avYCurve2);
    AdjustLimits(nbPts3, avXCurve3, avYCurve3);
    AutoAxes(nbPts, avXCurve, avYCurve, flaAxesMultiCurves [or Autres_flags]);
    nbPts, avXCurve et avYCurve ne seront pas pris en compte!
  -------------------------------------------------------------------------}

  procedure AutoScale(
  {==================}
        vMin,vMax    : Float;
        fBorderOnTic : boolean; { indique si les limites seront sur des Tics maj.}
        nCotation    : integer; { =0 si choix automatique }
    var vTicBigMin,  { position du Tic majeur minimal }
        vTicBigMax,  { position du Tic majeur maxmal }
        vTicSmlMin,  { position du Tic mineur minimal }
        vTicSmlMax,  { position du Tic mineur maximal }
        vTicBigDelta,{ distance entre deux Tics majeurs }
        vTicSmlDelta { distance entre deux Tics mineurs }
                     : Float;
    var nCotationUsed : integer; {=nb Tic mineurs entre deux Tic Majeur, -1 }
        nAxes        : integer; { indique si c'est pour l'axe X ou Y }
        lFlags       : longint  { c.f. flaAxes... }
    );
  { Dfini des chelles convenables pour un graphique.
    INPUT :
    vMin     : min de l'chelle <= vMin
    vMax     : max de l'chelle >= vMax
    OUTPUT :
    c.f. commentaires
    vTicBigMin est le plus grand multiple entier de vTicBigDelta <= vTicBigMin
    vTicBigMax est le plus petit multiple entier de vTicBigDelta >= vTicBigMax
    vTicSmlMin est le plus grand multiple entier de vTicSmlDelta <= vTicSmlMin
    vTicSmlMax est le plus petit multiple entier de vTicSmlDelta >= vTicSmlMax
    vTicBigDelta = vTicSmlDelta * nCotationUsed

    vTicBigDelta = x * 10^n, n entier, x = 0.5, 1, 2.5, ou 5,
         tel que il y ait entre 6 et 13 Tics munis d'un label. (=Tic majeurs)
  -------------------------------------------------------------------------}

  procedure DrawOneAxes(
    vIntercept, vTicBigMin, vMin1, vMax1, vTicBigMax, vTicSmlDelta : Float;
    vTicBigDelta:Float; nCotationUsed, nDirTicLoc : integer;
    lFlags:longint; { mmes flages que pour AutoAxes }
    nAxes:integer   { indique si c'est l'axe X ou Y que l'on dessine }
    );

  procedure DrawXAxes(lFlags:longint);
  procedure DrawYAxes(lFlags:longint);
  procedure DrawXYAxes(lFlags:longint);

  procedure AutoAxes(nbPts:integer; var avX,avY:array of Float; lFlags:longint); virtual;
  { calcul et dessine automatiquement des axes.
    C.f. les flags dfini en en-ttes pour des comportements particuliers.
    INPUT :
    nbPts  = nombre de points du graphe
    avX    = array[0..nbPts-1] of Float = coord. X des points du graphe
    avY    = array[0..nbPts-1] of Float = coord. Y des points du graphe

    lFlags = information de positionnement et d'affichage c.f. flaAxes de
             l'en-tte. Ils peuvent tre combin avec des  'or'
  -------------------------------------------------------------------------}

  procedure DrawGrid(
    nXBetween,nYBetween:integer;
    styleLineAA      : TStyle;    {=PS_SOLID, PS_DASH, PS_DOT, PS_DASHDOT, c.f. CreatePen }
    pixWidthLineAA   : integer;   { 1..infinity, paisseur minimale en pixels }
    mmWidthLineAA    : Float;     { paisseur minimale en milimtres }
    colLineAA        : TColorRef { =RGB(0..255, 0..255, 0..255) }
    );

  end; { - TAutoAxes = object - }

procedure FindMinMax(nbPts:integer; var avX:array of Float; var vXMin,vXMax:Float);
{=================================================================================}
{ Retourne le min et max de avX[0..nbPts-1] }

IMPLEMENTATION
{""""""""""""}

constructor TAutoAxes.Init(pGrBase1:PGraphicBase);
{==============================================}
begin { - TAutoAxes.Init - }
pGrBase:=pGrBase1;

fLimitsDefined:=FALSE;

{ choix automatique, pour que les Tics mineurs soient des puissances de 10 }
nXCotation:=0;
nYCotation:=0;

{ pour augmenter les limites des axes en X et/ou Y de 5% }
vXMinLimitIncrease :=0.05;
vXMaxLimitIncrease :=0.05;
vYMinLimitIncrease :=0.05;
vYMaxLimitIncrease :=0.05;

{ Direction des Tics lors du dessin des axes, c.f. flaDrawTic... }
nXDirTic  :=flaDrawTicAuto;
nYDirTic  :=flaDrawTicAuto;

{ longueurs des Tics }
mmXLenTicBig :=2;
mmXLenTicSml :=1;
mmYLenTicBig :=2;
mmYLenTicSml :=1;

{ Donnes pour les lignes du cadre }
styleFrame      :=PS_SOLID;
pixWidthFrame   :=1;
mmWidthFrame    :=0.2;
colFrame        :=RGB(0, 0, 0);

{ Ces limites ne doivent pas limiter l'affichage  priori. }
vXMinMin :=-1e35;
vXMinMax := 1e35;
vXMaxMin :=-1e35;
vXMaxMax := 1e35;
vYMinMin :=-1e35;
vYMinMax := 1e35;
vYMaxMin :=-1e35;
vYMaxMax := 1e35;
end;  { - TAutoAxes.Init - }

destructor  TAutoAxes.Done;
{========================}
begin { - TAutoAxes.Done - }
{ rien  faire }
end;  { - TAutoAxes.Done - }

procedure FindMinMax(nbPts:integer; var avX:array of Float; var vXMin,vXMax:Float);
{=================================================================================}
{ Retourne le min et max de avX[0..nbPts-1] }
var nn : integer;
begin { - FindMinMax - }
if (nbPts <= 0) then begin vXMin:=1e34; vXMax:=-1e34; exit; end; { pas dfini }

vXMin:=avX[0];   vXMax:=vXMin;
for nn:=1 to nbPts-1 do begin
   if      (vXMin > avX[nn]) then vXMin:=avX[nn]
   else if (vXMax < avX[nn]) then vXMax:=avX[nn];
   end; { for }
end;  { - FindMinMax - }

procedure TAutoAxes.ClearLimits;
{=============================}
{ Indique que on n'utilise pas vXMinAxe,vXMaxAxe,vYMinAxe,vYMaxAxe }
begin { - TAutoAxes.ClearLimits - }
fLimitsDefined:=FALSE;

{ Initialise, pour que si aucune courbe est utilise, avoir une graphique
  vide, mais standard par dfaut }
vXMinAxe:=-0.5;
vXMaxAxe:= 0.5;
vYMinAxe:=-0.5;
vYMaxAxe:= 0.5;
end;  { - TAutoAxes.ClearLimits - }

procedure TAutoAxes.AdjustLimits(nbPts:integer; var avX,avY:array of Float);
{=========================================================================}
{ Pour dfinir le min et max des axes en X et Y  partir d'une courbe avX,avY.
  Permet de dessiner plusieurs graphes sur le mme graphique:
  INPUT :
  nbPts, avX,avY = une courbe

  OUTPUT :
  Dpent de  fLimitsDefined.
  (Re)dfini les limites du graphique.

  Exemple d' UTILISATION :
  Pour 3 graph  dessiner sur le mme graphique
  ClearLimits;
  AdjustLimits(nbPts1, avXCurve1, avYCurve1);
  AdjustLimits(nbPts2, avXCurve2, avYCurve2);
  AdjustLimits(nbPts3, avXCurve3, avYCurve3);
  AutoAxes(nbPts, avXCurve, avYCurve, flaAxesMultiCurves [or Autres_flags]);
  nbPts, avXCurve et avYCurve ne seront pas pris en compte!
-------------------------------------------------------------------------}
var
  vXMinLoc,vXMaxLoc,
  vYMinLoc,vYMaxLoc  : Float;

begin { - TAutoAxes.AdjustLimits - }
FindMinMax(nbPts, avX, vXMinloc, vXMaxloc);
FindMinMax(nbPts, avY, vYMinloc, vYMaxloc);

if fLimitsDefined then begin
   if (vXMinAxe > vXMinLoc) then vXMinAxe:=vXMinLoc;
   if (vXMaxAxe < vXMaxLoc) then vXMaxAxe:=vXMaxLoc;
   if (vYMinAxe > vYMinLoc) then vYMinAxe:=vYMinLoc;
   if (vYMaxAxe < vYMaxLoc) then vYMaxAxe:=vYMaxLoc;
   end
else begin
   vXMinAxe:=vXMinLoc;
   vXMaxAxe:=vXMaxLoc;
   vYMinAxe:=vYMinLoc;
   vYMaxAxe:=vYMaxLoc;
   fLimitsDefined:=TRUE;
   end;

end;  { - TAutoAxes.AdjustLimits - }

function IntLog10(vX:Float):integer;
{==================================}
{ retourne nn tel que 10^nn <= abs(vX) < 10^(nn+1),
  c'est la partie entire du Log10(abs(vX)).  }
var nn : integer;
begin { - IntLog10 - }
vX:=abs(vX);
if (vX = 0) then begin IntLog10:=-36; exit; end;

nn:=0;
while (vX >= 10) do begin vX:=vX/10; inc(nn); end;
while (vX <   1) do begin vX:=vX*10; dec(nn); end;
IntLog10:=nn;
end;  { - IntLog10 - }

function Power10(nn:integer):Float;
{=================================}
{ retourne 10^nn  }
var vX : Float;
begin { - Power10 - }
vX:=1;
while (nn > 0) do begin vX:=vX*10; dec(nn); end;
while (nn < 0) do begin vX:=vX/10; inc(nn); end;
Power10:=vX;
end;  { - Power10 - }

procedure ComputeLabelSize(
{=========================}
      vDelta,                 { Distance entre les Tics majeurs }
      vMin,                   { min. de l'chelle }
      vMax         : Float;   { max. de l'chelle }
  var nTot,                   { nombre de places prisent  l'affichage }
      nVirg,                  { nombre de chiffres aprs virgule }
      nExp         : integer; { Si notation exponetielle ...E.. }
  var vMult        : Float    { =10^nExp }
  );
{ Calcule la taille des labels du graphique. }
var
  nDelta,          { =IntLog10(vDelta) }
  nMin,
  nMax             : integer;

begin { - ComputeLabelSize - }
nDelta :=IntLog10(vDelta);
nMin   :=IntLog10(vMin);
nMax   :=IntLog10(vMax);

if (nMax < nMin) then nMax:=nMin;

{ valeur habituelles par dfaut }
nExp:=0;
vMult:=1;

{ Calcule de la taille des labels pour un affichage optimis }
if (nDelta >= 0) then begin
   { pas de chiffres aprs la virgule }
   nVirg:=0;

   if (nMax <= 3) or (nDelta <= 2) then begin
      nTot:=1+nMax;
      if (vMin < 0) and (nMin >= nMax) then inc(nTot); { le signe prend aussi une place }
      exit;
      end;

   { impression avec exponentielle }
   nExp:=3 * (nDelta div 3);  { >= 1 }
   vMult:=Power10(-nExp);

   { -nExp car on divisera le nombre  afficher par vMult,
     +2  = place prise par  'e.'
     +IntLog10(nExp) = place prise si plusieurs chif. arps 'e' }
   nTot:=1+nMax - nExp + 2 + IntLog10(nExp);

   if (vMin < 0) and (nMin >= nMax) then inc(nTot); { le signe prend aussi une place }
   exit;
   end;

{ ici nDelta < 0 }
if (nMax >= -2) or (nDelta >= -4) then begin
   { Exemples : 12.34 ou 0.05 }
   nVirg:=-nDelta;
   if (nMax < 0) then nMax:=0;
   if (nMin < 0) then nMin:=0;
   nTot:=1+nMax + 1 + nVirg;
   if (vMin < 0) and (nMin >= nMax)  then inc(nTot); { le signe prend aussi une place }
   exit;
   end;

{ impression avec exponentielle }
nVirg:=0;
nExp:=-3 * ((2-nDelta) div 3);  { >= 1 }
vMult:=Power10(-nExp);

{ -nExp car on divisera le nombre  afficher par vMult,
  +3  = place prise par  'e-.'
  +IntLog10(-nExp) = place prise si plusieurs chif. arps 'e' }
nTot:=nMax - nExp + 3 + IntLog10(-nExp);

if (vMin < 0) and (nMin >= nMax)  then inc(nTot); { le signe prend aussi une place }
end;  { - ComputeLabelSize - }

function DetermineUnitFactor(
{===========================}
  pixWin2Delta,
  pixStrDelta      : integer;
  vUserDelta,
  vUnite           : Float;
  fBorderOnTic     : boolean
  ):integer;
{ Determine le facteur entier dans (1,2,5,10,20,50,100) qui multiplie Unite,
  pour que les Labels des Tics majeurs ne se recouvrent pas.
  Retourne 200 si pas de labels possible. }
var
  kk,nn            : integer;

begin { - DetermineUnitFactor - }
if (pixStrDelta >= pixWin2Delta) then begin
   DetermineUnitFactor:=200;
   exit;
   end;

if not fBorderOnTic then begin
   nn:=1 + Trunc(pixStrDelta * vUserDelta / (pixWin2Delta * vUnite));
   end
else begin
   { ici fBorderOnTic = TRUE }
   kk:=Trunc(pixWin2Delta / pixStrDelta); { >= 1 selon le teste du dbut }
   nn:=1 + Trunc(vUserDelta / (kk * vUnite));
   end;

     if (nn > 50) then nn:=100
else if (nn > 20) then nn:= 50
else if (nn > 10) then nn:= 20
else if (nn >  5) then nn:= 10
else if (nn >  2) then nn:=  5;

DetermineUnitFactor:=nn;
end;  { - DetermineUnitFactor - }

procedure TAutoAxes.AutoScale(
{===========================}
      vMin,vMax    : Float;
      fBorderOnTic : boolean; { indique si les limites seront sur des Tics maj.}
      nCotation    : integer; { =0 si choix automatique }
  var vTicBigMin,  { position du Tic majeur minimal }
      vTicBigMax,  { position du Tic majeur maxmal }
      vTicSmlMin,  { position du Tic mineur minimal }
      vTicSmlMax,  { position du Tic mineur maximal }
      vTicBigDelta,{ distance entre deux Tics majeurs }
      vTicSmlDelta { distance entre deux Tics mineurs }
                   : Float;
  var nCotationUsed : integer; {=nb Tic mineurs entre deux Tic Majeur, -1 }
      nAxes        : integer; { indique si c'est pour l'axe X ou Y }
      lFlags       : longint
  );
{ Dfini des chelles convenables pour un graphique.
  INPUT :
  vMin     : min de l'chelle <= vMin
  vMax     : max de l'chelle >= vMax
  OUTPUT :
  c.f. commentaires
  vTicBigMin est le plus grand multiple entier de vTicBigDelta <= vTicBigMin
  vTicBigMax est le plus petit multiple entier de vTicBigDelta >= vTicBigMax
  vTicSmlMin est le plus grand multiple entier de vTicSmlDelta <= vTicSmlMin
  vTicSmlMax est le plus petit multiple entier de vTicSmlDelta >= vTicSmlMax
  vTicBigDelta = vTicSmlDelta * nCotationUsed

  vTicBigDelta = x * 10^n, n entier, x = 0.5, 1, 2.5, ou 5,
       tel que il y ait entre 6 et 13 Tics munis d'un label. (=Tic majeurs)
-------------------------------------------------------------------------}
var
  vUnite           : Float; { distance entre deux Tic majeurs }
  Delta            : Float;
  DeltaS100        : Float;
  DeltaS10         : Float;
  nKeepTic         : integer; { nombre de Tics mineurs finalement gards }
//  vNbTicsLeft,
//  vNbTicsRight     : Float;
  nTot,nVirg,nExp  : integer;
  vMult            : Float;
  strS             : string[30];
  mmTotLen,
  mmTotHeight      : Float;
  pixWin2Delta,
  pixStrDelta      : integer;

   function mInt(v1:Float):Float;
   {============================}
   { --- mathematical def. of integral part of a real number --- }
   begin { - mint - }
   if (v1 >= 0) then mInt:=Int(v1 {+ vUnite/1000})    { vUnite/1000 pour arrondi }
   else              mInt:=Int(v1 {- vUnite/1000})-1; { vUnite/1000 pour arrondi }
   end;  { - mint - }

begin { - TAutoAxes.AutoScale - }
{==============================}
{ cas ou les min et max sont mal dfini }
if (vMin >= vMax) then begin vMin:=vMin-0.5;  vMax:=vMin + 1.0;  end;

{ --- recherche de vUnite = 10^n satisfaisant :
      vUnite * 10 <= Delta = vMax - vMin < vUnite * 100 --- }
Delta:=vMax - vMin;

vUnite:=1.0;
DeltaS100:=Delta / 100;
DeltaS10 :=Delta / 10;
while vUnite <= DeltaS100 do vUnite:=vUnite * 10.0;
while vUnite >  DeltaS10  do vUnite:=vUnite / 10.0;

{ Min et Max de l'chelle, en Units multiples de : 'Unite' }
//vNbTicsLeft  :=vMin / vUnite;  { peut tre ngatif et trs grand }
//vNbTicsRight :=vMax / vUnite;  { idem }

{ Calcule la taille des labels }
ComputeLabelSize(vUnite, vMin, vMax, nTot, nVirg, nExp, vMult);
if (nTot > 20) then nTot:=20; { impossible }
strS:='0';
while (length(strS) < nTot) do strS:=strS + '0';

{ Donnes pour dterminer l'espacement entre les tics majeurs }
pGrBase^.GetMMTextSize(strS, mmTotLen, mmTotHeight);

if (nAxes = flaAxesX) then begin
   pixStrDelta:=Round(mmTotLen * pGrBase^.vPixXSmm);
   pixWin2Delta:=pGrBase^.pixX2Width;
   end
else begin
   pixStrDelta:=Round(mmTotHeight * pGrBase^.vPixYSmm);
   pixWin2Delta:=pGrBase^.pixY2Height;
   end;

{ Dtermine l'espacement entre les tics majeurs }
nKeepTic:=DetermineUnitFactor( pixWin2Delta, pixStrDelta,
                               vMax - vMin, vUnite, fBorderOnTic);

{ distance entre deux Tics majeurs }
vTicBigDelta := vUnite * nKeepTic;

{ distance entre deux Tics mineurs }
if (nCotation = 0) then begin { choix automatique des tics mineurs }
   if (nKeepTic = 1) then nCotationUsed:=10
   else                   nCotationUsed:=nkeepTic;
   end
else                      nCotationUsed:=nCotation;

{ distance entre deux Tics mineurs }
vTicSmlDelta := vTicBigDelta/nCotationUsed;

vTicBigMin:= vTicBigDelta*mInt( vMin/vTicBigDelta); { position du Tic majeur minimal }
vTicBigMax:=-vTicBigDelta*mInt(-vMax/vTicBigDelta);{ position du Tic majeur maxmal }
{ position du Tic mineur minimal }
vTicSmlMin:= vTicSmlDelta*mInt( vMin/vTicSmlDelta);
{ position du Tic mineur maxmal }
vTicSmlMax:=-vTicSmlDelta*mInt(-vMax/vTicSmlDelta);
end;  { - TAutoAxes.AutoScale - }

function FloatToStr(vX:Float; nTot,nVirg,nExp:integer):String;
{============================================================}
{ procedure de conversion de Float en string, avec formattage, pour
  impression.
  nTot  = formattage pour que la virgule soit  la place nTot - nVirg -1
          Si pas de virgule, c'est l'exponentielle qui est align.
           (=0 pas de formattage)
  nVirg = nombre de chiffres aprs la virgule
  nExp  = nombre  placer aprs le  'e' pour puissances de 10
          (0= => n'est pas affich).
-------------------------------------------------------------------------}
var
  nSize            : integer;
  vMult,
  vFact,
  vDiv             : Float;
  strSigne         : string[1];
  strS3,
  strS2,
  strS             : string[30];

begin { - FloatToStr - }
if (vX = 0) then begin FloatToStr:='0'; exit; end;

if (abs(vX) > 1000000000) then begin
   { affichage avec exponentielles }
   { traitement  complter }
   nSize:=3 * ((5+IntLog10(vX)) div 3);
   vDiv:=Power10(nSize);
   vX:=vX/vDiv;
   nExp:=nExp+nSize;
   nVirg:=0;
   end;

strS:='';
strSigne:='';
{ place le signe si ngatif }
if (vX < 0) then begin
   strSigne:='-';
   vX:=-vX;
   end;

{ pour les arrondis }
vMult:=Power10(-nVirg);
vX:=vX + 0.5*vMult;

{ place la partie avant la virgule }
strS:=strS+IntToStr(Trunc(vX),0);

{ partie aprs la virgule, si existe }
if (nVirg > 0) then begin
   { nombre aprs la virgule }
   vMult:=Power10(nVirg);
   vFact:=Frac(vX)*vMult;

   { vite des erreurs de Trunc }
   strS2:='';
   while (vFact > 1e9) do begin
      strS2:=IntToStr(Trunc(vX*10), 0) + strS2;
      vX:=Frac(vX*10);
      vMult:=vMult/10;
      vFact:=Frac(vX)*vMult;
      end;

{   while (vFact > 1e9) do begin vFact:=vFact/10; strS2:=strS2 + '9'; end;}
   strS2:=strS2 + IntToStr(Trunc(vFact) ,0);

   strS3:=strS2;
   while (length(strS2) < nVirg) do strS2:='0' + strS2;

   { le '0' est particulier }
   if (strS <> '0') or (strS3 <> '0') then strS:=strS + '.' + strS2;
   end;

{ partie aprs l'exponentielle si existe }
if (nExp <> 0) and(strS <> '0') then strS:=strS + 'e' + IntToStr(nExp,0);

if (strS <> '0') then strS:=strSigne + strS;

{ Formattage pour avoir une longueur fixe, ne sert  rien en mode proportionnelle }
{while (length(strS) < nTot) do strS:='' + strS;}
FloatToStr:=strS;
end;  { - FloatToStr - }

procedure TAutoAxes.DrawOneAxes(
{=============================}
  vIntercept, vTicBigMin, vMin1, vMax1, vTicBigMax, vTicSmlDelta : Float;
  vTicBigDelta:Float; nCotationUsed, nDirTicLoc : integer; lFlags:longint;
  nAxes:integer { indique si c'est l'axe X ou Y que l'on dessine }
  );
var
  vPosMin,vPosMax,
  vPos             : Float;
  mmXLenTicBig0,
  mmYLenTicBig0,
  mmXLenTicSml0,
  mmYLenTicSml0    : Float;
  relX,relY        : Float;
  nTot,                   { nombre de places prisent  l'affichage }
  nVirg,                  { nombre de chiffres aprs virgule }
  nExp             : integer; { Si notation exponetielle ...E.. }
  vMult            : Float;    { =10^nExp }
  nCpt
                   : integer;
  fDraw00,
  fWithTic,
  fWithLabel       : boolean;
  strS             : string[20];
  vX05mm,vY05mm,v05mm,   { pour le cas ou on est proche des bords }
  mmTotLen,
  mmTotHeight      : Float;
  nAxePos,
  pixXLabelWidth,
  pixYLabelHeight  : integer;
  fDraw            : boolean;

begin { - TAutoAxes.DrawOneAxes - }
ComputeLabelSize(vTicBigDelta, vMin1, vMax1, nTot, nVirg, nExp, vMult);

{ Test si les axes ne se croisent pas en (0,0), et donc on crit le label 0 }
fDraw00:=(vXIntercept <> 0.0 ) or (vYIntercept <> 0.0);

pGrBase^.MMToUser(0.5{mm},0.5{mm}, vX05mm,vY05mm);
if (nAxes = flaAxesX) then begin
   { dessin de l'axe X }
   pGrBase^.Line0(vMin1, vIntercept, vMax1, vIntercept);

   if (vXTicBigDelta >= vXMaxAxe - vXMinAxe) then
      lFlags:=lFlags or flaAxesNoTicX;

   fWithTic   :=((lFlags and flaAxesNoTicX) = 0);
   fWithLabel :=((lFlags and flaAxesWithoutXLabel) = 0);

   { sens de la graduation }
   mmXLenTicBig0:=0;
   mmYLenTicBig0:=mmYLenTicBig;
   mmXLenTicSml0:=0;
   mmYLenTicSml0:=mmYLenTicSml;

   relX:=0.5;
   relY:=0;

   if (nDirTicLoc = flaDrawTicUp) then begin
      relY:=1;
      mmYLenTicBig0:=-mmYLenTicBig;
      mmYLenTicSml0:=-mmYLenTicSml;
      end;

   mmTotLen:=0;
   v05MM:=vX05mm;

   nAxePos:=nXAxePos;

   { Cas particuliers rares }
   if (vYIntercept = vYMinAxe) then nAxePos:=-1; { haut }
   if (vYIntercept = vYMaxAxe) then nAxePos:= 1; { bas}

   fDraw00:=fDraw00 or (nXAxePos <> 0);
   end
else begin
   { dessin de l'axe Y }
   pGrBase^.Line0(vIntercept, vMin1, vIntercept, vMax1);

   if (vYTicBigDelta >= vYMaxAxe - vYMinAxe) then
      lFlags:=lFlags or flaAxesNoTicY;

   fWithTic   :=((lFlags and flaAxesNoTicY) = 0);
   fWithLabel :=((lFlags and flaAxesWithoutYLabel) = 0);

   { sens de la graduation }
   mmXLenTicBig0:=-mmXLenTicBig;
   mmYLenTicBig0:=0;
   mmXLenTicSml0:=-mmXLenTicSml;
   mmYLenTicSml0:=0;

   relX:=1;
   relY:=0.5;
   mmTotLen:=0;

   if (nDirTicLoc = flaDrawTicRight) then begin
      relX:=1;
      mmXLenTicBig0:=mmXLenTicBig;
      mmXLenTicSml0:=mmXLenTicSml;

      { calcule du dcalage pour alignement.
        On suppose que tous les chiffres prennent la mme place. }
      strS:='0';
      while (length(strS) < nTot) do strS:=strS + '0';

      pGrBase^.GetMMTextSize(strS, mmTotLen, mmTotHeight);
      end;

   v05MM:=vY05mm;
   nAxePos:=nYAxePos;

   { Cas particuliers rares }
   if (vXIntercept = vXMinAxe) then nAxePos:=-1; {  gauche }
   if (vXIntercept = vXMaxAxe) then nAxePos:= 1; {  droite }

   fDraw00:=fDraw00 or (nYAxePos <> 0);
   end;

if not fWithTic then exit;

{if (nAxes = flaAxesX) then nTot:=0; // pas de formattage des labels }
nTot:=0;

with pGrBase^ do begin
   { pour viter de perdre les Tics extrmes  cause d'erreurs d'arrondi }
   vPosMin :=vMin1 - v05mm;
   vPosMax :=vMax1 + v05mm;
   { ancienne manire, moins bonne
   vPosMin :=vMin1 - vTicSmlDelta / 10;
   vPosMax :=vMax1 + vTicSmlDelta / 10;}

   { dessin des Tics }
   vPos    :=vTicBigMin;
   nCpt    :=0;

   { Saute la partie non visible du dbut }
   while (vPos < vPosMin) do begin
       vPos:=vPos + vTicSmlDelta;
       inc(nCpt); if (nCpt > 1000) then exit; { erreur!}
       end;

   { Dessin des Tics et des labels }
   while (vPos < vPosMax) do begin
      if (nAxes = flaAxesX) then MoveTo0(vPos, vIntercept)
      else                       MoveTo0(vIntercept, vPos);

      if ((nCpt mod nCotationUsed) = 0) then begin
         if (nDirTicLoc = flaDrawTicCross) then
             MoveMMRelTo(-mmXLenTicBig0/2, -mmYLenTicBig0/2);
         LineMMRelTo(mmXLenTicBig0, mmYLenTicBig0);

         { Si avec labels et label n'est pas en 0 avec axes qui se croisent }
         if fWithLabel and
            (fDraw00 or (vPos < -vTicSmlDelta/2) or(vPos > vTicSmlDelta/2)) then begin
            { - Met le Label - }
            MoveMMRelTo(mmTotLen, 0);

            strS:=FloatToStr(vPos*vMult, nTot, nVirg, nExp);
            GetPixTextSize(strS, pixXLabelWidth, pixYLabelHeight);

            { n'affiche pas le texte si touche le cadre du haut }
            fDraw:=TRUE;

            { pas de label si demand et touche le cadre }
            if (nAxes = flaAxesY) and (nAxePos = 0) and
               fNoYLabelOnFrame and
               ((pixYPos + relY *pixYLabelHeight <= pixY2Top) or
                (pixYPos - relY *pixYLabelHeight >= pixY2Top+ pixY2Height)) then
               fDraw:=FALSE;

            { pas de label si demand et touche le cadre }
            if (nAxes = flaAxesX) and (nAxePos = 0) and
               fNoXLabelOnFrame and
               ((pixXPos + relX *pixXLabelWidth <= pixX2Left) or
                (pixXPos - relX *pixXLabelWidth >= pixX2Left+ pixX2Width)) then
               fDraw:=FALSE;

            if fDraw then writeMoveRel(relX, relY, strS);
            end;
         end
      else begin
         if (nDirTicLoc = flaDrawTicCross) then
             MoveMMRelTo(-mmXLenTicSml0/2, -mmYLenTicSml0/2);
         LineMMRelTo(mmXLenTicSml0, mmYLenTicSml0);
         end;

      vPos:=vPos + vTicSmlDelta;
      inc(nCpt);
      if (nCpt > 3000) then exit; { erreur!}
      end; { - while (vPos < vPosMax) do begin - }

   end; { - with pGrBase^ do begin - }
end;  { - TAutoAxes.DrawOneAxes - }

procedure TAutoAxes.DrawXAxes(lFlags:longint);
{===========================================}
begin { - TAutoAxes.DrawXAxes - }
DrawOneAxes( vYIntercept, vXTicBigMin, vXMinAxe, vXMaxAxe, vXTicBigMax,
             vXTicSmlDelta, vXTicBigDelta,
             nXCotationUsed, nXDirTicUsed, lFlags, flaAxesX);
end;  { - TAutoAxes.DrawXAxes - }

procedure TAutoAxes.DrawYAxes(lFlags:longint);
{===========================================}
begin { - TAutoAxes.DrawYAxes - }
DrawOneAxes( vXIntercept, vYTicBigMin, vYMinAxe, vYMaxAxe, vYTicBigMax,
             vYTicSmlDelta, vYTicBigDelta,
             nYCotationUsed, nYDirTicUsed, lFlags, flaAxesY);
end;  { - TAutoAxes.DrawYAxes - }

procedure TAutoAxes.DrawXYAxes(lFlags:longint);
{===========================================}
begin { - TAutoAxes.DrawXYAxes - }
DrawXAxes(lFlags);
DrawYAxes(lFlags);
end;  { - TAutoAxes.DrawXYAxes - }

procedure TAutoAxes.AutoAxes(nbPts:integer; var avX,avY:array of Float; lFlags:longint);
{=====================================================================================}
{ calcul et dessine automatiquement des axes.
  C.f. les flags dfini en en-ttes pour des comportements particuliers.
  INPUT :
  nbPts  = nombre de points du graphe
  avX    = array[0..nbPts-1] of Float = coord. X des points du graphe
  avY    = array[0..nbPts-1] of Float = coord. Y des points du graphe

  lFlags = information de positionnement et d'affichage c.f. const. de l'en-tte.
           Ils peuvent tre combin avec des  'or'
-------------------------------------------------------------------------}
var
  fBorderXOnTic : boolean;
  fBorderYOnTic : boolean;
  vTemp         : Float;

begin { - TAutoAxes.AutoAxes - }
{ --- recherche des limites inf. et sup. en X et Y des axes --- }
if ((lFlags and flaAxesMultiCurves) = 0) then begin
   ClearLimits;
   { Dtermination des limites du graphique }

   AdjustLimits(nbPts, avX, avY);
   end;

{ --- cas particuliers pour la compatibilit avec Quinn-Curtis v7.x --- }
     if (lFlags = 0) then lFlags:=flaAxesInterceptX0 or flaAxesInterceptY0
else if (lFlags = 2) then lFlags:=flaAxesInterceptXUp or flaAxesInterceptYRight
else if (lFlags = 3) then lFlags:=flaAxesInterceptXUp;

fBorderXOnTic:=(lFlags and flaAxesBorderXOnTic) <> 0;
fBorderYOnTic:=(lFlags and flaAxesBorderYOnTic) <> 0;

{ --- si on veut d'agrandir l'axe X --- }
if ((lFlags and flaAxesNotSmallX) <> 0) then begin
   vXMinAxe:=vXMinAxe - (vXMaxAxe - vXMinAxe) * vXMinLimitIncrease;
   vXMaxAxe:=vXMaxAxe + (vXMaxAxe - vXMinAxe) * vXMaxLimitIncrease;
   end;

{ --- si on veut d'agrandir l'axe Y --- }
if ((lFlags and flaAxesNotSmallY) <> 0) then begin
   vYMinAxe:=vYMinAxe - (vYMaxAxe - vYMinAxe) * vYMinLimitIncrease;
   vYMaxAxe:=vYMaxAxe + (vYMaxAxe - vYMinAxe) * vYMaxLimitIncrease;
   end;

{ --- si on veut imposer des limites aux dimensions des axes --- }
if ((lFlags and flaAxesWithLimitsValuesX) <> 0) then begin
   if (vXMinAxe < vXMinMin) then vXMinAxe:=vXMinMin;
   if (vXMinAxe > vXMinMax) then vXMinAxe:=vXMinMax;
   if (vXMaxAxe < vXMaxMin) then vXMaxAxe:=vXMaxMin;
   if (vXMaxAxe > vXMaxMax) then vXMaxAxe:=vXMaxMax;
   end;

if ((lFlags and flaAxesWithLimitsValuesY) <> 0) then begin
   if (vYMinAxe < vYMinMin) then vYMinAxe:=vYMinMin;
   if (vYMinAxe > vYMinMax) then vYMinAxe:=vYMinMax;
   if (vYMaxAxe < vYMaxMin) then vYMaxAxe:=vYMaxMin;
   if (vYMaxAxe > vYMaxMax) then vYMaxAxe:=vYMaxMax;
   end;

{ --- si on veut l'axe X Symtrique --- }
if ((lFlags and flaAxesSymetricXAxes) <> 0) and
   ((vXMaxAxe - vXMinAxe) * 5 > abs(vXMinAxe) + abs(vXMaxAxe)) then begin
   { Test que cela ait un sens, par exemple que vXMinAxe < 0 < vXMaxAxe }
   if (vXMaxAxe < abs(vXMinAxe)) then vXMaxAxe:=abs(vXMinAxe);
   vXMinAxe:=-vXMaxAxe;
   end;

{ --- si on veut l'axe Y Symtrique --- }
if ((lFlags and flaAxesSymetricYAxes) <> 0) and
   ((vYMaxAxe - vYMinAxe) * 5 > abs(vYMinAxe) + abs(vYMaxAxe)) then begin
   { Test que cela ait un sens, par exemple que vYMinAxe < 0 < vYMaxAxe }
   if (vYMaxAxe < abs(vYMinAxe)) then vYMaxAxe:=abs(vYMinAxe);
   vYMinAxe:=-vYMaxAxe;
   end;

{ --- si on veut des axes identiques --- }
if ((lFlags and flaAxesSameXY) > 0) then begin
   if (vXMinAxe > vYMinAxe) then vXMinAxe:=vYMinAxe
   else                          vYMinAxe:=vXMinAxe;
   if (vXMaxAxe < vYMaxAxe) then vXMaxAxe:=vYMaxAxe
   else                          vYMaxAxe:=vXMaxAxe;
   end;

{ cas ou les min et max sont mal dfini }
if (vXMinAxe >= vXMaxAxe) then begin vXMinAxe:=-0.5;  vXMaxAxe:=vXMinAxe + 1.0;  end;
{if (vYMinAxe >= vYMaxAxe) then begin vYMinAxe:=-0.5;  vYMaxAxe:=vYMinAxe + 1.0;  end;}
if (vYMaxAxe - vYMinAxe <= (abs(vYMinAxe) + abs(vYMaxAxe)) / 10000) then begin
   if (abs(vYMinAxe) > 1e6)  or (abs(vYMaxAxe) > 1e6) or
      (abs(vYMinAxe) < 1e-6) or (abs(vYMaxAxe) < 1e-6) then begin
      { probablement l'axe n'est pas dfini }
      vYMinAxe:=-0.5; vYMaxAxe:=0.5;
      end
   else begin
      vTemp    :=(vYMaxAxe + vYMinAxe) / 2 -  (abs(vYMaxAxe) + abs(vYMinAxe))/1000;
      vYMaxAxe :=(vYMaxAxe + vYMinAxe) / 2 +  (abs(vYMaxAxe) + abs(vYMinAxe))/1000;
      vYMinAxe :=vTemp;
      end;
   end;

{ Calcul de l'chelle en X }
AutoScale( vXMinAxe,vXMaxAxe,fBorderXOnTic, nXCotation,
           vXTicBigMin, vXTicBigMax, vXTicSmlMin, vXTicSmlMax,
           vXTicBigDelta, vXTicSmlDelta, nXCotationUsed, flaAxesX, lFlags);

{ Calcul de l'chelle en Y }
AutoScale( vYMinAxe,vYMaxAxe,fBorderYOnTic, nYCotation,
           vYTicBigMin, vYTicBigMax, vYTicSmlMin, vYTicSmlMax,
           vYTicBigDelta, vYTicSmlDelta, nYCotationUsed, flaAxesY,  lFlags);

{DebugWrite(1, IntToStr(Round(vXMinAxe),9) + IntToStr(Round(vXMaxAxe),9));
DebugWrite(1, IntToStr(Round(vXTicBigMin),9) + IntToStr(Round(vXTicBigMax),9));
DebugWrite(1, IntToStr(Round(vXTicSmlMin),9) + IntToStr(Round(vXTicSmlMax),9));

DebugWrite(1, IntToStr(Round(vYMinAxe),9) + IntToStr(Round(vYMaxAxe),9));
DebugWrite(1, IntToStr(Round(vYTicBigMin),9) + IntToStr(Round(vYTicBigMax),9));
DebugWrite(1, IntToStr(Round(vYTicSmlMin),9) + IntToStr(Round(vYTicSmlMax),9));}

{ --- sauf demand, les bords du graphique ne coincident pas forcment
      avec des tics --- }
if fBorderXOnTic then begin
   vXMinAxe:=vXTicBigMin;
   vXMaxAxe:=vXTicBigMax;
   end;

if fBorderYOnTic then begin
   vYMinAxe:=vYTicBigMin;
   vYMaxAxe:=vYTicBigMax;
   end;

{ --- si on veut des axes identiques --- }
if ((lFlags and flaAxesSameXY) > 0) then begin
   if (vXMinAxe > vYMinAxe) then vXMinAxe:=vYMinAxe
   else                          vYMinAxe:=vXMinAxe;
   if (vXMaxAxe < vYMaxAxe) then vXMaxAxe:=vYMaxAxe
   else                          vYMaxAxe:=vXMaxAxe;
   end;

{ limites des coordonnes utilisateur. }
with pGrBase^ do begin
   useXMin :=vXMinAxe;
   useYMin :=vYMinAxe;
   useXMax :=vXMaxAxe;
   useYMax :=vYMaxAxe;
   UpdateWin12Pos;
   end;

{ --- localisation du croisement des axes --- }
vXIntercept:=vXMinAxe;
vYIntercept:=vYMinAxe;
nXAxePos   :=-1; { axe X est  gauche }
nYAxePos   :=-1; { axe Y est en bas }
nXDirTicUsed:=nXDirTic;
if (nXDirTic = 0) then nXDirTicUsed:=flaDrawTicDown; { choix automatique }
nYDirTicUsed:=nYDirTic;
if (nYDirTic = 0) then nYDirTicUsed:=flaDrawTicLeft; { choix automatique }

{ --- X --- }
if ((lFlags and flaAxesInterceptXUp) > 0) then begin
   nXAxePos   := 1; { axe X est en haut }
   vYIntercept:=vYMaxAxe;
   if (nXDirTic = 0) then nXDirTicUsed:=flaDrawTicUp; { tics vers le haut }
   end;

{ --- Y --- }
if ((lFlags and flaAxesInterceptYRight) > 0) then begin
   nYAxePos   := 1; { axe Y est  droite }
   vXIntercept:=vXMaxAxe;
   if (nYDirTic = 0) then nYDirTicUsed:=flaDrawTicRight; { tics vers la droite }
   end;

{ --- X --- }
if ((lFlags and flaAxesInterceptX0) > 0) and
   (vYMinAxe <= 0.0) and (vYMaxAxe >= 0.0) then begin
   { l'axe X croise l'axe Y en Y=0 }
   nXAxePos   := 0; { axe X est en Y=0 }
   vYIntercept:=0.0;
   if (nXDirTic = 0) then nXDirTicUsed:=flaDrawTicCross; { tics croisent l'axe }
   end;

{ --- Y --- }
if ((lFlags and flaAxesInterceptY0) > 0) and
   { l'axe Y croise l'axe X en X=0 }
   (vXMinAxe <= 0.0) and (vXMaxAxe >= 0.0) then begin
   nYAxePos   := 0; { axe Y est en X=0 }
   vXIntercept:=0.0;
   if (nYDirTic = 0) then nYDirTicUsed:=flaDrawTicCross; { tics croisent l'axe }
   end;

fNoXLabelOnFrame:=((lFlags and (flaAxesNoXLabelOnFrame or flaAxesWithFrame)) <> 0);
fNoYLabelOnFrame:=((lFlags and (flaAxesNoYLabelOnFrame or flaAxesWithFrame)) <> 0);

if ((lFlags and flaAxesNoDraw) <> 0) then exit;

{ --- Dessins du cadre et des axes --- }
if ((lFlags and flaAxesWithFrame) > 0) then
   pGrBase^.BorderWin2(styleFrame, pixWidthFrame, mmWidthFrame, colFrame);

{ --- dessin des axes --- }
{ --- X --- }
DrawXAxes(lFlags);

{ --- Y --- }
DrawYAxes(lFlags);
end;  { - TAutoAxes.AutoAxes - }

procedure TAutoAxes.DrawGrid(
{==========================}
  nXBetween,nYBetween:integer;
  styleLineAA      : TStyle;    {=PS_SOLID, PS_DASH, PS_DOT, PS_DASHDOT, c.f. CreatePen }
  pixWidthLineAA   : integer;   { 1..infinity, paisseur minimale en pixels }
  mmWidthLineAA    : Float;     { paisseur minimale en milimtres }
  colLineAA        : TColorRef  { =RGB(0..255, 0..255, 0..255) }
  );
{ Dessin d'une Grille, avec le style, l'paisseur et la couleur de ligne
  demande. Pour des impressions sur imprimantes, il peut tre utile
  d'exprimer une paisseur minimale de ligne en milimtres. C'est la raison
  pour laquelle l'paisseur est donnes en pixels et en milimtres,
  chacune des deux donnes est une valeur minimale d'paisseur.
-------------------------------------------------------------------------}
var
  pixWidth         : integer;  { le style n'est valable que avec une
                                 paisseur de 1 pixel }
  fDoReleaseDC     : boolean;
  vX05mm,vY05mm    : Float;  { corresp. de 0.5 mm en coord. utilisateur }
  recLineAt        : TypeLineAttrib;

   procedure DrawOneGrid(
   {====================}
      vTicBigMin, vMin1, vMax1, vTicDelta : Float;
      vMin2,vMax2,v05mm : Float; { min et max de l'autre axe }
      nAxes:integer { indique si c'est l'axe X ou Y que l'on dessine }
      );
   var
     vPosMin,vPosMax,
     vPos             : Float;
     vPixX1,vPixY1,
     vPixX2,vPixY2    : Float;
     nn               : integer;

   begin { - DrawOneGrid - }
   { pour viter de perdre les Tics extrmes  cause d'erreurs d'arrondi }
   vPosMin :=vMin1 - v05mm;
   vPosMax :=vMax1 + v05mm;

   { dessin des Tics }
   vPos    :=vTicBigMin;
   { Saute la partie non visible du dbut }
   while (vPos < vPosMin) do vPos:=vPos + vTicDelta;

   { Dessin de la grille perpendiculaire  l'axe choisi }
   with pGrBase^do while (vPos < vPosMax) do begin
      if (nAxes = flaAxesX) then begin
         UserToPix(vPos, vMin2, vPixX1, vPixY1);
         UserToPix(vPos, vMax2, vPixX2, vPixY2);
         vPixX1:=vPixX1-pixWidth div 2;

         { trace plusieurs lignes, pour donner une paisseur }
         for nn:=0 to pixWidth-1 do begin
            MovePixTo(vPixX1+nn, vPixY1);
            LinePixTo(vPixX1+nn, vPixY2);
            end;
         end
      else begin
         UserToPix(vMin2, vPos, vPixX1, vPixY1);
         UserToPix(vMax2, vPos, vPixX2, vPixY2);
         vPixY1:=vPixY1-pixWidth div 2;

         { trace plusieurs lignes, pour donner une paisseur }
         for nn:=0 to pixWidth-1 do begin
            MovePixTo(vPixX1, vPixY1+nn);
            LinePixTo(vPixX2, vPixY1+nn);
            end;
         end; { - else de if (nAxes = flaAxesX) then begin - }

      vPos:=vPos + vTicDelta;
      end; { - with pGrBase^do while (vPos < vPosMax) do begin - }
   end; { - DrawOneGrid - }

begin { - TAutoAxes.DrawGrid - }
{==============================}
with pGrBase^ do begin
   { calcule ici la largeur en pixels, car le style de ligne n'est
     valable que avec une largeur de 1 pixel }
   pixWidth:=Trunc(mmWidthLineAA * vPixXSmm);
   if (pixWidth < pixWidthLineAA) then pixWidth:=pixWidthLineAA;

   { mmorise le style de ligne actuelle }
   GetLineAttribRec(recLineAt);

   { change le style de ligne pour le dessin.
     1 pixel de largeur, sinon le style ne marche pas! }
   SetLineAttrib(styleLineAA, 1{pixWidthLineAA}, 0{mmWidthLineAA}, colLineAA);

   { Si ncssaire, ouvre le Device contexte, pour aller plus vite }
   fDoReleaseDC:=(hDC0 = 0);
   if (hDC0 = 0) then begin
      if (GetDC0 = 0) then begin {error} exit; end;
      end;

   MMToUser(0.5{mm},0.5{mm}, vX05mm,vY05mm);

   if (nXBetween <=  0) then nXBetween:= 1;
   if (nXBetween >= 21) then nXBetween:=20;
   if (nYBetween <=  0) then nYBetween:= 1;
   if (nYBetween >= 21) then nYBetween:=20;

   { Dessin de la grille verticale }
   DrawOneGrid( vXTicBigMin, vXMinAxe,vXMaxAxe, vXTicBigDelta/nXBetween,
                vYMinAxe,vYMaxAxe,vX05mm,flaAxesX);

   { Dessin de la grille horizontale }
   DrawOneGrid( vYTicBigMin, vYMinAxe,vYMaxAxe, vYTicBigDelta/nYBetween,
                vXMinAxe,vXMaxAxe,vY05mm,flaAxesY);

   if fDoReleaseDC then ReleaseDC0;

   { remet l'ancien style de ligne }
   SetLineAttribRec(recLineAt);
   end; { - with pGrBase^ do begin - }
end;  { - TAutoAxes.DrawGrid - }


{ *** UNIT wAutoAxes0 *** }
end.

