unit wFontManager;

{ Unit pour le traitement de Fonts.
  Cration de fonts,
  Slection de fonts dans un DC
  Efface un font.

  rem.  c.f. Fonts in Windows.
    72  points = 1 inch
  1440  twips  = 1 inch
    20  twips  = 1 point
-------------------------------------------------------------------------}

// To tell FreePascal that the application is a Graphical User Interface
// i.e. is running in windows, not in text mode.
{$APPTYPE GUI}

interface
{"""""""}

Uses
//  wWindow0
  wErrors0
// ,wWindowBase0
 ,wGlobals2
 ,wDebug0
 ,Strings
 ,Windows;

function SelectFont(hDC1:HDC; var LogFont:TLogFont):HFont;
{========================================================}
{ Est souvant utilis
  1) Cherche une font correspondant  la description LogFont.
  2) Si pas trouv, cr la font correspondante
  3) Slectionne la font dans le DC
  4) Mmorise le hDC, et la font qui s'y trouvait, si c'est le premier select
  5) Incrmente le count lock de 100, et
     si une font tait dj slectionn dans le DC,
     dcrment son count lock de 100.
  *) Retourne la font prcedemment slectionn.
-------------------------------------------------------------------------}

function ReleaseFont(hDC1:HDC):HFont;
{===================================}
{ Est souvant utilis
  remet l'ancien font dans ce DC, si il est dans la liste, et
  retourne le dernier font qui tait dans ce DC.
  Dcrmente le cont lock de 100.
-------------------------------------------------------------------------}

function GetFont(var LogFont:TLogFont):HFont;
{===========================================}
{ Est parfois utilis.
  1) Cherche une font correspondant  la description LogFont.
  2) Si pas trouv, cr la font correspondante.
     Retourne le font.
     Retourne 0 en cas d'erreur.
-------------------------------------------------------------------------}

function IsHFontRegistered(hFnt:HFont):boolean;
{=============================================}
{ Indique si la Font dfini par le handle hFnt est dans la liste des
  fonts gr par cette unit.
  Si oui, elle sera dtruite automatiquement en fin de programme,
   l'appelle de DoneFontUtilities, et donc il ne faut
  pas la dtruire autrement. }

function LockFont(hFnt:HFont):integer;
{====================================}
{ Incrmente de 1 le count lock, et retourne sa valeur.
  Retourne -1111 si pas trouv.
-------------------------------------------------------------------------}

function UnLockFont(hFnt:HFont):integer;
{======================================}
{ Dcrmente de 1 le count lock, et retourne sa valeur.
  Retourne -1111 si pas trouv.
-------------------------------------------------------------------------}

function DeleteFont(var LogFont:TLogFont):integer;
{================================================}
{ Rarement utilis.
  A utiliser particulirement quand on veut remplacer une font par une
  autre, et qu'il n'y a plus de raison de garder l'ancienne.
  Efface la font nFont de la liste et de la mmoire.
  Retourne 0 si OK
           errFontBadNum si nFont <= 0 ou nFont > nLastFont
           errFontBadDeleteObject si erreur au DeleteObject(ahFont[nFont])
           errFontLocked si Locked, i.e. si Count Lock > 0,
              n'efface pas, ce qui fait perdre de la mmoire ressources.
-------------------------------------------------------------------------}

function TextOutWithSymbol(
{==========================}
  var fontText     : TLogFont; { transmis par var pour simplicit, mais pas modifi }
      pixXP0,      { position X d'affichage }
      pixYP0       { position Y d'affichage }
                   : integer;
  var strS         : String;  { texte  afficher }
      ptSizeText   : integer;
      hDC0         : HDC;
      wFlags       : word
      ):longint;
{ Affichage de texte, avec caractres grec, indices et exposant.
  \g(abc) => abc en grec.
  \+      => suite en exposant, ou retour aprs un \-
  \-      => suite en indice, ou retour aprs un \+
  wFlags <- 4 => criture verticale, sinon horizontale.
  wFlags <- 8 => n'crit rient, fait seulement les calculs de dimension
                 i.e. ne fait que de mesurer le texte Extend.
  wFlags <- 16 => affiche le texte avec un font, sinon, le font est transparent.
  Retourne la taille en pixels prise par le string  l'affichage.
-------------------------------------------------------------------}

IMPLEMENTATION
{""""""""""""}

const
  DIX              = 10; { nombre de DC qui peuvent tre li  une font.
                           En principe il ne peut pas avoir plus de
                           5 DC ouvert en mme temps! }
  QUINZE           = 15; { nombre de fonts mmoris possible }

{ erreurs possibles }
errReleaseFont               = 595;
errAllFontLocked             = 596;
errFontBadDeleteObject       = 597;
errFontBadNum                = 598;
errFontLocked                = 599;

var
  aLogFont         : array[1..QUINZE] of TLogFont;
  { description du font cr }

  ahFont           : array[1..QUINZE] of HFont;
  { font slectionnable dans un DC }

  anLockCount      : array[1..QUINZE] of integer;
  { Indique si le nombre de ReleaseFont est le mme que le nombre
    de SelectFont. Si non, c'est une erreur, qui ne permet pas
    de dtruire la Font.
    Cette erreur entraine des pertes de ressources Windows.
    Ce n'est pas fatal, mais il faut la corriger.
    Normalement un Device Contexte ne reste pas ouvert trs longtemps,
    et d-SelecteObject des fonts qui lui sont associs.
    Si dans un DC, incrment de 100. }

  anUsedOrder      : array[1..QUINZE] of integer;
  { Ordre des dernires utilisation.
    1 = plus rcemment utilise, 2 = suivante, ... nLastFont = dernire.
    Si il est ncssaire de remplacer une font par une autre de la liste
    ahFont, ceci permet de remplacer la font qui n'a pas t utilise
    depuis le plus longtemps. }

  nLastFont        : integer;
  { dernier indice correspondant  un font }

  ahDC             : array[1..DIX] of HDC;
  { List des DC ayant slectionn un font }

  anDCFont         : array[1..DIX] of integer;
  { font slectionn par le DC de mme indice }

  ahDCOldFont      : array[1..DIX] of HFont;
  { Ancien font du DC de mme indice }

  nLastDC          : integer;
  { dernier indice correspondant  un DC }

type TStr80 = string[80];

procedure ErrorMsg1(nErr:integer; strS:TStr80);
{=============================================}
begin { - ErrorMsg1 - }

if (nErr > 0) then ErrorWrite(word(nErr), strS)
else               ErrorWrite(0, strS);
end;  { - ErrorMsg1 - }

procedure InitFontsUtilities;
{===========================}
{ Initialisation  faire en dbut de programme }
var
  nn               : integer;
begin { - InitFontsUtilities - }
for nn:=1 to QUINZE do begin
   ahFont[nn]      :=0; { font slectionnable dans un DC }
   anLockCount[nn] :=0; { indique si les fonts sont effacable (<=0) ou non }
   anUsedOrder[nn] :=nn;
   end;

for nn:=1 to DIX do begin
   ahDC[nn]        :=0; { List des DC ayant slectionn un font }
   anDCFont[nn]    :=0; { font slectionn par le DC de mme indice }
   ahDCOldFont[nn] :=0; { Ancien font du DC de mme indice }
   end;

nLastFont:=0;
{ dernier indice correspondant  un font }

nLastDC:=0;
{ dernier indice correspondant  un DC }

end;  { - InitFontsUtilities - }

procedure DoneFontsUtilities;
{===========================}
{ Libration  faire en fin de programme.
  Si nErr <> 0, ce n'est pas normal, il y a une erreur dans le programme.
  Cela indique que une font ne peut pas tre libre. }
var
  nRep,
  nn               : integer;
begin { - DoneFontsUtilities - }
for nn:=1 to nLastFont do begin
   if (anLockCount[nn] > 0) then begin
      ErrorMsg1(errFontLocked, IntToStr(nn,0));
      end;

   { On DeleteObject la font, dans tous les cas }
   nRep:=byte(DeleteObject(ahFont[nn]));
   if (nRep = 0) then ErrorMsg1(0, 'Error in DoneFontsUtilities');
   end;

nLastFont:=0;
{ dernier indice correspondant  un font }

nLastDC:=0;
{ dernier indice correspondant  un DC }
end;  { - DoneFontsUtilities - }

function EqualFont(var F1,F2:TLogFont):boolean;
{=============================================}
{ Retourn TRUE si les deux fonts sont gaux }
begin { - EqualFont - }
EqualFont:=FALSE;
if (F1.lfHeight <> F2.lfHeight) then exit;
if (StrComp(F1.lfFaceName, F2.lfFaceName) <> 0) then exit;
if (F1.lfEscapement <> F2.lfEscapement) then exit;

if (F1.lfWeight <> F2.lfWeight) then exit;
if (F1.lfItalic <> F2.lfItalic) then exit;
if (F1.lfUnderline <> F2.lfUnderline) then exit;
if (F1.lfStrikeOut <> F2.lfStrikeOut) then exit;
if (F1.lfCharSet <> F2.lfCharSet) then exit;

if (F1.lfWidth <> F2.lfWidth) then exit;
if (F1.lfOrientation <> F2.lfOrientation) then exit;
if (F1.lfOutPrecision <> F2.lfOutPrecision) then exit;
if (F1.lfClipPrecision <> F2.lfClipPrecision) then exit;
if (F1.lfQuality <> F2.lfQuality) then exit;
if (F1.lfPitchAndFamily <> F2.lfPitchAndFamily) then exit;

EqualFont:=TRUE;
end;  { - EqualFont - }

function SearchFont(var LogFont:TLogFont):integer;
{================================================}
{ Cherche la description de font.
  Retourne l'indice correspondant si trouv, 0 sinon }
var
  nn               : integer;
begin { - SearchFont - }
for nn:=1 to nLastFont do begin
   if EqualFont(LogFont, aLogFont[nn]) then begin
      SearchFont:=nn;
      exit;
      end;
   end;

SearchFont:=0;
end;  { - SearchFont - }

function SearchDC(hDC1:HDC):integer;
{==================================}
{ Cherche le DC parmis les ahDC.
  Retourne l'indice correspondant si trouv, 0 sinon }
var
  nn               : integer;
begin { - SearchDC - }
if (hDC1 = 0) then begin
   ErrorMsg1(0, 'Error in SearchDC, hDC = 0');
   SearchDC:=0;
   exit;
   end;

for nn:=1 to nLastDC do begin
   if (hDC1 = ahDC[nn]) then begin
      SearchDC:=nn;
      exit;
      end;
   end;

SearchDC:=0;
end;  { - SearchDC - }

function UsedOrderPermuttation(nFont,nPlace:integer):integer;
{===========================================================}
{ Cherche la position de nFont, et effectue la permuttation cyclique
  pour que nFont prenne la place anUsedOrder[nPlace].
  Si nPlace <= 0, ne fait pas de permuttation.
  Retourne la position dans anUsedOrder de nFont,
  retourne < 0 si erreur.
------------------------------------------------------------}
var
  nn,
  nPos             : integer;

begin { - UsedOrderPermuttation - }
{ Cherche la position de nFont dans anUsedOrder }
nPos:=0;
repeat
   inc(nPos);
   until (nPos > QUINZE{nLastFont}) or (anUsedOrder[nPos] = nFont);

if (nPos > QUINZE{nLastFont}) then begin
   { erreur impossible en principe }
   ErrorMsg1(0, 'Impossible error in UsedOrderPermuttation.');
   UsedOrderPermuttation:=-1;
   exit;
   end;

UsedOrderPermuttation:=nPos;

if (nPlace <= 0) then exit;

{ ici anUsedOrder[nPos] = nFont }
{ permuttation, pour amener anUsedOrder[nPos] en anUsedOrder[nPlace]. }
if (nPlace < nPos) then begin
   { dcalage vers la droite }
   for nn:=nPos downto nPlace+1 do anUsedOrder[nn]:=anUsedOrder[nn-1];
   anUsedOrder[nPlace]:=nFont;
   end
else if (nPlace > nPos) then begin
   { dcalage vers la gauche }
   for nn:=nPos to nPlace-1 do anUsedOrder[nn]:=anUsedOrder[nn+1];
   anUsedOrder[nPlace]:=nFont;
   end
end;  { - UsedOrderPermuttation - }

function DeleteFontNum(nFont:integer):integer;
{============================================}
{ efface la font nFont de la liste et de la mmoire.
  Ne change rien dans anUsedOrder.
  Retourne 0 si OK
           errFontBadNum si nFont <= 0 ou nFont > nLastFont
           errFontBadDeleteObject si erreur au DeleteObject(ahFont[nFont])
           errFontLocked si Locked, i.e. si Count Lock > 0,
              n'efface pas, ce qui fait perdre de la mmoire ressources.
-------------------------------------------------------------------------}
var
  nn : integer;

begin { - DeleteFontNum - }
DeleteFontNum:=0;

if (nFont <= 0) or (nFont > nLastFont) then begin
   { la font n'existe pas }
   ErrorMsg1(errFontBadNum, IntToStr(nFont,0));
   DeleteFontNum:=errFontBadNum;
   exit;
   end;

if (anLockCount[nFont] > 0) then begin
   { la font est locked, on ne peut pas l'effacer }
   ErrorMsg1(errFontLocked, IntToStr(nFont,0));
   DeleteFontNum:=errFontLocked;
   exit;
   end;

{ ici la Font ne peut pas tre dans les rfrences de anDCFont, car sinon elle
  serait Locked. }

{ on efface la font }
if not DeleteObject(ahFont[nFont]) then begin
   ErrorMsg1(errFontBadDeleteObject, IntToStr(nFont,0));
   DeleteFontNum:=errFontBadDeleteObject;
   end;

{ ----- renumrote les tableaux }
aLogFont[nFont]    :=aLogFont[nLastFont];
ahFont[nFont]      :=ahFont[nLastFont];
anLockCount[nFont] :=anLockCount[nLastFont];
ahFont[nLastFont]  :=0;
anLockCount[nLastFont]:=0;

{ remet  jours l'indication du font. utilis par les DC }
for nn:=1 to nLastDC do
   if (anDCFont[nn] = nLastFont) then anDCFont[nn] := nFont;

dec(nLastFont);
end;  { - DeleteFontNum - }

function CreateFont(var LogFont:TLogFont):integer;
{================================================}
{ Cr une font selon la description de font.
  la rajoute dans la liste et
  retourne l'indice correspondant.
  Au besoin libre une font qui n'est pas Locked.
  Retourne 0 si pas possible. }
var
  nFont,
  nErr,
  nn               : integer;
begin { - CreateFont - }
if (nLastFont >= QUINZE) then begin
   { recherche la premire font qui n'est pas Locked, dans l'ordre inverse
     d'historique d'utilisation. }
   nn:=nLastFont+1; { = QUINZE+1 }
   repeat
      dec(nn);
      until (nn <= 0) or (anLockCount[anUsedOrder[nn]] <= 0);

   if (nn <= 0) then begin
      { pas normal, erreur du programme qui block trop de fonts }
      ErrorMsg1(errAllFontLocked, 'in CreateFont');
      CreateFont:=0;
      exit;
      end;

   { Font que l'on efface, et remplace par la nouvelle.
     en gnral nn = nLastFont = QUINZE ici. }
   nFont:=anUsedOrder[nn];

   { efface la font de la mmoire, donc decr. nLastFont }
   DeleteFontNum(nFont);
   end;

{ ajoute une font supplmentaire. }
inc(nLastFont);
aLogFont[nLastFont]    :=LogFont;
ahFont[nLastFont]      :=CreateFontIndirect(LogFont);
anLockCount[nLastFont] :=0;

{ dcalage vers la droite et insertion de nLastFont en premire position.
  Cette font est la dernire utilise. }
nErr:=UsedOrderPermuttation(nLastFont, 1);
if (nErr < 0) then begin
   { erreur impossible en principe }
   ErrorMsg1(0, 'Error in CreateFont from UsedOrderPermuttation, nLastFont='
              + IntToStr(nLastFont,4));
   end;

CreateFont:=nLastFont;
end;  { - CreateFont - }

function GetFontNum(var LogFont:TLogFont):HFont;
{==============================================}
{ 1) Cherche une font correspondant  la description LogFont.
  2) Si pas trouv, cr la font correspondante.
     Retourne l'indice de  ahFont  ou la font est stocke.
     retourne 0 si erreur.
-------------------------------------------------------------------------}
var
  nErr,
  nFont            : integer;

begin { - GetFontNum - }
GetFontNum:=0;

{ cherche si la font existe dans la liste aLogFont }
nFont:=SearchFont(LogFont);

if (nFont = 0) then begin
   nFont:=CreateFont(LogFont); { cre la font }
   if (nFont = 0) then exit; { erreur, dj affiche }
   end
else begin { la Font existait dj }
   { indique que cette font est la dernire utilise. }
   nErr:=UsedOrderPermuttation(nFont, 1);
   { erreur impossible en principe }
   if (nErr < 0) then
      ErrorMsg1(0, 'Error in GetFontNum from UsedOrderPermuttation, nFont='
              + IntToStr(nFont,4));
   end;

GetFontNum:=nFont;
end;  { - GetFontNum - }

function GetFont(var LogFont:TLogFont):HFont;
{===========================================}
{ 1) Cherche une font correspondant  la description LogFont.
  2) Si pas trouv, cr la font correspondante.
     Retourne le font.
     retourne 0 si erreur.
-------------------------------------------------------------------------}
var
  nFont            : integer;

begin { - GetFont - }
{ cherche si la font existe dans la liste aLogFont }
nFont:=GetFontNum(LogFont);

if (nFont > 0) then GetFont:=ahFont[nFont]
else begin ErrorMsg1(0, 'Error in GetFont'); GetFont:=0; end;
end;  { - GetFont - }

function SelectFont(hDC1:HDC; var LogFont:TLogFont):HFont;
{========================================================}
{ 1) Cherche une font correspondant  la description LogFont.
  2) Si pas trouv, cr la font correspondante
  3) Slectionne la font dans le DC
  4) Mmorise le hDC, et la font qui s'y trouvait, si c'est le premier select
     Retourne la font prcedemment slectionn.
-------------------------------------------------------------------------}
var
  nFont            : integer;
  nDC              : integer;
  hOldFont        : HFont;

begin { - SelectFont - }
{ cherche si la font existe dans la liste aLogFont, si non, cre la font }
nFont:=GetFontNum(LogFont);

if (nFont = 0) then begin
   SelectFont:=0; { erreur, dj affiche }
   ErrorMsg1(0, 'Error in SelectFont from GetFontNum, nFont=0.');
   exit;
   end;

{ slectionne la nouvelle font dans le DC }
hOldFont:=SelectObject(hDC1, ahFont[nFont]);

{ cherche si ce DC tait dj reli a une font dcrite ici }
nDC:=SearchDC(hDC1);

if (nDC = 0) then begin
   { mmorise la font qui tait dans le DC }
   inc(nLastDC);
   ahDC[nLastDC]       :=hDC1;
   ahDCOldFont[nLastDC]:=hOldFont;
   nDC:=nLastDC;
   end
else begin
   { ce DC contenait dj une font, on peut diminuer son count lock }
   dec(anLockCount[anDCFont[nDC]], 100);
   if (anLockCount[anDCFont[nDC]] < 0) then begin
      ErrorMsg1(0, 'Error in SelectFont LockCount < 0!');
      anLockCount[anDCFont[nDC]]:=0;
      end;
   end;

{ mmorise la font slectionne dans le DC }
anDCFont[nDC]:=nFont;

{ Le font est slectionn dans un DC, sont il est Locked }
inc(anLockCount[nFont], 100);

SelectFont:=hOldFont;
end;  { - SelectFont - }

function ReleaseFont(hDC1:HDC):HFont;
{===================================}
{ remet l'ancien font dans ce DC, si il est dans la liste, et
  retourne le dernier font qui tait dans ce DC.
  Dcrmente le cont lock.
-------------------------------------------------------------------------}
var
  nDC              : integer;
  hOldFont        : HFont;
begin { - ReleaseFont - }
{ cherche si ce DC tait dj reli a une font dcrite ici }
nDC:=SearchDC(hDC1);

if (nDC = 0) then begin
   ErrorMsg1(errReleaseFont, 'in ReleaseFont');
   ReleaseFont:=0;
   exit;
   end;

{ reslectionne la font d'origine du DC }
hOldFont:=SelectObject(hDC1, ahDCOldFont[nDC]);

dec(anLockCount[anDCFont[nDC]], 100);

{ un DC de moins dans la liste }
ahDC[nDC]        :=ahDC[nLastDC];
anDCFont[nDC]    :=anDCFont[nLastDC];
ahDCOldFont[nDC] :=ahDCOldFont[nLastDC];
ahDC[nLastDC]:=0;
anDCFont[nLastDC]:=0;
ahDCOldFont[nLastDC]:=0;
dec(nLastDC);

ReleaseFont:=hOldFont;
end;  { - ReleaseFont - }

function IsHFontRegistered(hFnt:HFont):boolean;
{=============================================}
{ Indique si la Font dfini par le handle hFnt est dans la liste des
  fonts gr par cette unit.
  Si oui, elle sera dtruite automatiquement en fin de programme,
   l'appelle de DoneFontUtilities, et donc il ne faut
  pas la dtruire autrement. }
var nn : integer;
begin { - IsHFontRegistered - }
IsHFontRegistered:=TRUE;
for nn:=1 to nLastFont do if (ahFont[nn] = hFnt) then exit;

IsHFontRegistered:=FALSE;
end;  { - IsHFontRegistered - }

function LockFont(hFnt:HFont):integer;
{====================================}
{ Incrmente de 1 le count lock, et retourne sa valeur.
  Retourne -1111 si pas trouv.
-------------------------------------------------------------------------}
var
  nFont            : integer;

begin { - LockFont - }
nFont:=0;

repeat
   inc(nFont);
   until (nFont > nLastFont) or (hFnt = ahFont[nFont]);

if (nFont > nLastFont) then begin
   { pas trouv }
   LockFont:=-1111;
   exit;
   end;

{ incrmente le count lock }
inc(anLockCount[nFont]);

LockFont:=anLockCount[nFont];
end;  { - LockFont - }

function UnLockFont(hFnt:HFont):integer;
{======================================}
{ Dcrmente de 1 le count lock, et retourne sa valeur.
  Retourne -1111 si pas trouv.
-------------------------------------------------------------------------}
var
  nFont            : integer;

begin { - UnLockFont - }
nFont:=0;

repeat
   inc(nFont);
   until (nFont > nLastFont) or (hFnt = ahFont[nFont]);

if (nFont > nLastFont) then begin
   { pas trouv }
   UnLockFont:=-1111;
   exit;
   end;

{ dcrmente le count lock }
dec(anLockCount[nFont]);

UnLockFont:=anLockCount[nFont];
end;  { - UnLockFont - }

function DeleteFont(var LogFont:TLogFont):integer;
{================================================}
{ A utiliser particulirement quand on veut remplacer une font par une
  autre, et qu'il n'y a plus de raison de garder l'ancienne.
  Efface la font nFont de la liste et de la mmoire.
  Retourne 0 si OK
           errFontBadNum si nFont <= 0 ou nFont > nLastFont
           errFontBadDeleteObject si erreur au DeleteObject(ahFont[nFont])
           errFontLocked si Locked, i.e. si Count Lock > 0,
                         mais efface quand-mme.
-------------------------------------------------------------------------}
var
  nErr,
  nFont            : integer;

begin { - DeleteFont - }
{ cherche si la font existe dans aLogFont }
nFont:=SearchFont(LogFont);

DeleteFont:=DeleteFontNum(nFont);

{ Cherche la position de nFont dans anUsedOrder, et effectue des permut.
  pour mettre nFont en fin. } {on peut mettre nLastFont+1  la place de QUINZE }
nErr:=UsedOrderPermuttation(nFont, QUINZE);
if (nErr < 0) then begin
   ErrorMsg1(0, 'Error in DeleteFont from UsedOrderPermuttation, nFont='
            + IntToStr(nFont,4));
   end;

end;  { - DeleteFont - }

function TextOutWithSymbol(
{==========================}
  var fontText     : TLogFont; { transmis par var pour simplicit, mais pas modifi }
      pixXP0,      { position X d'affichage }
      pixYP0       { position Y d'affichage }
                   : integer;
  var strS         : String;  { texte  afficher }
      ptSizeText   : integer;
      hDC0         : HDC;
      wFlags       : word
      ):longint;
{ Affichage de texte, avec caractres grec, indices et exposant.
  \g(abc) => abc en grec.
  \+      => suite en exposant, ou retour aprs un \-
  \-      => suite en indice, ou retour aprs un \+
  wFlags <-  4 => criture verticale, sinon horizontale.
  wFlags <-  8 => n'crit rient, fait seulement les calculs de dimension
                 i.e. ne fait que de mesurer le texte Extend.
  wFlags <- 16 => affiche le texte avec un font, sinon, le font est transparent.
  Retourne la taille en pixels prise par le string  l'affichage.
-------------------------------------------------------------------}
label EndProc;
var
  pixXP,
  pixYP,
  nPos1,
  nBeg,
  nLen
                : integer;
  nExposant
  { niveau dans la ligne.
     0 = normal
     1 = exposant
    -1 = indice }
  ,pixExpInd    { dcalage en pixels pour indices et exposant }
                : integer;
  lfHeightMem,
  lfEscapementMem  :integer;
  fontSymbol       : TLogFont;
  fTxt,
  fVert            : boolean; { si criture verticale }
  rSize            : Size;

   procedure ChangeFontSize(nExposant:integer);
   {==========================================}
   { change la taille des fonts en fonction des exposants et indices }
   begin { - ChangeFontSize - }
   if (nExposant <> 0) then begin
      fontText.lfHeight :=Round(-0.8*ptSizeText);
      fontSymbol.lfHeight :=Round(-0.8*ptSizeText);
      end
   else begin
      fontText.lfHeight :=Round(-ptSizeText);
      fontSymbol.lfHeight :=Round(-ptSizeText);
      end;

   SelectFont(hDC0, fontText);
   end;  { - ChangeFontSize - }

begin { - TextOutWithSymbol - }
{=============================}
lfEscapementMem:=fontText.lfEscapement;
lfHeightMem    :=fontText.lfHeight;

{nLogPixelsy:=GetDeviceCaps(hDC0, LogPixelsy); ???? }
fontText.lfHeight :=-Round(ptSizeText);

fTxt  :=((wFlags and 8)  = 0);
fVert :=((wFlags and 4) <> 0);

if ((wFlags and 16) = 0) then SetBkMode(hDC0, TRANSPARENT);

{ pour dfinir une font de symbole mathmatiques. }
with fontSymbol do begin
    lfHeight           :=fontText.lfHeight;
    lfWidth            :=0;
    lfEscapement       :=0; { en 1/10 de degrs }
    lfOrientation      :=0;   { ignor }
    lfWeight           :=FW_BOLD;  {FW_DONTCARE;}
    lfItalic           :=0;
    lfUnderline        :=0;
    lfStrikeOut        :=0;
    lfCharSet          :=ANSI_CHARSET;
    lfOutPrecision     :=OUT_DEFAULT_PRECIS;
    lfClipPrecision    :=CLIP_DEFAULT_PRECIS;
    lfQuality          :=DEFAULT_QUALITY;
    lfPitchAndFamily   :=DEFAULT_PITCH or FF_SWISS;
    StrPCopy(@lfFaceName, 'Symbol');
   end; { - with fontSymbol do begin - }

if fVert then begin
   { pour criture verticale }
   fontText.lfEscapement:=900;
   fontSymbol.lfEscapement:=900;
   SelectFont(hDC0, fontText);
   end;

pixXP  :=pixXP0;
pixYP  :=pixYP0;
nBeg :=1;
nLen :=length(strS);
nExposant:=0;

{ dcalage des exposants et indices }
GetTextExtentPoint32(hDC0, @strS[nBeg], nLen, rSize);
pixExpInd :=rSize.cy div 2;
SelectFont(hDC0, fontText);

{ boucle, pour affichage, en cherchant les caractres spciaux }
repeat
   nPos1:=Pos('\', Copy(strS,nBeg, 250));

   if (nPos1 = 0) then begin
      { plus de caractres spciaux }
{         SelectFont(hDC0, fontText); dj, par dfaut }
      nLen:=length(strS)+1 - nBeg;
      if fTxt then TextOut(hDC0, pixXP, pixYP, @strS[nBeg], nLen);
      fontText.lfHeight :=Round(-ptSizeText);
      fontSymbol.lfHeight :=Round(-ptSizeText);

      { retourne la dimension du texte affich }
      GetTextExtentPoint32(hDC0, @strS[nBeg], nLen, rSize);
      if fVert then begin
         pixYP:=pixYP - rSize.cx;
         pixXP:=pixXP + rSize.cy;
         TextOutWithSymbol:=MAKELong(pixYP0 - pixYP, pixXP - pixXP0);
         end
      else begin
          pixXP:=pixXP + rSize.cx;
          pixYP:=pixYP - rSize.cy;
          TextOutWithSymbol:=MAKELong(pixXP - pixXP0, pixYP0 - pixYP);
          end;
      goto EndProc;
      end;

   nLen:=nPos1 - 1;
   if (nLen > 0) then begin
      { imprime le texte normal }
{         SelectFont(hDC0, fontText);  dj par dfaut }
      if fTxt then TextOut(hDC0, pixXP, pixYP, @strS[nBeg], nLen);
      GetTextExtentPoint32(hDC0, @strS[nBeg], nLen, rSize);
      if fVert then pixYP:=pixYP - rSize.cx
      else          pixXP:=pixXP + rSize.cx;
      end;

   { Suite du texte  }
   nBeg:=nBeg + nLen+1;  { strS[nBeg-1] = '\'  }

   if (strS[nBeg] = 'g') and (strS[nBeg+1] = '(') then begin
      { dbut du texte avec symboles }
      nBeg:=nBeg + 2;  { saute le '\g(' }

      { imprime le texte de caractres spciaux }
      nPos1:=Pos(')', Copy(strS,nBeg,250));
      if (nPos1 > 0) then nLen:=nPos1-1;

      if (nLen > 0) then begin
         SelectFont(hDC0, fontSymbol);
         if fTxt then TextOut(hDC0, pixXP, pixYP, @strS[nBeg], nLen);
         GetTextExtentPoint32(hDC0, @strS[nBeg], nLen, rSize);
         if fVert then pixYP:=pixYP - rSize.cx
         else          pixXP:=pixXP + rSize.cx;

         { reslectionne la font par dfaut }
         SelectFont(hDC0, fontText);
         end;

      { Suite du texte  }
      nBeg:=nBeg + nLen+1;  { strS[nBeg-1] = ')' }
      end
   else if (strS[nBeg] = '+') then begin
      { exposant }
      inc(nExposant);
      nBeg:=nBeg + 1;  { saute le '\+' }
      if fVert then pixXP:=pixXP - pixExpInd
      else          pixYP:=pixYP - pixExpInd;
      if (nExposant > 0) then begin
         { Sinon l'exposant est trop haut }
         if fVert then pixXP:=pixXP + pixExpInd div 2
         else          pixYP:=pixYP + pixExpInd div 2;
         end;

      ChangeFontSize(nExposant);
      end
   else if (strS[nBeg] = '-') then begin
      { indice }
      if (nExposant > 0) then begin
         { ramne l'exposant au niveau normal }
         if fVert then pixXP:=pixXP - pixExpInd div 2
         else          pixYP:=pixYP - pixExpInd div 2;
         end;

      dec(nExposant);
      nBeg:=nBeg + 1;  { saute le '\-' }
      if fVert then pixXP:=pixXP + pixExpInd
      else          pixYP:=pixYP + pixExpInd;

      ChangeFontSize(nExposant);
      end
   else begin { par une commande spciale }
       { strS[nBeg] = caractre qui suit le '\' }
{         SelectFont(hDC0, fontText); dj, par dfaut }
      { Si deux  '\'  se suivent immdiatement, ignore le premier. }
      if (strS[nBeg] = '\') then inc(nBeg);

      { affiche le '\' }
      if fTxt then TextOut(hDC0, pixXP, pixYP, @strS[nBeg-1], 1);
      GetTextExtentPoint32(hDC0, @strS[nBeg-1], 1, rSize);
      if fVert then pixYP:=pixYP - rSize.cx
      else          pixXP:=pixXP + rSize.cx;
      end;
   until FALSE; { for ever, on sort avec un goto EndProc;; }

EndProc:;
if fVert
   or (fontText.lfHeight <> lfHeightMem)
   or (fontText.lfEscapement <> lfEscapementMem) then begin
   fontText.lfEscapement:=lfEscapementMem;
   fontText.lfHeight    :=lfHeightMem;
{   SelectFont(hDC0, fontText);}
   end;

ReleaseFont(hDC0);
end;  { - TextOutWithSymbol - }

initialization
InitFontsUtilities;
{ Initialisation  faire en dbut de programme }

finalization
DoneFontsUtilities;
{ Libration  faire en fin de programme.
  Si nErr <> 0, ce n'est pas normal, il y a une erreur dans le programme.
  Cela indique que une font ne peut pas tre libre. }

{ *** UNIT wFontManager *** }
end.
