unit wGrBase0;

{ Unit pour simplifier le graphisme de base sous windows.
  Il simplifie :
  - Le dessin de lignes, avec attribut de couleurs, de styles, de taille, ...
  - Le dessin de marques, carrs, croix, triangles, ... de couleur, taille, ...
  - L'affichage de texte, avec font, taille, attribut, ...
  - La correspondance entre coordonnes utilisateur et device (cran ou imprimante).
  - La gestion des Device Context.
  - Le positionnnement dans la fentre
-------------------------------------------------------------------------}

// To tell FreePascal that the application is a Graphical User Interface
// i.e. is running in windows, not in text mode.
{$APPTYPE GUI}

interface
{"""""""}

Uses
  wWindow0
 ,wFontManager 
 ,wPrinter0
 ,wErrors0
 ,wWindowBase0
 ,wGlobals2
 ,wDebug0
 ,Strings
 ,Windows;

const
  markPlus     = 1;
  markCross    = 2;
  markStar     = 3;
  markSquare   = 4;
  markDiamond  = 5;
  markCircle   = 6;
  markTriangle = 7;

const
flaClipAuto  =-1; { => choix automatique du clipping }
flaClipNone  = 0; { => pas de clipping }
flaClipWin1  = 1; { => clipping limit  la fentre 1 }
flaClipWin2  = 2; { => clipping limit  la fentre 2 }

type
  TStyle           = integer;
  TSize            = integer;

  TypeLineAttrib = record
     style2     : TStyle;
     pixWidth2  : integer;
     mmWidth2   : Float;
     col2       : TColorRef;
     end; { - TypeLineAttrib = record - }

  TypeMarkAttrib = record
     style2      : TStyle;
     pixWidth2   : integer;
     mmWidth2    : Float;
     col2        : TColorRef;
     mmSize2     : Float;     { = taille de la marque en milimtres }
     end; { - TypeMarkAttrib = record - }

  TypeTextAttrib = record
     font             : TLogFont;
     col              : integer;
     size             : integer; { en points }
     end; { - TypeTextAttrib = record - }

PGraphicBase = ^TGraphicBase;
TGraphicBase = object
{===================}
  { pointeur sur la fentre associe }
  pWnd0         : PWindow0;

  { - Handle sur la fentre lie au graphisme }
  hWnd0            : HWnd;

  { - Pour Simplifier les sorties sur imprimante. }
  pPrinter0        : PGAPPrinter;

  { - Donnes des devices contextes }
  { ------------------------------- }
  hDC0             : HDC;  { Device Contexte ncssaire au graphisme }

  { =TRUE pour que un cercle reste un cercle, pour toute taille d'cran }
  fIsometricXY     : boolean;

  { indique le device d'affichage, c.f.  cteOutScreen, cteOutPrinter, ... }
  nOutputDevice    : integer;

  vScaleDPI        : Float;
  { Permet de dimimuer la rsolution de l'image alloue en mmoire pour
    l'imprimante.  = 1 normalement }

  { - Donnes des Lignes }
  { -------------------- }
  styleLine        : TStyle;    {=PS_SOLID, PS_DASH, PS_DOT, PS_DASHDOT, c.f. CreatePen }
  pixWidthLine     : integer;   { 0..infinity }
  mmWidthLine      : Float;     { paisseur minimale en milimtres }
  colLine          : TColorRef; { =RGB(0..255, 0..255, 0..255) }

  { L'paisseur choisie est celle maximale entre celle donne en
    pixels et celle donnes en milimtres }
  pixWidthLineUsed : integer; { taille en pixels utilis pour les lignes }

  { - Donnes des Marques }
  { --------------------- }
  styleMark        : TStyle;    {=PS_SOLID, PS_DASH, PS_DOT, PS_DASHDOT, c.f. CreatePen }
  pixWidthMark     : integer;     { 0..infinity }
  mmWidthMark      : Float;     { paisseur minimale en milimtres }
  colMark          : TColorRef; { =RGB(0..255, 0..255, 0..255) }
  mmSizeMark       : Float;     { = taille de la marque en milimtres }

  { L'paisseur choisie est celle maximale entre celle donne en
    pixels et celle donnes en milimtres }
  pixWidthMarkUsed : integer; { taille en pixels utilis pour les Mark }

  { - Donnes des Textes }
  { -------------------- }
  fontText         : TLogFont;
  colText          : TColorRef;
  sizeText         : integer;  { en points }

  { Dfini la manire dont les pixels ragissent avec le font.
    Les valeurs habituelles sont :
    R2_COPYPEN    = dessin normal
    R2_NOTXORPEN  = dessin s'annulle si il est fait deux fois de suite }
  ROP2Mode         : integer;

  { position de la fentre 1 en combinant des milimtres et coord. relatives }
  mmX1Left   , relX1Left   : Float;
  mmY1Top    , relY1Top    : Float;
  mmX1Width  , relX1Width  : Float;
  mmY1Height , relY1Height : Float;

  { position de la fentre 2 en combinant des milimtres et coord. relatives }
  mmX2Left   , relX2Left   : Float;
  mmY2Top    , relY2Top    : Float;
  mmX2Width  , relX2Width  : Float;
  mmY2Height , relY2Height : Float;

  { Position sur papier pour imprimer le graphisme et le texte.
    Les marges gauches et hautes, sont relativement aux bords du papier,
    mais des valeurs minimales sont imposs par les limites de l'imprimante.
    Elles sont testes automatiquement. }
  mmPrinterLeft   , relPrinterLeft   : Float;
  mmPrinterTop    , relPrinterTop    : Float;
  mmPrinterWidth  , relPrinterWidth  : Float;
  mmPrinterHeight , relPrinterHeight : Float;

  { Est mis  TRUE si les dimensions sur l'imprimante, sont pour la
    fentre 1 en entier,
    est mis  FALSE si les dimensions sur l'impr. correspondent aux
    dimensions d'une fentre, i.e. le positions de la fentre 1 correspondent 
    (0,0) = haut, gauche, (1,1)=bas,droite dans la rgion sur papier
    dfinie ci-dessus. }
  fPrinterBigWin1  : boolean;

  { limites des coordonnes utilisateur. }
  useXMin, useYMin, useXMax, useYMax  : Float;

  { - Donnes prives }
  { ----------------- }
{  Private}
  {"""""}
  vPixXSmm,vPixYSmm  : Float; { transf. de mm en pixels du device contexte }

  { Position dans le device (cran ou imprimante), en coordonnes pixels }
  pixXPos,pixYPos   : integer;
  vPixXPos,vPixYPos : Float;  { position en pixels, prcise pour ne pas accumuler les
                                erreurs d'arrondi }

  { Pour transformation de coordonnes utilisateur en pixels }
  vPixX0user, vPixXSuser,
  vPixY0user, vPixYSuser  : Float;

  { position de la fentre 1 en pixels }
  pixX1Left,pixY1Top,
  pixX1Width,pixY1Height    : integer;

  { position de la fentre 2 en pixels }
  pixX2Left,pixY2Top,
  pixX2Width,pixY2Height    : integer;

  { position sur la page en pixels, soit de la Window,
    soit de la fentre 1. C.f. fPrinterBigWin1 }
  pixPrinterLeft,pixPrinterTop,
  pixPrinterWidth,pixPrinterHeight    : integer;

  { Taille du papier en pixels }
  pixPrinterPaperWidth,
  pixPrinterPaperHeight : integer;

  { Largeur en pixels, accessible sur le papier }
  pixPrinterTotWidth,
  pixPrinterTotHeight : integer;

  { marges aux bords du papier, non accessible  l'criture }
  pixPrinterBorderLeft,
  pixPrinterBorderTop   : integer;

  { Rgion de clipping pour les fentres 1 et 2. Ici, se sont des rectangles }
  hRgnWin1,
  hRgnWin2
                   : HRgn;

  nWinCLip         : integer;
  { = flaClipNone si le clipping est enlev, i.e. est pour la fentre entire (client),
    = flaClipWin1 si le Clipping est dfini pour la fentre 1,
    = flaClipWin2 si le clipping est dfini pour la fentre 2. }

  { indique si le clipping a t choisi par l'utilisateur (=FALSE)
    ou si le choix se fait automatiquement }
  fClippingAuto    : boolean;

  { Pour mmoriser le style de ligne }
  hPenUsed,     { mmorise le Pen Actuellement utilis }
  hFirstPenUsed,
  hLinePenNew,   { Style de ligne, taille, couleur }
  hMarkPenNew    { Style des marques, taille, couleur }
                : HPen;

  hMemoryBitMap,   { pour mmoriser ce que l'on veut sortir sur l'imprimante }
  hOldBitMap       : HBitMap;

  { Si nOutputDevice = cteOutPrinterOneShoot, permet d'indiquer que
    le device contexte est celui d'une fentre. Sinon la valeur est nulle }
  hWindowTemp      : HWnd;

  { - Mthodes publiques }
  { ==================== }
  public
  {""""}

  { - Mthodes de coordonnes utilisateurs et de dfinition de fentres graphiques }
  { ------------------------------------------------------------------------------ }
  procedure UpdateWin12Pos;
  {=======================}
  { Redfini la position et la taille des fentres 1 et 2, en combinant des coordonnes absolues
    en milimtres et des coordonnes relative  la taille de la Window. C.f. DOC }

{  procedure SetUserCoord(    useXMin1,useYMin1,useXMax1,useYMax1:Float);
  procedure GetUserCoord(var useXMin1,useYMin1,useXMax1,useYMax1:Float);}

  { - constructeur et destructeur - }
  { ------------------------------- }
  { Init, en donnant le pointeur sur la fentre associe }
  constructor Init(pWnd1:PWindow0);
  destructor  Done;

  procedure CopyGrBaseTo(pGB:PGraphicBase; Flags:word); virtual;
  { duplifie les donnes du graphique de base.
    Ne duplifie pas tout, est  complter.
  -------------------------------------------------------}

  procedure SaveGrBase(var fOut:Text; var nErr:integer); virtual;
  procedure LoadGrBase(var fIn:Text; var nErr:integer); virtual;

  { - Mthodes des Device Contexte }
  { ------------------------------ }
  procedure ReleaseDC0; virtual;
  { Ferme le device contexte, met hDC0  0.
     chaque LineTo, Mark, write, polyline,... il faut tester si le Device
    Contexte est ouvert, et si non, il faut l'ouvrir avant traitement, et le
    fermer aprs. }

  procedure SetHWindow(hWinDC:HWnd); virtual;
  { Indique que le Device contexte donn dans SetNewDC0 est celui
    de la fentre hWinDC.
    Ne s'utilise que avant d'appeler SetNewDC0(handle > 0).
    Typiquement s'utilise dans la procedure de traitement du WM_PAINT :
    BeginPaint(Self.HWindow, PaintInfo);  ( PaintInfo donn par WM_PAINT )
    SetHWindow(Self.HWindow);
    SetNewDC0(PaintDC);
    ... on fait le dessin
    GrWin.ReleaseDC0;
    EndPaint(Self.HWindow, PaintInfo);
    end;
  -------------------------------------------------------------------------}

  procedure GetPrinterSizes(hDCPrn:HDC); virtual;
  { Dfini des dimensions de papier d'imprimantes dfinie par hDCPrn }

  procedure SetNewDC0(hDC1:HDC); virtual;
  { Ne jamais utilis avec hDC1 = 0, hors de cette unit.
    Permet de changer de device contexte, pour l'associer  l'imprimante par exemple.
    On peut aussi l'associ  l'crant dans le traitement de WM_PAINT.
    c.f. SetHWindow. }

  function GetDC0:HDC; virtual;
  { Ouvre un Device Contexte pour le graphisme, et indique qu'il est ouvert,
    pas le fait que hDC0 <> 0. Retourne hDC0 = handle sur le device contexte.
    Si il y a eut une erreur, hDC0 = 0. }

  function fIsDCOpen:boolean; virtual;
  { retourne TRUE si  hDC0 <> 0,  i.e. si le device contexte est ouvert }

  procedure CallPrintPage(pPrOut:PMsgPrintOut; var nErr:integer); virtual;
  { Demande d'appeler la mthode  PrintPage  de  pPrOut^.
    Elle doit donc avoir t redfinie dans l'application qui utilise
    cette unit, ici tGrBase.pas
    Valeurs possibles pour nErr : (en plus de 0)
    cteErrPPrinterNotInitialised
    cteErrPPrintOutNotInitialised
  --------------------------------------------------------------------------}

  procedure ClearPrinterImage(var nErr:integer); virtual;
  { Efface l'image destine  l'imprimante.
    Si le device contexte d'imprimante est ouvert, ne fait que d'effacer l'image,
    sinon libre la mmoire utilise par le bit map.
    Valeurs possibles pour nErr : (en plus de 0)
    cteErrNoMemoryBitMap
  -------------------------------------------------------------------------}

  procedure OutputOnPrinter(var nErr:integer); virtual;
  { Indique que les prochains graphismes doivent tre fait sur l'imprimante,
    et ouvre le device contexte.
    On peut utiliser :
    ReleaseDC0     pour le refermer,
    GetDC0         pour l'ouvrir denouveau,
    PrinterPrint   pour sortir effectivement sur l'imprimante le graphique
    OutputOnScreen pour l'affichage  l'cran
    Valeurs possibles pour nErr : (en plus de 0)
    cteErrNULLhDC0
    cteErrNULLPrinterDC
    cteErrCreateCompatibleBitMap
    cteErrSelectObjectBitMap
  -------------------------------------------------------------------------}

  procedure SetPrinterTopHere; virtual;
  { pour positionner le graphique sous la position courrante,
    Utile aprs un certain nombre de  Writeln0,  pour que le graphique
    soit en dessous du texte. }

  procedure PrinterPrint(var nErr:integer); virtual;
  { Pour sortir  l'imprimante les dessins qui ont t fait pour l'imprimante, c--d
    les dessins fait aprs l'appelle  OutputOnPrinter.
    Valeurs possibles pour nErr : (en plus de 0)
    cteErrNoMemoryBitMap
    cteErrPPrinterNotInitialised
    cteErrPPrintOutNotInitialised
  -------------------------------------------------------------------------}

  procedure ScreenPrint(var nErr:integer); virtual;
  { Pour sortir  l'cran les dessins qui ont t fait pour l'imprimante, c--d
    les dessins fait aprs l'appelle  OutputOnPrinter.
    Valeurs possibles pour nErr : (en plus de 0)
    cteErrNoMemoryBitMap
  -------------------------------------------------------------------------}

  procedure OutputOnScreen(var nErr:integer); virtual;
  { Indique que les prochains graphismes doivent tre fait sur l'cran,
    ferme le device contexte d'imprimante, mais
    n'ouvre pas le device contexte de l'cran.
    Ceci est utilis gnralement aprs PrinterPrint, pour fermer
    l'affichage sur l'imprimante.
  -------------------------------------------------------------------------}

  procedure CloseOutputOnPrinter(var nErr:integer); virtual;
  { Imprime ce qui a t prpar sur l'imprimante,
    Indique que les prochains graphismes doivent tre fait sur l'cran,
    ferme le device contexte d'imprimante, mais
    n'ouvre pas le device contexte de l'cran.
    Ceci est utilis gnralement aprs OutputOnPrinter.
    Est quivalent  :
    PrintPrinter(nErr);  OutputOnScreen(nErr);  ClearPrinter(nErr);
  -------------------------------------------------------------------------}

  procedure SetGrClipRgn(nWinClip1:integer);
  { Pour utilisation interne, ne pas utiliser.
    Dfini le clipping a utiliser, et mmorise celui utilis }

  procedure SetClipping(nClipType:integer);
  { Choix du clipping, flaClipAuto = par dfaut.
    nClipType =
    flaClipAuto => choix automatique du clipping
    flaClipNone => pas de clipping
    flaClipWin1 => clipping limit  la fentre 1
    flaClipWin2 => clipping limit  la fentre 2. }

  procedure ClearWin2( pixFromXBorder, mmFromXBorder,
                       pixFromYBorder, mmFromYBorder:Float);
  { permet d'effacer le contenu de la fentre 2, sans effacer ce qui est
    trs proche des bords }

  { - Mthodes de positionnement }
  { ------------------------------------------ }
  procedure MovePixTo(vPixX,vPixY:Float); virtual;
  { dplacement absolu en pixels, utile pour revenir  un point de dpart. }

  procedure MoveTo0(XX,YY:Float); virtual;        { dpl. absolu en coord. utilisateur }
  procedure MoveMMRelTo(mmX,mmY:Float); virtual;  { dpl. relatif en milimtres }
  procedure MoveWin1To(relX,relY:Float); virtual; { dpl. absolu dans la fentre 1 }
  procedure GetXYPos(var XX,YY:Float); virtual;   { position en coordonnes utilisateur }
  procedure GetXYMMPos(var mmX,mmY:Float); virtual;
  procedure GetXYWin1Pos(var relX,relY:Float); virtual;

  { MoveWin1To  Permet des positionnements dans la fentre 1.
    (0  ,0  ) = en haut  gauche.
    (1  ,1  ) = en bas   droite.
    (0.5,0.5) = au milieu de la fentre 1.  }

  { - Conversion de coordonnes - }
  { ----------------------------- }
  procedure MMToPix(mmX,mmY:Float; var vPixX,vPixY:Float); virtual;
  procedure MMToPixInt(mmX,mmY:Float; var pixX,pixY:integer); virtual;
  { conversion de milimtres en pixels. }
  { conversion en relatif }

  procedure PixToMM(vPixX,vPixY:Float; var mmX,mmY:Float); virtual;
  { conversion de pixels en milimtres en relatif. }

  procedure MMToUser(mmX,mmY:Float; var vX,vY:Float); virtual;
  { conversion de milimtres coord utilisateur en relatif. }

  procedure UserToMM(vX,vY:Float; var mmX,mmY:Float); virtual;
  { conversion en relatif }

  procedure PixToUser(vPixX,vPixY:Float; var vX,vY:Float); virtual;
  { conversion en absolu }

  procedure UserToPix(vX,vY:Float; var vPixX,vPixY:Float); virtual;
  procedure UserToPixInt(vX,vY:Float; var pixX,pixY:integer); virtual;
  { conversion en absolu }

  { - Mthodes des Lignes }
  { --------------------- }
  procedure LinePixTo(vPixX,vPixY:Float); virtual;      { en pixels }
  { Dessin de ligne en pixels, clipping dans la fentre 1. }

  procedure LineTo0(XX,YY:Float); virtual;      { en coord. utilisateur }
  procedure Line0(X1,Y1,X2,Y2:Float); virtual;
  procedure PolyLine0(nbData:integer; var aX,aY:array of Float); virtual;

  procedure LineMMRelTo(mmX,mmY:Float); virtual;
  { dplacement relatif en millimtres }

  procedure PolyLineMMRel(nbData:integer; vHomot:Float; var ammX,ammY:array of Float); virtual;
  { Fait une succession de lignes relatives au point prcdent, les longueurs sont
    donnes en milimtres * vHomot. }

  procedure LineWin1To(relX,relY:Float); virtual;  { c.f.  MoveWin1To  }

  procedure GetLineAttribRec( var recLineAt:TypeLineAttrib); virtual;
  { Lit l'attribut de ligne actuelle }

  procedure SetLineAttrib( style2:TStyle;  pixWidth2:integer;
                           mmWidth2:Float; col2:TColorRef); virtual;
  { Change l'attribut de ligne }

  procedure SetLineAttrib2( style2:TStyle;  pixWidth2:integer;
                            mmWidth2:Float; col2:TColorRef; var recLineAt:TypeLineAttrib); virtual;
  { Change l'attribut de ligne, en retournant l'attibut prcdent }

  procedure SetLineAttribRec(var recLineAt:TypeLineAttrib); virtual;
  { Change l'attribut de ligne, en donnant un Record, comme celui obtenu
    par GetLineAttribRec. }

  procedure DrawBox(nPts:integer; x1,y1,x2,y2,x3,y3,x4,y4:Float); virtual;
  { nPts = 2  5
    Dessine de nPts-1 lignes de (x1,y1)  (x4,y4),
    avec retour  (x1,y1) si  nPts >= 5.
    Les coordonnes sont ceux utilisateur, dans la fentre 2,
    la couleur, le style et width des lignes sont utilis.
  -------------------------------------------------------------------------}

  { Sert  mettre des cadres autour des fentres }
  procedure BorderWin1( Style1:TStyle; pixWidth1:integer; mmWidth1:Float;
                        Color1:TColorRef); virtual;
  procedure BorderWin2( Style1:TStyle; pixWidth1:integer; mmWidth1:Float;
                        Color1:TColorRef); virtual;

  { est-ce utile? Au dbut, non! }
{   procedure MoveRelTo(X1,Y1:Float); ???}
{   procedure LineRelTo(X1,Y1:Float); ???}

  { - Mthodes des Marques }
  { ---------------------- }
  procedure Mark0(nMark:integer); virtual;
  { Place une marque :  +  x  *  []  O  }

  procedure MarkTo(X1,Y1:Float; nMark:integer); virtual;
  { Se positionne avant de placer une marque }

  procedure PolyMark(nbData,nMark:integer; var aX,aY:array of Float); virtual;

  procedure GetMarkAttribRec( var recMarkAt:TypeMarkAttrib); virtual;
  { Lit l'attribut de Marque actuelle }

  procedure SetMarkAttrib( style2:TStyle;  pixWidth2:integer; mmWidth2:Float;
                           col2:TColorRef; mmSize2:Float); virtual;
  { Change l'attribut de Marque }

  procedure SetMarkAttribRec(var recMarkAt:TypeMarkAttrib); virtual;
  { Change l'attribut de Mark, en donnant un Record, comme celui obtenu
    par GetLineAttribRec. }

  { - Mthodes des Textes }
  { ---------------------- }
  procedure WriteMove(rX,rY:Float; strS:string; flags:word);
  { Dplacement relatif avant d'afficher le string, et indique dans flags
    le type de dplacement dsir aprs l'affichage.
    Flags <- 1 => en fin de string,
    Flags <- 2 => sous le string.
    Flags <- 4 => criture verticale }

  procedure Write0(strS:string); virtual;
  { avec positionnement en fin de ligne }

  procedure Writeln0(strS:string); virtual;
  { Avec retour sous la position de dpart }

  procedure WriteMoveRel(rX,rY:Float; strS:string);
  { comme WritelnMoveRel, mais se positionne en fin de ligne }

  procedure WritelnMoveRel(rX,rY:Float; strS:string);
  { Souvant, pour placer un texte sur un graphique, on veut se placer  un
  endroit sur
    le graphique avec  MoveTo  , puis se positionner plus prcicment en
  utilisant
    des coordonnes de taille caractres ou des coordonnes pixels.
    se dplace relativement  la position courante de :
    rX * longueur(strS)  en X.
    rY * hauteur(strS)   en Y.
    Donc rX = 0   => dbut  de string en nXPos.
    Donc rX = 0.5 => milieu de string en nXPos.
    Donc rX = 1   => fin    de string en nXPos.
    Donc rY = 0   => bas    de string en nYPos.
    Donc rY = 0.5 => mi-H   de string en nYPos.
    Donc rY = 1   => haut   de string en nYPos. }

  procedure WriteVert(strS:string);
  { Ecriture verticale,
    Avec continuation en fin de dernire position. }

  procedure WritelnVert(strS:string);
  { Ecriture verticalement.
    Avec retour sous la position de dpart. }

  procedure WriteVertMoveRel(rX,rY:Float; strS:string);
  { c.f. WriteMovRel }

  procedure WritelnVertMoveRel(rX,rY:Float; strS:string);
  { c.f. WritelnMovRel }

  procedure GetPixTextSize(strS:string; var pixX,pixY:integer);
  procedure GetMMTextSize(strS:string; var mmX,mmY:Float);

  procedure SaveTextAttribRec(var recTextAt:TypeTextAttrib);
  procedure GetTextAttribRec( var recTextAt:TypeTextAttrib);
  end; { - TGraphicBase = object - }

const
{ indication du device de sortie des graphiques }
cteOutScreen           = 1;
cteOutPrinter          = 2;

{ Pour la mthode dessinant tout d'un coup sur l'imprimante. Donc au prochain
  ReleaseDC0,  le device de sortie est remis   cteOutScreen. }
cteOutPrinterOneShoot  = 6;

implementation
{""""""""""""}

const
{ ces erreurs peuvent tre lies  des messages d'erreurs }
{ 600..699}
cteErrNULLhDC0                         = 601;
cteErrNoRasterPrinter                  = 602;
cteErrNULLPrinterDC                    = 603;


cteErrPPrinterNotInitialised           = 611;
cteErrPPrintOutNotInitialised          = 612;
cteErrNoMemoryBitMap                   = 615;
cteErrCreateCompatibleBitMap           = 619;
cteErrSelectObjectBitMap               = 621;

{ erreurs 52600..52699 (errBaseLibrary = 52000) }
errBaseLibrary = 52000;
msgGetDCis0                            = cteErrNULLhDC0 + errBaseLibrary;
msgNoRasterPrinter                     = cteErrNoRasterPrinter + errBaseLibrary;
msgCreateDCis0                         = cteErrNULLPrinterDC + errBaseLibrary;
msgGetDCis0ForPrinter                  = 604 + errBaseLibrary;
msgFontNotSelected                     = 605 + errBaseLibrary;
msgBadUseOfSetNewDC0                   = 606 + errBaseLibrary;


msgNoGraphicToPrint                    = 650 + errBaseLibrary;
msgNoPrinterCompatibleDC               = 651 + errBaseLibrary;


type
  { Pour pouvoir imprimer le contenu de hMemoryBitMap }
  TPrintOutLocal = object(TMsgPrintout)
    procedure PrintPage(Page: Word; var Rect: TRect; Flags:Word); virtual;
    { Flags = 0, rserv pour futur }
    end;  { - TPrintOutLocal = object(TPrintout) - }

var
  { pour indiquer  TPrintOutLocal.PrintPage, des donnes sur TGraphicBase, tel que
    le bit map  imprimer }
//  glpGrWin            : PGraphicBase;

  { pour corriger l'inconsistance de Windows, qui ne tient pas compte de l'origine du
    Viewport pour dfinir la rgion  Clipper.
    = (0,0) si affichage  l'cran, mais = dcalage des marges du papier si impression. }
  glpixXDec,
  glpixYDec : integer;

{************************************************************************}
{*****  TPrintOutLocal *****}
{************************************************************************}

procedure TPrintOutLocal.PrintPage(Page: Word; var Rect: TRect; Flags:Word);
{==========================================================================}
begin { - TPrintOutLocal.PrintPage - }
end;  { - TPrintOutLocal.PrintPage - }

{************************************************************************}
{*****  TGraphicBase *****}
{************************************************************************}

{ - Mthodes de coordonnes utilisateurs et de dfinition de fentres graphiques }
{ ------------------------------------------------------------------------------ }
procedure TGraphicBase.UpdateWin12Pos;
{====================================}
{ Redfini la position et la taille des fentres 1 et 2, en combinant des coordonnes absolues
  en milimtres et des coordonnes relative  la taille de la Window. C.f. DOC }
var
  tRect0           : TRect;
  pixOutputWidth,
  pixOutputHeight  : integer;
  pixX2Right,
  pixY2Bottom      : integer;
  fPrinterOutput,
  fModifyWin1TopLeft : boolean;
  pixX0,
  pixY0            : integer; { dcalage en cas d'impression directe sur imprimante }

     procedure SetForPrinter;
     {======================}
     { Position d'impression sur papier }
     begin { - SetForPrinter - }
     pixPrinterLeft   := Round( mmPrinterLeft   * vPixXSmm +
                                relPrinterLeft   * pixPrinterTotWidth);
     pixPrinterTop    := Round( mmPrinterTop    * vPixYSmm +
                                relPrinterTop    * pixPrinterTotHeight);
     pixPrinterWidth  := Round( mmPrinterWidth  * vPixXSmm +
                                relPrinterWidth  * pixPrinterTotWidth);
     pixPrinterHeight := Round( mmPrinterHeight * vPixYSmm +
                                relPrinterHeight * pixPrinterTotHeight);

     if (pixPrinterLeft < pixPrinterBorderLeft) then
         pixPrinterLeft :=pixPrinterBorderLeft;
     if (pixPrinterTop < pixPrinterBorderTop) then
         pixPrinterTop :=pixPrinterBorderTop;
     if (pixPrinterWidth > pixPrinterTotWidth) then
         pixPrinterWidth :=pixPrinterTotWidth;
     if (pixPrinterHeight > pixPrinterTotHeight) then
         pixPrinterHeight :=pixPrinterTotHeight;

     pixOutputWidth  :=pixPrinterWidth;
     pixOutputHeight :=pixPrinterHeight;
     fModifyWin1TopLeft:=fPrinterBigWin1;

     pixX0:=pixPrinterLeft;
     pixY0:=pixPrinterTop;
     fPrinterOutput:=TRUE;  fPrinterOutput:=fPrinterOutput;
     end;  { - SetForPrinter - }

begin { - TGraphicBase.UpdateWin12Pos - }
{=======================================}
fPrinterOutput:=FALSE;
fModifyWin1TopLeft:=FALSE;
pixX0:=0;
pixY0:=0;

{ - pour la taille de la fentre en pixels - }
if (nOutputDevice = cteOutScreen) then begin
   GetClientRect(hWnd0, tRect0);
   pixOutputWidth  :=tRect0.Right;
   pixOutputHeight :=tRect0.Bottom;
   end
else if (nOutputDevice = cteOutPrinter) then begin
   SetForPrinter;
   end
else begin {  if (nOutputDevice = cteOutPrinterOneShoot) then begin }
   if (hWindowTemp <> 0) then begin
      { Le device contexte est celui d'une fentre }
      GetClientRect(hWindowTemp, tRect0);
      pixOutputWidth  :=tRect0.Right;
      pixOutputHeight :=tRect0.Bottom;
      end
   else begin
      {if (hDC0 = 0) then exit; // erreur, pas normal }
      SetForPrinter;
      end;

{   DebugWrite(1, 'w,h=' + IntToStr(GetDeviceCaps(hDC0, HORZRES), 9)
                        + IntToStr(GetDeviceCaps(hDC0, VERTRES), 9));}
   end;

{ conversion de coordonnes pour la fentre 1 }
if fModifyWin1TopLeft then begin
   { Modifie les positions de la fentre 1, pour que elles correspondent
      la rgion dfinie sur le papier }
   { Le dcalage se fera  l'impression }
   pixX1Left   :=pixX0;  { = pixPrinterLeft }
   pixY1Top    :=pixY0;  { = pixPrinterTop }
   pixX1Width  :=pixPrinterWidth;
   pixY1Height :=pixPrinterHeight;
{   SetRectRgn( hRgnWin1, pixX1Left+glpixXDec, pixY1Top+glpixYDec,
               pixX1Left + pixPrinterWidth+1+glpixXDec, pixY1Top + pixPrinterHeight+1+glpixXDec);}
   end
else begin
   pixX1Left   := pixX0 + Round(mmX1Left   * vPixXSmm + relX1Left * pixOutputWidth);
   pixY1Top    := pixY0 + Round(mmY1Top    * vPixYSmm + relY1Top  * pixOutputHeight);
   pixX1Width  := Round(mmX1Width  * vPixXSmm + relX1Width  * pixOutputWidth);
   pixY1Height := Round(mmY1Height * vPixYSmm + relY1Height * pixOutputHeight);
   end;

if (pixX1Width  = 0) then pixX1Width :=1;
if (pixY1Height = 0) then pixY1Height:=1;
SetRectRgn( hRgnWin1, pixX1Left+glpixXDec, pixY1Top+glpixYDec,
            pixX1Left + pixX1Width+1+glpixXDec, pixY1Top + pixY1Height+1+glpixYDec);

{ conversion de coordonnes pour la fentre 2 }
{ Les pixels extrmes qui se dessinent sont :
  (pixX2Left, pixY2Top) et (pixX2Left + pixX2Width, pixY2Top + pixY2Height) }
pixX2Left   := pixX1Left + Round(mmX2Left   * vPixXSmm + relX2Left   * pixX1Width);
pixY2Top    := pixY1Top  + Round(mmY2Top    * vPixYSmm + relY2Top    * pixY1Height);
pixX2Width  :=             Round(mmX2Width  * vPixXSmm + relX2Width  * pixX1Width);
pixY2Height :=             Round(mmY2Height * vPixYSmm + relY2Height * pixY1Height);
if (pixX2Width  = 0) then pixX2Width :=1;
if (pixY2Height = 0) then pixY2Height:=1;
SetRectRgn(hRgnWin2, pixX2Left+glpixXDec, pixY2Top+glpixYDec,
           pixX2Left + pixX2Width+1+glpixXDec, pixY2Top + pixY2Height+1+glpixYDec);

{ dfini la conversion de coord. utilisateur en coord. pixels }
{ useXMin = pixX2Left,
  useXMax = pixX2Right:=pixX2Left + pixX2Width. }
vPixXSuser :=(pixX2Width) / (useXMax - useXMin);
vPixX0user :=pixX2Left - useXMin * vPixXSuser;

{ useYMin = pixY2Top,
  useYMax = pixY2Bottom:=pixY2Top + pixY2Height. }
vPixYSuser :=(pixY2Height) / (useYMin - useYMax);
vPixY0user :=pixY2Top - useYMax * vPixYSuser;

if fIsometricXY then begin
   { Pour que un cercle reste un cercle, pour toute taille d'cran }
   if (vPixXSuser * vPixYSmm > -vPixYSuser * vPixXSmm) then begin
      { Diminue les dimensions en  X  pour avoir isomtrie }
      pixX2Right :=pixX2Left + pixX2Width;
      vPixXSuser := -vPixYSuser * vPixXSmm / vPixYSmm;
      pixX2Width :=Round((useXMax - useXMin) * vPixXSuser);
      pixX2Left  :=pixX2Right - pixX2Width;
      vPixX0user :=pixX2Left - useXMin * vPixXSuser;
      end
   else begin
      { Diminue les dimensions en  Y  pour avoir isomtrie }
      pixY2Bottom :=pixY2Top + pixY2Height;
      vPixYSuser  := -vPixXSuser * vPixYSmm / vPixXSmm;
      pixY2Height :=Round((useYMin - useYMax) * vPixYSuser);
      pixY2Top    :=pixY2Bottom - pixY2Height;
      vPixY0user  :=pixY2Top - useYMax * vPixYSuser;
      end;

   SetRectRgn( hRgnWin2, pixX2Left+glpixXDec, pixY2Top+glpixYDec,
               pixX2Left + pixX2Width+1+glpixXDec, pixY2Top + pixY2Height+1+glpixYDec);
   end;  { - if fIsometricXY then begin - }

end;  { - TGraphicBase.UpdateWin12Pos - }

{ - constructeur et destructeur - }
{ ------------------------------- }
constructor TGraphicBase.Init(pWnd1:PWindow0);
{============================================}
{ Init, en donnant le pointeur sur la fentre associe }
var
  nPixX_Width
 ,nPixY_Height
                 : integer;
  rRect : RECT;

begin { - TGraphicBase.Init - }
pWnd0:=pWnd1;

{ Handle de la fentre lie au graphisme }
hWnd0:=pWnd0^.hWin;

{ trait erreurs }
hDC0:=GetDC(hWnd0);

{ rsolution en pixels / milimtes }
vPixXSmm:=GetDeviceCaps(hDC0, HORZRES)  / GetDeviceCaps(hDC0, HORZSIZE);
vPixYSmm:=GetDeviceCaps(hDC0, VERTRES) / GetDeviceCaps(hDC0, VERTSIZE);

{ trait erreurs }
ReleaseDC(hWnd0, hDC0);

GetClientRect(hWnd0, rRect);

nPixX_Width:=rRect.right;
nPixY_Height:=rRect.bottom;

{ Coordonnes utilisateur par dfaut }
useXMin :=  0;
useXMax :=nPixX_Width;
useYMin :=  0;
useYMax :=nPixY_Height;

{ le device contexte est ferm. }
hDC0:=0;
fIsometricXY:=FALSE;

{ Sortie  l'cran, par dfaut }
nOutputDevice :=cteOutScreen;

hWindowTemp:=0;

{ Permet de dimimuer la rsolution de l'image alloue en mmoire pour
  l'imprimante.  = 1 normalement }
vScaleDPI  :=1;

{ Clipping rectangle, ils seront redfini dans UpdateWin12Pos }
hRgnWin1:=CreateRectRgn(0, 0, nPixX_Width, nPixY_Height);
hRgnWin2:=CreateRectRgn(0, 0, nPixX_Width, nPixY_Height);

{ indique si le clipping a t choisi par l'utilisateur (=FALSE)
 ou si le choix se fait automatiquement }
fClippingAuto :=TRUE;

{ par dfaut, la fentre 1 = tout l'cran }
mmX1Left:=0;    mmY1Top:=0;   mmX1Width:=0;   mmY1Height:=0;
relX1Left:=0;   relY1Top:=0;  relX1Width:=1;  relY1Height:=1;

{ par dfaut, la fentre 2 = tout l'cran }
mmX2Left:=0;   mmY2Top:=0;  mmX2Width:=0;  mmY2Height:=0;
relX2Left:=0;   relY2Top:=0;  relX2Width:=1;  relY2Height:=1;

{ Position sur papier pour imprimer le graphisme et le texte.
  Les marges gauches et hautes, sont relativement aux bords du papier,
  mais des valeurs minimales sont imposs par les limites de l'imprimante.
  Elles sont testes automatiquement. }
mmPrinterLeft   := 20; {mm}
mmPrinterTop    :=  8; {mm}
mmPrinterWidth  :=100; {mm}
mmPrinterHeight :=100; {mm}
relPrinterLeft:=0;  relPrinterTop:=0;  relPrinterWidth:=0;  relPrinterHeight:=0;

{ par dfaut, la fentre 1 = la rgion dfinie sur papier }
fPrinterBigWin1:=TRUE;
{fPrinterBigWin1:=FALSE;}

{ couleurs et tailles par dfaut des lignes }
styleLine    :=PS_SOLID;
pixWidthLine :=1;
mmWidthLine  :=0.1; { taille minimale en milimtres }
colLine      :=RGB(0,0,0); { = BLACK }

{ couleurs et tailles par dfaut des marques }
styleMark     :=PS_SOLID;
pixWidthMark  :=1;
mmWidthMark   :=0.1; { taille minimale en milimtres }
colMark       :=RGB(0,0,0); { = BLACK }
mmSizeMark    :=1;  { = taille de la marque en milimtres }

{ ligne se dessine normalement }
ROP2Mode      := R2_COPYPEN;

{ dfini les fentres et le systme de coordonnes }
UpdateWin12Pos;

{ Initialisation pour sortie sur imprimante }
if (glpPrinter = NIL) then new(glpPrinter, Init);
pPrinter0:=glpPrinter;

//??glpGrWin:=nil;
hMemoryBitMap:=0;
hOldBitMap:=0;

{ Taille en pixels de la sortie sur imprimante }
{ Sera dfini au GetDC0 }
pixPrinterWidth  := 700;
pixPrinterHeight :=1000;

{ Dfini les caractristiques de la font }
colText         :=RGB(0,0,0);
sizeText        :=10; { taille en points }

{ cre une font pour sortie sur imprimante }
with fontText do begin
    lfHeight           :=Round(-sizeText * vPixYSmm * 25.4 / 72);
    lfWidth            :=0;
    lfEscapement       :=0; { en 1/10 de degrs }
    lfOrientation      :=0;   { ignor }
    lfWeight           :=FW_BOLD;  {FW_DONTCARE;}
    lfItalic           :=0;
    lfUnderline        :=0;
    lfStrikeOut        :=0;
    lfCharSet          :=ANSI_CHARSET;
    lfOutPrecision     :=OUT_DEFAULT_PRECIS;
    lfClipPrecision    :=CLIP_DEFAULT_PRECIS;
    lfQuality          :=DEFAULT_QUALITY;
    lfPitchAndFamily   :=DEFAULT_PITCH or FF_SWISS;
    StrCopy(@lfFaceName, 'Arial');

   end; { - with fontText do begin - }

{ C.F.
  TPrinter.DeviceSettings (Field) ou
  TPrintOut.Size (Field) ou
  TDEVMODE (Record)
  DeviceCapabilities (function)

  Pour dfinir la rgion sur l'imprimante, ou on dessins.
  Position gauche, haut, largeur, hauteur
------------------------------------------------------------}
end;  { - TGraphicBase.Init - }

destructor TGraphicBase.Done;
{===========================}
begin { - TGraphicBase.Done - }
DeleteObject(hRgnWin2);
DeleteObject(hRgnWin1);

if (pPrinter0 <> glpPrinter) then begin
   { cas non habituelle }
//   dispose(pPrinter0, Done);
   pPrinter0 :=NIL;
   end;
{else le dispose de glpPrinter est fait dans TSKEAPPLICATION}

if (hDC0 <> 0) then ReleaseDC0;
end;  { - TGraphicBase.Done - }

procedure TGraphicBase.CopyGrBaseTo(pGB:PGraphicBase; Flags:word);
{================================================================}
{ Duplifie les donnes du graphique de base.
  Ne duplifie pas tout, est  complter.
-------------------------------------------------------}
begin { - TGraphicBase.CopyGrBaseTo - }
if (pGB = NIL) then exit;

pGB^.sizeText :=sizeText;

pGB^.mmX1Left    :=mmX1Left;
pGB^.relX1Left   :=relX1Left;
pGB^.mmY1Top     :=mmY1Top;
pGB^.relY1Top    :=relY1Top;
pGB^.mmX1Width   :=mmX1Width;
pGB^.relX1Width  :=relX1Width;
pGB^.mmY1Height  :=mmY1Height;
pGB^.relY1Height :=relY1Height;

pGB^.mmX2Left    :=mmX2Left;
pGB^.relX2Left   :=relX2Left;
pGB^.mmY2Top     :=mmY2Top;
pGB^.relY2Top    :=relY2Top;
pGB^.mmX2Width   :=mmX2Width;
pGB^.relX2Width  :=relX2Width;
pGB^.mmY2Height  :=mmY2Height;
pGB^.relY2Height :=relY2Height;
end;  { - TGraphicBase.CopyGrBaseTo - }

procedure TGraphicBase.SaveGrBase(var fOut:Text; var nErr:integer);
{================================================================}
begin { - TGraphicBase.SaveGrBase - }
{$i-}
{ version dveloppeur-utilisateur }
writeln(fOut, '# ...Signature : _TGraphicBase');
writeln(fOut,'# ', 100);

writeln(fOut,'# ', mmX1Left   :7:2);
writeln(fOut,'# ', relX1Left  :7:2);
writeln(fOut,'# ', mmY1Top    :7:2);
writeln(fOut,'# ', relY1Top   :7:2);
writeln(fOut,'# ', mmX1Width  :7:2);
writeln(fOut,'# ', relX1Width :7:2);
writeln(fOut,'# ', mmY1Height :7:2);
writeln(fOut,'# ', relY1Height:7:2);

writeln(fOut,'# ', mmX2Left   :7:2);
writeln(fOut,'# ', relX2Left  :7:2);
writeln(fOut,'# ', mmY2Top    :7:2);
writeln(fOut,'# ', relY2Top   :7:2);
writeln(fOut,'# ', mmX2Width  :7:2);
writeln(fOut,'# ', relX2Width :7:2);
writeln(fOut,'# ', mmY2Height :7:2);
writeln(fOut,'# ', relY2Height:7:2);

nErr:=IOResult;
{$i+}
end;  { - TGraphicBase.SaveGrBase - }

procedure TGraphicBase.LoadGrBase(var fIn:Text; var nErr:integer);
{================================================================}
var
  strS             : string[50];
  nVersion         : integer;
  c1,c2            : char;
begin { - TGraphicBase.LoadGrBase - }
{$i-}
{ version dveloppeur-utilisateur }
readln(fIn, strS); { strS = '# ...Signature : _TGraphicBase'  }
readln(fIn,c1,c2, nVersion); nVersion:=nVersion;

readln(fIn,c1,c2,mmX1Left);
readln(fIn,c1,c2,relX1Left);
readln(fIn,c1,c2,mmY1Top);
readln(fIn,c1,c2,relY1Top);
readln(fIn,c1,c2,mmX1Width);
readln(fIn,c1,c2,relX1Width);
readln(fIn,c1,c2,mmY1Height);
readln(fIn,c1,c2,relY1Height);

readln(fIn,c1,c2,mmX2Left);
readln(fIn,c1,c2,relX2Left);
readln(fIn,c1,c2,mmY2Top );
readln(fIn,c1,c2,relY2Top);
readln(fIn,c1,c2,mmX2Width);
readln(fIn,c1,c2,relX2Width);
readln(fIn,c1,c2,mmY2Height);
readln(fIn,c1,c2,relY2Height);

nErr:=IOResult;
{$i+}
end;  { - TGraphicBase.LoadGrBase - }

{ - Mthodes des Device Contexte }
{ ------------------------------ }
procedure TGraphicBase.ReleaseDC0;
{================================}
var
  nOK              : integer;

begin { - TGraphicBase.ReleaseDC0 - }
if (hDC0 = 0) then exit;

SelectObject(hDC0, hFirstPenUsed);
nOK:=Byte(DeleteObject(hMarkPenNew));
{if (nOK = 0) then DBW('***Error in DeleteObject(hMarkPenNew)');}
nOK:=Byte(DeleteObject(hLinePenNew));  nOK:=nOK;
{if (nOK = 0) then DBW('***Error in DeleteObject(hLinePenNew)');}

ReleaseFont(hDC0); // from wFontManager

if (nOutputDevice = cteOutScreen) then nOK:=ReleaseDC(hWnd0, hDC0)  {=1 si OK, 0 si pas OK }
else if (nOutputDevice = cteOutPrinter) then begin
   { efface le device contexte en mmoire }
   if (hOldBitMap <> 0) then SelectObject(hDC0, hOldBitMap);
   DeleteDC(hDC0);
   hOldBitmap:=0;
   end
else if (nOutputDevice = cteOutPrinterOneShoot) then begin
   { la prochaine fois que l'on ouvre le Device contexte, ce sera pour l'cran }
   nOutputDevice:=cteOutScreen;
   end;

hDC0:=0;
hWindowTemp:=0;
end;  { - TGraphicBase.ReleaseDC0 - }

procedure TGraphicBase.SetHWindow(hWinDC:HWnd);
{=============================================}
{ Indique que le Device contexte donn dans SetNewDC0 est celui
  de la fentre hWinDC.
  Ne s'utilise que avant d'appeler SetNewDC0(handle > 0).
  Typiquement s'utilise dans la procedure de traitement du WM_PAINT :
  BeginPaint(Self.HWindow, PaintInfo);  ( PaintInfo donn par WM_PAINT )
  SetHWindow(Self.HWindow);
  SetNewDC0(PaintDC);
  ... on fait le dessin
  GrWin.ReleaseDC0;
  EndPaint(Self.HWindow, PaintInfo);
  end;
-------------------------------------------------------------------------}
begin { - TGraphicBase.SetHWindow - }
hWindowTemp:=hWinDC;
{ il est remis  0 dans ReleaseDC }
end;  { - TGraphicBase.SetHWindow - }

procedure TGraphicBase.GetPrinterSizes(hDCPrn:HDC);
{=================================================}
{ Dfini des dimensions de papier d'imprimantes dfinie par hDCPrn }
var
  nRep             : integer;
  rPoint           : TPoint;

begin { - TGraphicBase.GetPrinterSizes - }
{ utilise les coordonnes pixels du Device Contexte }
SetMapMode(hDCPrn, MM_TEXT);

{ Largeur et hauteur accessible du papier }
{ dfinition des marges haut, gauche,  pour sortie sur imprimante ou plotter }
pixPrinterTotWidth  :=Round(GetDeviceCaps(hDCPrn, HORZRES) / vScaleDPI);
pixPrinterTotHeight :=Round(GetDeviceCaps(hDCPrn, VERTRES) / vScaleDPI);

{ Lit les dimensions physiques totales du papier }
{ (nRep = 0) => pas implment. }
nRep:=Escape(hDCPrn, GETPHYSPAGESIZE, SizeOf(rPoint), NIL, @rPoint);
if (nRep <= 0) then begin
   { Pas normal, on donne une dimension physique trop petite }
   pixPrinterPaperWidth  :=pixPrinterTotWidth;
   pixPrinterPaperHeight :=pixPrinterTotHeight;
   end
else begin
   { dimension physique du papier en pixels }
   pixPrinterPaperWidth  :=Round(rPoint.x / vScaleDPI);
   pixPrinterPaperHeight :=Round(rPoint.y / vScaleDPI);
   end;

{ Lit les marges gauche, haut non accessible  l'criture sur papier }
{ (nRep = 0) => pas implment. }
nRep:=Escape(hDCPrn, GETPRINTINGOFFSET, SizeOf(rPoint), NIL, @rPoint);
if (nRep <= 0) then begin rPoint.x:=0; rPoint.y:=0; end;
{DebugWrite(1, IntToStr(nRep,8) + IntToStr(rPoint.x,8) + IntToStr(rPoint.y,8));}

pixPrinterBorderLeft :=Round(rPoint.x / vScaleDPI);
pixPrinterBorderTop  :=Round(rPoint.y / vScaleDPI);
end;  { - TGraphicBase.GetPrinterSizes - }

procedure TGraphicBase.SetNewDC0(hDC1:HDC);
{=========================================}
{ permet de changer de device contexte, pour l'associer  l'imprimante par exemple. }
var
  hDCPrn           : HDC;
  nRep             : integer;
  hFontText        : HFont;

begin { - TGraphicBase.SetNewDC0 - }
{ pour corriger l'inconsistance de Windows, qui ne tient pas compte de l'origine du
  Viewport pour dfinir la rgion  Clipper.
  = (0,0) si affichage  l'cran, mais = dcalage des marges du papier si impression. }
glpixXDec:=0;
glpixYDec:=0;

if (hDC1 <> 0) then begin
   { on dfini un nouveau Device contexte pour le graphisme }
   ReleaseDC0;
   hDC0:=hDC1;

   { ouverture pour affichage sur l'imprimante d'une traite, sans refermer le device contexte
     avant l'impression. Donc au ReleaseDC0, on indique que les prochains affichages se feront
      l'cran }
   nOutputDevice:=cteOutPrinterOneShoot;

   { rsolution en pixels / milimtes }
   vPixXSmm:=GetDeviceCaps(hDC0, HORZRES) / GetDeviceCaps(hDC0, HORZSIZE) / vScaleDPI;
   vPixYSmm:=GetDeviceCaps(hDC0, VERTRES) / GetDeviceCaps(hDC0, VERTSIZE) / vScaleDPI;

   if (hWindowTemp = 0) then begin
      GetPrinterSizes(hDC0);

      { Dfinit l'origine de la page dans le coin haut gauche }
      SetViewportOrgEx(hDC0, -pixPrinterBorderLeft, -pixPrinterBorderTop, nil);
      glpixXDec:=-pixPrinterBorderLeft;
      glpixYDec:=-pixPrinterBorderTop;
      end;

   end; { - if (hDC1 <> 0) then begin - }

if (hDC0 = 0) then begin
   if (nOutputDevice = cteOutScreen) then begin
      hDC0:=GetDC(hWnd0);

      if (hDC0 = 0) then begin
         ErrorWrite(msgGetDCis0, '');
         exit;
         end;
      end
   else if (nOutputDevice = cteOutPrinter) then begin
      { - Ouvre un Device Contexte en mmoire }
      {with pPrinter0^ do hDCPrn:=CreateDC(Driver, Device, Port, DevSettings);}
      hDCPrn:=0;  //?? hDCPrn:=pPrinter0^.GetDC;

      if (hDCPrn = 0) then begin
         ErrorWrite(msgCreateDCis0, '');
         exit;
         end;

      nRep:= GetDeviceCaps(hDCPrn, RasterCaps);
      if ((nRep and rc_BitBlt) <> rc_BitBlt) then begin
         { erreur, Device compatible n'accepte pas le raster scan }
         DeleteDC(hDCPrn);
         ErrorWrite(msgNoRasterPrinter, '');
         exit;
         end;

      { Cration du device contexte en mmoire }
      hDC0 := CreateCompatibleDC(hDCPrn);
      DeleteDC(hDCPrn);
      if (hDC0 = 0) then begin
         ErrorWrite(msgGetDCis0ForPrinter, '');
         exit;
         end;

      { on ne peut selectionner un bitMap que dans un device contexte en mmoire,
        c.f. commentaires de SelectObject }
      hOldBitMap := SelectObject(hDC0, hMemoryBitMap);
      end
   else begin { C'est une erreur d'arriver ici }
      ErrorWrite(msgBadUseOfSetNewDC0, '');
      exit;
      end;

   end;

{ ici hDC0 <> 0 }
{ """"""""""""" }

{ rsolution en pixels / milimtes }
vPixXSmm:=GetDeviceCaps(hDC0, HORZRES) / GetDeviceCaps(hDC0, HORZSIZE) / vScaleDPI;
vPixYSmm:=GetDeviceCaps(hDC0, VERTRES) / GetDeviceCaps(hDC0, VERTSIZE) / vScaleDPI;

{ Slection d'un font de le device contexte }
fontText.lfHeight :=Round(-sizeText * vPixYSmm * 25.4 / 72);

hFontText:=SelectFont(hDC0, fontText);  // from wFontManager

if (hFontText = 0) then begin
   ErrorWrite(msgFontNotSelected, '');
   end;

{ utilise les coordonnes pixels du Device Contexte }
SetMapMode(hDC0, MM_TEXT);

{ Si les lignes se dessinent en se combinant avec l'cran }
SetROP2(hDC0, ROP2Mode);

{ Innutile si MapMode = MM_TEXT.
SetWindowOrg(hDC0, 0,0);
SetViewPortOrg(hDC0, 0,0);
SetWindowExt(hDC0, 1,1);
SetViewPortExt(hDC0, 1,1);}

{ Dfini les conversions d'units, et les rgions de clipping }
UpdateWin12Pos;
nWinClip:=flaClipNone; { force le clipping }
SetGrClipRgn(flaClipWin2);

{ Pour mmoriser le style de ligne, et le style de mark }
pixWidthLineUsed:=Trunc(mmWidthLine * vPixXSmm);
if (pixWidthLineUsed < pixWidthLine) then pixWidthLineUsed:=pixWidthLine;

pixWidthMarkUsed:=Trunc(mmWidthMark * vPixXSmm);
if (pixWidthMarkUsed < pixWidthMark) then pixWidthMarkUsed:=pixWidthMark;

hLinePenNew :=CreatePen(styleLine, pixWidthLineUsed, colLine);
hMarkPenNew :=CreatePen(styleMark, pixWidthMarkUsed, colMark);
hFirstPenUsed :=SelectObject(hDC0, hLinePenNew);
hPenUsed    :=hLinePenNew;
end;  { - TGraphicBase.SetNewDC0 - }

function TGraphicBase.GetDC0:HDC;
{===============================}
{ retourn 0 si erreur, Le handle sur le device contexte sinon. }
begin { - TGraphicBase.GetDC0 - }
if (hDC0 <> 0) then begin
   GetDC0:=hDC0;
   exit;
   end;

SetNewDC0(0);
GetDC0:=hDC0;
end;  { - TGraphicBase.GetDC0 - }

function TGraphicBase.fIsDCOpen:boolean;
{======================================}
{ retourne TRUE si  hDC0 <> 0,  i.e. si le device contexte est ouvert }
begin { - TGraphicBase.fIsDCOpen - }
fIsDCOpen:=(hDC0 <> 0);
end;  { - TGraphicBase.fIsDCOpen - }

procedure TGraphicBase.CallPrintPage(pPrOut:PMsgPrintOut; var nErr:integer);
{=======================================================================}
{ Demande d'appeler la mthode  PrintPage  de  pPrOut^.
  Elle doit donc avoir t redfinie dans l'application qui utilise cette unit. }
begin { - TGraphicBase.CallPrintPage - }
if (pPrinter0 = nil) then begin nErr:=cteErrPPrinterNotInitialised; exit; end;
if (pPrOut    = nil) then begin nErr:=cteErrPPrintOutNotInitialised; exit; end;

//??pPrOut^.Banding := true;
//??pPrinter0^.Print(pWnd0, pPrOut);
nErr:=0;
end;  { - TGraphicBase.CallPrintPage - }

procedure TGraphicBase.ClearPrinterImage(var nErr:integer);
{=========================================================}
{ Efface l'image destine  l'imprimante.
  Si le device contexte d'imprimante est ouvert, ne fait que d'effacer l'image,
  sinon libre la mmoire utilise par le bit map.
  Valeurs possibles pour nErr : (en plus de 0)
  cteErrNoMemoryBitMap
-------------------------------------------------------------------------}
begin { - TGraphicBase.ClearPrinterImage - }
if (hMemoryBitMap = 0) then begin
   nErr:=cteErrNoMemoryBitMap; exit; end; { pas d'image }

if (hDC0 <> 0) and (nOutputDevice = cteOutPrinter) then begin
   { efface le BitMap }
   BitBlt( hDC0,  pixPrinterBorderLeft, pixPrinterBorderTop,
           pixPrinterTotWidth-1, pixPrinterTotHeight-1, 0, 0, 0, WHITENESS);
   end
else begin
   { libre la mmoire du BitMap d'impression sur imprimante }
   DeleteObject(hMemoryBitMap);
   hMemoryBitMap:=0;
   end;

nErr:=0;
end;  { - TGraphicBase.ClearPrinterImage - }

procedure TGraphicBase.OutputOnPrinter(var nErr:integer);
{=======================================================}
{ Indique que les prochains graphismes doivent tre fait sur l'imprimante,
  et ouvre le device contexte.
  Si il existait dj un  BitMap  associ  l'imprimante, on le garde sans le modifier,
  sinon on en cre un vierge.
  On peut utiliser :
  ReleaseDC0     pour le refermer,
  GetDC0         pour l'ouvrir denouveau,
  PrinterPrint   pour sortir effectivement sur l'imprimante le graphique
  OutputOnScreen pour l'affichage  l'cran
  Valeurs possibles pour nErr : (en plus de 0)
  cteErrNULLhDC0
  cteErrNULLPrinterDC
  cteErrCreateCompatibleBitMap
  cteErrSelectObjectBitMap
-------------------------------------------------------------------------}
begin { - TGraphicBase.OutputOnPrinter - }
nErr:=cteErrNULLhDC0;
end;  { - TGraphicBase.OutputOnPrinter - }

procedure TGraphicBase.PrinterPrint(var nErr:integer);
{=====================================================}
{ Pour sortir  l'imprimante les dessins qui ont t fait pour l'imprimante, c--d
  les dessins fait aprs l'appelle  OutputOnPrinter.
  Il n'est pas ncessaire que le device contexte soit ouvert ou soit celui de l'imprimante
  Valeurs possibles pour nErr : (en plus de 0)
  cteErrNoMemoryBitMap
  cteErrPPrinterNotInitialised
  cteErrPPrintOutNotInitialised
-------------------------------------------------------------------------}
begin { - TGraphicBase.PrinterPrint - }
nErr:=cteErrPPrinterNotInitialised;
end;  { - TGraphicBase.PrinterPrint - }

procedure TGraphicBase.ScreenPrint(var nErr:integer);
{===================================================}
{ Pour sortir  l'cran les dessins qui ont t fait pour l'imprimante, c--d
  les dessins fait aprs l'appelle  OutputOnPrinter.
  Valeurs possibles pour nErr : (en plus de 0)
  cteErrNoMemoryBitMap
-------------------------------------------------------------------------}
begin { - TGraphicBase.ScreenPrint - }
nErr:=cteErrNoMemoryBitMap;
end;  { - TGraphicBase.ScreenPrint - }

procedure TGraphicBase.SetPrinterTopHere;
{=======================================}
{ pour positionner le graphique sous la position courrante,
  Utile aprs un certain nombre de  Writeln0,  pour que le graphique
  soit en dessous du texte. }
begin { - TGraphicBase.SetPrinterTopHere - }
mmPrinterTop:= pixYPos / vPixYSmm;
UpdateWin12Pos;
end;  { - TGraphicBase.SetPrinterTopHere - }

procedure TGraphicBase.OutputOnScreen(var nErr:integer);
{======================================================}
{ Indique que les prochains graphismes doivent tre fait sur l'cran,
  ferme le device contexte d'imprimante, mais
  n'ouvre pas le device contexte de l'cran.
  Ceci est utilis gnralement aprs PrinterPrint, pour fermer
  l'affichage sur l'imprimante.
-------------------------------------------------------------------------}
begin { - TGraphicBase.OutputOnScreen - }
ReleaseDC0;
nOutputDevice:=cteOutScreen;
nErr:=0;
vScaleDPI:=1; { la prochaine fois, les dimensions ne seront pas modifies }
//??wPrn.glvScaleDPIPrn:=vScaleDPI; { indique la modif  la unit wPrn.pas }
end;  { - TGraphicBase.OutputOnScreen - }

procedure TGraphicBase.CloseOutputOnPrinter(var nErr:integer);
{============================================================}
{ Imprime ce qui a t prpar sur l'imprimante,
  Indique que les prochains graphismes doivent tre fait sur l'cran,
  ferme le device contexte d'imprimante, mais
  n'ouvre pas le device contexte de l'cran.
  Ceci est utilis gnralement aprs OutputOnPrinter.
  Est quivalent  :
  PrinterPrint(nErr);  OutputOnScreen(nErr);  ClearPrinterImage(nErr);
-------------------------------------------------------------------------}
var nErr1,nErr2, nErr3 : integer;
begin { - TGraphicBase.CloseOutputOnPrinter - }
PrinterPrint(nErr1);
OutputOnScreen(nErr2);
ClearPrinterImage(nErr3);
nErr:=nErr1 or nErr2 or nErr3;
end;  { - TGraphicBase.CloseOutputOnPrinter - }

procedure TGraphicBase.SetGrClipRgn(nWinClip1:integer);
{====================================================}
{ Pour utilisation interne, ne pas utiliser.
  Dfini le clipping a utiliser, et mmorise celui utilis }
var
  fDoReleaseDC     : boolean;

begin { - TGraphicBase.SetGrClipRgn - }
if not fClippingAuto then exit;  { clip. dfini par l'utilisateur }

if (nWinClip1 = nWinClip) then exit; { clipping dj OK }

fDoReleaseDC:=(hDC0 = 0);
if (hDC0 = 0) then begin
   if (GetDC0 = 0) then exit;  { traitement d'erreur ????? }
   end;

nWinClip:=nWinClip1;
case nWinClip of
   flaClipNone : SelectClipRgn(hDC0, 0);
   flaClipWin1 : SelectClipRgn(hDC0, hRgnWin1);
   flaClipWin2 : SelectClipRgn(hDC0, hRgnWin2);
   else SelectClipRgn(hDC0, 0);
   end; { - case - }

if fDoReleaseDC then ReleaseDC0;
end;  { - TGraphicBase.SetGrClipRgn - }

procedure TGraphicBase.SetClipping(nClipType:integer);
{====================================================}
{ Choix du clipping, flaClipAuto = par dfaut.
  nClipType =
  flaClipAuto => choix automatique du clipping
  flaClipNone => pas de clipping
  flaClipWin1 => clipping limit  la fentre 1
  flaClipWin2 => clipping limit  la fentre 2. }
begin { - TGraphicBase.SetClipping - }
fClippingAuto:=TRUE;

if (nClipType <> flaClipAuto) then begin
   SetGrClipRgn(nClipType);
   fClippingAuto:=FALSE;
   end;

end;  { - TGraphicBase.SetClipping - }

procedure TGraphicBase.ClearWin2( pixFromXBorder, mmFromXBorder,
{===============================} pixFromYBorder, mmFromYBorder:Float);
{ permet d'effacer le contenu de la fentre 2, sans effacer ce qui est
  trs proche des bords }
var
  fDoReleaseDC     : boolean;
  hBrush0          : HBrush;
  hRect            : HRgn;
  pixL,pixR,
  pixT,pixB        : integer;
  fOK              : boolean;

begin { - TGraphicBase.ClearWin2 - }
fDoReleaseDC:=(hDC0 = 0);
if (hDC0 = 0) then begin
   if (GetDC0 = 0) then exit;  { traitement d'erreur ????? }
   end;

SetGrClipRgn(flaClipWin2);

pixL:=Trunc(mmFromXBorder * vPixXSmm);
if (pixL < pixFromXBorder) then pixL:=Round(pixFromXBorder);
pixR:=-pixL;

pixT:=Trunc(mmFromYBorder * vPixYSmm);
if (pixT < pixFromYBorder) then pixT:=Round(pixFromYBorder);
pixB:=-pixT;

pixL:=pixL + pixX2Left;
pixT:=pixT + pixY2Top;
pixR:=pixR + pixX2Left + pixX2Width+1;
pixB:=pixB + pixY2Top  + pixY2Height+1;

hBrush0:=CreateSolidBrush(GetBkColor(hDC0));
hRect:=CreateRectRgn(pixL, pixT, pixR, pixB);

{if (hBrush = 0) then error}
{if (hRect = 0) then error}

fOK:=FillRgn(hDC0, hRect, hBrush0);  fOK:=fOK;

{if not fOK then error}

DeleteObject(hRect);
DeleteObject(hBrush0);
if fDoReleaseDC then ReleaseDC0;
end;  { - TGraphicBase.ClearWin2 - }

{ - Mthodes de positionnement }
{ ---------------------------- }
procedure TGraphicBase.MovePixTo(vPixX,vPixY:Float);
{==================================================}
{ dplacement absolu en pixels, utile pour revenir  un point de dpart. }
begin { - TGraphicBase.MovePixTo - }
vPixXPos:=vPixX;
vPixYPos:=vPixY;

pixXPos:=Round(vPixXPos);
pixYPos:=Round(vPixYPos);

{ - Si le Device contexte est ouvert, fait le dplacement dsir - }
if (hDC0 <> 0) then MoveToEx(hDC0, pixXPos, pixYPos, nil);
end;  { - TGraphicBase.MovePixTo - }

procedure TGraphicBase.MoveTo0(XX,YY:Float);
{=========================================}
{ dpl. absolu en coord. utilisateur. (Dans la fentre 2) }
begin { - TGraphicBase.MoveTo0 - }
vPixXPos:=vPixX0user + XX * vPixXSuser;
if (vPixXPos >  32767) then vPixXPos:= 32767;
if (vPixXPos < -32767) then vPixXPos:=-32767;

vPixYPos:=vPixY0user + YY * vPixYSuser;
if (vPixYPos >  32767) then vPixYPos:= 32767;
if (vPixYPos < -32767) then vPixYPos:=-32767;

pixXPos:=Round(vPixXPos);
pixYPos:=Round(vPixYPos);

{ - Si le Device contexte est ouvert, fait le dplacement dsir - }
if (hDC0 <> 0) then MoveToEx(hDC0, pixXPos, pixYPos, nil);
end;  { - TGraphicBase.MoveTo0 - }

procedure TGraphicBase.MoveMMRelTo(mmX,mmY:Float);
{================================================}
{ dpl. relatif en milimtres. (Dans la fentre 1) }
begin { - TGraphicBase.MoveMMRelTo - }
vPixXPos:=vPixXPos + mmX * vPixXSmm;
if (vPixXPos >  32767) then vPixXPos:= 32767;
if (vPixXPos < -32767) then vPixXPos:=-32767;

vPixYPos:=vPixYPos + mmY * vPixYSmm;
if (vPixYPos >  32767) then vPixYPos:= 32767;
if (vPixYPos < -32767) then vPixYPos:=-32767;

pixXPos:=Round(vPixXPos);
pixYPos:=Round(vPixYPos);

{ - Si le Device contexte est ouvert, fait le dplacement dsir - }
if (hDC0 <> 0) then MoveToEx(hDC0, pixXPos, pixYPos, nil);
end;  { - TGraphicBase.MoveMMRelTo - }

procedure TGraphicBase.MoveWin1To(relX,relY:Float);
{=================================================}
{ dpl. absolu dans la fentre 1 }
begin { - TGraphicBase.MoveWin1To - }
vPixXPos:=pixX1Left + relX * pixX1Width;
if (vPixXPos >  32767) then vPixXPos:= 32767;
if (vPixXPos < -32767) then vPixXPos:=-32767;

vPixYPos:=pixY1Top + relY * pixY1Height;
if (vPixYPos >  32767) then vPixYPos:= 32767;
if (vPixYPos < -32767) then vPixYPos:=-32767;

pixXPos:=Round(vPixXPos);
pixYPos:=Round(vPixYPos);

{ - Si le Device contexte est ouvert, fait le dplacement dsir - }
if (hDC0 <> 0) then MoveToEx(hDC0, pixXPos, pixYPos, nil);
end;  { - TGraphicBase.MoveWin1To - }

procedure TGraphicBase.GetXYPos(var XX,YY:Float);
{===============================================}
{ position en coordonnes utilisateur }
begin { - TGraphicBase.GetXYPos - }
XX:=vPixXPos;  YY:=vPixYPos;
end;  { - TGraphicBase.GetXYPos - }

procedure TGraphicBase.GetXYMMPos(var mmX,mmY:Float);
{===================================================}
{ position en milimtres }
begin { - TGraphicBase.GetXYMMPos - }
mmX:=vPixXPos / vPixXSmm;  mmY:=vPixYPos / vPixYSmm;
end;  { - TGraphicBase.GetXYMMPos - }

procedure TGraphicBase.GetXYWin1Pos(var relX,relY:Float);
{=======================================================}
{ position en coordonnes relatives  la fentre 1 }
begin { - TGraphicBase.GetXYWin1Pos - }
relX:=(vPixXPos - pixX1Left) / pixX1Width;
relY:=(vPixYPos - pixY1Top)  / pixY1Height;
end;  { - TGraphicBase.GetXYWin1Pos - }

{ - Conversion de coordonnes - }
{ ----------------------------- }
procedure TGraphicBase.MMToPix(mmX,mmY:Float; var vPixX,vPixY:Float);
{===================================================================}
{ conversion de milimtres en pixels. Utiliser Round si on veut des entiers }
{ dpl. relatif en milimtres. (Dans la fentre 1) }
begin { - TGraphicBase.MMToPix - }
vPixX:=mmX * vPixXSmm;
vPixY:=mmY * vPixYSmm;
end;  { - TGraphicBase.MMToPix - }

procedure TGraphicBase.MMToPixInt(mmX,mmY:Float; var pixX,pixY:integer);
{======================================================================}
{ conversion de milimtres en pixels. Utiliser Round si on veut des entiers }
{ dpl. relatif en milimtres. (Dans la fentre 1) }
begin { - TGraphicBase.MMToPixInt - }
pixX:=Round(mmX * vPixXSmm);
pixY:=Round(mmY * vPixYSmm);
end;  { - TGraphicBase.MMToPixInt - }

procedure TGraphicBase.PixToMM(vPixX,vPixY:Float; var mmX,mmY:Float);
{===================================================================}
{ conversion de pixels en milimtres. }
begin { - TGraphicBase.PixToMM - }
if (vPixXSmm <> 0) then mmX:=vPixX / vPixXSmm else mmX:=0;
if (vPixYSmm <> 0) then mmY:=vPixY / vPixYSmm else mmY:=0;
end;  { - TGraphicBase.PixToMM - }

procedure TGraphicBase.MMToUser(mmX,mmY:Float; var vX,vY:Float);
{==============================================================}
{ conversion de milimtres coord utilisateur. }
begin { - TGraphicBase.MMToUser - }
vX:= mmX * vPixXSmm / vPixXSuser;
vY:=-mmY * vPixYSmm / vPixYSuser;
end;  { - TGraphicBase.MMToUser - }

procedure TGraphicBase.UserToMM(vX,vY:Float; var mmX,mmY:Float);
{=============================================================}
begin { - TGraphicBase.UserToMM - }
mmX:= vX * vPixXSuser / vPixXSmm;
mmY:=-vY * vPixYSuser / vPixYSmm;
end;  { - TGraphicBase.UserToMM - }

procedure TGraphicBase.PixToUser(vPixX,vPixY:Float; var vX,vY:Float);
{===================================================================}
begin { - TGraphicBase.PixToUser - }
vX:=(vPixX - vPixX0user) / vPixXSuser;
vY:=(vPixY - vPixY0user) / vPixYSuser;
end;  { - TGraphicBase.PixToUser - }

procedure TGraphicBase.UserToPix(vX,vY:Float; var vPixX,vPixY:Float);
{===================================================================}
begin { - TGraphicBase.UserToPix - }
vPixX:=vPixX0user + vX * vPixXSuser;
vPixY:=vPixY0user + vY * vPixYSuser;
end;  { - TGraphicBase.UserToPix - }

procedure TGraphicBase.UserToPixInt(vX,vY:Float; var pixX,pixY:integer);
{======================================================================}
begin { - TGraphicBase.UserToPixInt - }
pixX:=Round(vPixX0user + vX * vPixXSuser);
pixY:=Round(vPixY0user + vY * vPixYSuser);
end;  { - TGraphicBase.UserToPixInt - }

{ - Mthodes des Lignes }
{ --------------------- }
procedure TGraphicBase.LinePixTo(vPixX,vPixY:Float);
{=================================================}
{ Dessin de ligne en pixels, clipping dans la fentre 1. }
begin { - TGraphicBase.LinePixTo - }
vPixXPos:=vPixX;
if (vPixXPos >  32767) then vPixXPos:= 32767;
if (vPixXPos < -32767) then vPixXPos:=-32767;

vPixYPos:=vPixY;
if (vPixYPos >  32767) then vPixYPos:= 32767;
if (vPixYPos < -32767) then vPixYPos:=-32767;

{ Slectionne le bon style de trait si ncessaire }
if (hPenUsed <> hLinePenNew) then begin
   SelectObject(hDC0, hLinePenNew);
   hPenUsed :=hLinePenNew;
   end;

if (hDC0 = 0) then begin
   if (GetDC0 = 0) then exit;  { traitement d'erreur ????? }
   SetGrClipRgn(flaClipWin1);

   MoveToEx(hDC0, pixXPos, pixYPos, nil);

   pixXPos:=Round(vPixXPos);
   pixYPos:=Round(vPixYPos);

   LineTo(hDC0, pixXPos, pixYPos);
   SetPixel(hDC0, pixXPos, pixYPos, colLine);  { pour terminer le dessin de lineTo }
   ReleaseDC0;
   exit;
   end;

pixXPos:=Round(vPixXPos);
pixYPos:=Round(vPixYPos);

SetGrClipRgn(flaClipWin1);

LineTo(hDC0, pixXPos, pixYPos);
SetPixel(hDC0, pixXPos, pixYPos, colLine);  { pour terminer le dessin de lineTo }
end;  { - TGraphicBase.LinePixTo - }

procedure TGraphicBase.LineTo0(XX,YY:Float);
{==========================================}
{ en coord. utilisateur }
begin { - TGraphicBase.LineTo0 - }
vPixXPos:=vPixX0user + XX * vPixXSuser;
if (vPixXPos >  32767) then vPixXPos:= 32767;
if (vPixXPos < -32767) then vPixXPos:=-32767;

vPixYPos:=vPixY0user + YY * vPixYSuser;
if (vPixYPos >  32767) then vPixYPos:= 32767;
if (vPixYPos < -32767) then vPixYPos:=-32767;

{ Slectionne le bon style de trait si ncessaire }
if (hPenUsed <> hLinePenNew) then begin
   SelectObject(hDC0, hLinePenNew);
   hPenUsed :=hLinePenNew;
   end;

if (hDC0 = 0) then begin
   if (GetDC0 = 0) then exit;  { erreur }
   MoveToEx(hDC0, pixXPos, pixYPos, nil);

   pixXPos:=Round(vPixXPos);
   pixYPos:=Round(vPixYPos);

   SetGrClipRgn(flaClipWin2);

   LineTo(hDC0, pixXPos, pixYPos);
   ReleaseDC0;
   exit;
   end;

pixXPos:=Round(vPixXPos);
pixYPos:=Round(vPixYPos);

SetGrClipRgn(flaClipWin2);

LineTo(hDC0, pixXPos, pixYPos);
end;  { - TGraphicBase.LineTo0 - }

procedure TGraphicBase.Line0(X1,Y1,X2,Y2:Float);
{=============================================}
begin { - TGraphicBase.Line0 - }
if (hDC0 = 0) then begin
   if (GetDC0 = 0) then exit; {erreur}
   MoveTo0(X1,Y1);
   LineTo0(X2,Y2);
   ReleaseDC0;
   exit;
   end;

MoveTo0(X1,Y1);
LineTo0(X2,Y2);
end;  { - TGraphicBase.Line0 - }

procedure TGraphicBase.PolyLine0(nbData:integer; var aX,aY:array of Float);
{========================================================================}
var
  fDoReleaseDC     : boolean;
  nn               : integer;

begin { - TGraphicBase.PolyLine0 - }
fDoReleaseDC:=(hDC0 = 0);

{ Ouvre le device contexte si ncessaire }
if (hDC0 = 0) then begin
   if (GetDC0 = 0) then exit;  { traitement d'erreur ????? }
   end;

MoveTo0(aX[0], aY[0]);
for nn:=1 to nbData-1 do LineTo0(aX[nn], aY[nn]);

if fDoReleaseDC then ReleaseDC0;
end;  { - TGraphicBase.PolyLine0 - }

procedure TGraphicBase.LineMMRelTo(mmX,mmY:Float);
{================================================}
{ Dessin de ligne par dpl. relatif en milimtres. (Dans la fentre 1) }
begin { - TGraphicBase.LineMMRelTo - }
vPixXPos:=vPixXPos + mmX * vPixXSmm;
if (vPixXPos >  32767) then vPixXPos:= 32767;
if (vPixXPos < -32767) then vPixXPos:=-32767;

vPixYPos:=vPixYPos + mmY * vPixYSmm;
if (vPixYPos >  32767) then vPixYPos:= 32767;
if (vPixYPos < -32767) then vPixYPos:=-32767;

{ Slectionne le bon style de trait si ncessaire }
if (hPenUsed <> hLinePenNew) then begin
   SelectObject(hDC0, hLinePenNew);
   hPenUsed :=hLinePenNew;
   end;

if (hDC0 = 0) then begin
   if (GetDC0 = 0) then exit;  { traitement d'erreur ????? }
   SetGrClipRgn(flaClipWin1);

   MoveToEx(hDC0, pixXPos, pixYPos, nil);

   pixXPos:=Round(vPixXPos);
   pixYPos:=Round(vPixYPos);

   LineTo(hDC0, pixXPos, pixYPos);
   ReleaseDC0;
   exit;
   end;

pixXPos:=Round(vPixXPos);
pixYPos:=Round(vPixYPos);

SetGrClipRgn(flaClipWin1);

LineTo(hDC0, pixXPos, pixYPos);
end;  { - TGraphicBase.LineMMRelTo - }

procedure TGraphicBase.PolyLineMMRel(nbData:integer; vHomot:Float; var ammX,ammY:array of Float);
{===============================================================================================}
{ Fait une succession de lignes relatives au point prcdent, les longueurs sont
  donnes en milimtres * vHomot. }
var
  fDoReleaseDC     : boolean;
  nn               : integer;

begin { - TGraphicBase.PolyLineMMRel - }
{ Slectionne le bon style de trait si ncessaire }
if (hPenUsed <> hLinePenNew) then begin
   SelectObject(hDC0, hLinePenNew);
   hPenUsed :=hLinePenNew;
   end;

fDoReleaseDC:=(hDC0 = 0);

{ Ouvre le device contexte si ncessaire }
if (hDC0 = 0) then begin
   if (GetDC0 = 0) then exit;  { traitement d'erreur ????? }
   SetGrClipRgn(flaClipWin1);

   MoveToEx(hDC0, pixXPos, pixYPos, nil);
   end;

{ dfini le bon clipping si ncessaire }
SetGrClipRgn(flaClipWin1);

{ fait une succession de lignes relatives au point prcdent }
for nn:=0 to nbData-1 do begin
   vPixXPos:=vPixXPos + ammX[nn] * vPixXSmm * vHomot;
   if (vPixXPos >  32767) then vPixXPos:= 32767;
   if (vPixXPos < -32767) then vPixXPos:=-32767;

   vPixYPos:=vPixYPos + ammY[nn] * vPixYSmm * vHomot;
   if (vPixYPos >  32767) then vPixYPos:= 32767;
   if (vPixYPos < -32767) then vPixYPos:=-32767;

   pixXPos:=Round(vPixXPos);
   pixYPos:=Round(vPixYPos);

   LineTo(hDC0, pixXPos, pixYPos);
   end;  { - for nn:=1 to nbData do begin - }

if fDoReleaseDC then ReleaseDC0;
end;  { - TGraphicBase.PolyLineMMRel - }

procedure TGraphicBase.LineWin1To(relX,relY:Float);
{=================================================}
{ Dessin de ligne par dpl. relatif  la fentre 1.
  (0,0) = haut gauche,  (1,1) = bas droit
------------------------------------------------------}
begin { - TGraphicBase.LineWin1To - }
vPixXPos:=pixX1Left + relX * pixX1Width;
if (vPixXPos >  32767) then vPixXPos:= 32767;
if (vPixXPos < -32767) then vPixXPos:=-32767;

vPixYPos:=pixY1Top + relY * pixY1Height;
if (vPixYPos >  32767) then vPixYPos:= 32767;
if (vPixYPos < -32767) then vPixYPos:=-32767;

{ Slectionne le bon style de trait si ncessaire }
if (hPenUsed <> hLinePenNew) then begin
   SelectObject(hDC0, hLinePenNew);
   hPenUsed :=hLinePenNew;
   end;

if (hDC0  = 0) then begin
   if (GetDC0 = 0) then exit;  { traitement d'erreur ????? }
   SetGrClipRgn(flaClipWin1);

   MoveToEx(hDC0, pixXPos, pixYPos, nil);

   pixXPos:=Round(vPixXPos);
   pixYPos:=Round(vPixYPos);

   LineTo(hDC0, pixXPos, pixYPos);
   ReleaseDC0;
   exit;
   end;

pixXPos:=Round(vPixXPos);
pixYPos:=Round(vPixYPos);

SetGrClipRgn(flaClipWin1);

LineTo(hDC0, pixXPos, pixYPos);
end;  { - TGraphicBase.LineWin1To - }

procedure TGraphicBase.GetLineAttribRec( var recLineAt:TypeLineAttrib);
{==================================================================}
begin { - TGraphicBase.GetLineAttribRec - }
with RecLineAt do begin
   style2     :=styleLine;
   pixWidth2  :=pixWidthLine;
   mmWidth2   :=mmWidthLine;
   col2       :=colLine;
   end;
end;  { - TGraphicBase.GetLineAttribRec - }

procedure TGraphicBase.SetLineAttrib(
{===================================}
   style2     : TStyle;
   pixWidth2  : integer;
   mmWidth2   : Float;
   col2       : TColorRef
   );
{ Change l'attribut de ligne }
var
  hOldPen,
  hNewPen          : HPen;
  fOK              : boolean;

begin { - TGraphicBase.SetLineAttrib - }
styleLine    :=style2;
pixWidthLine :=pixWidth2;
mmWidthLine  :=mmWidth2;
colLine      :=col2;

{ Si le device contexte est ouvert, il faut changer l'attribut de ligne }
if (hDC0 <> 0) then begin
   pixWidthLineUsed:=Trunc(mmWidthLine * vPixXSmm);
   if (pixWidthLineUsed < pixWidthLine) then pixWidthLineUsed:=pixWidthLine;

   { cre un nouveau style de ligne, et efface l'ancien }
   hNewPen :=CreatePen(styleLine, pixWidthLineUsed, colLine);

   if (hPenUsed = hLinePenNew) then begin
      { le style de ligne utilis actuellement est celui que l'on
        va changer. Donc il faut le dslectionner }
      hOldPen :=SelectObject(hDC0, hNewPen);  { ici hOldPen = hLinePenNew }  hOldPen:=hOldPen;
      hPenUsed :=hNewPen;
      end;

   { efface l'ancien }
   fOK:=DeleteObject(hLinePenNew); fOK:=fOK;

{   if not fOK then error...;}

   { mmorise le nouveau style utilis }
   hLinePenNew:=hNewPen;
   end; { - if (hDC0 <> 0) then begin - }
end;  { - TGraphicBase.SetLineAttrib - }

procedure TGraphicBase.SetLineAttrib2(
{====================================}
  style2:TStyle;  pixWidth2:integer;
  mmWidth2:Float; col2:TColorRef; var recLineAt:TypeLineAttrib);
{ Change l'attribut de ligne, en retournant l'attibut prcdent }
begin { - TGraphicBase.SetLineAttrib2 - }
GetLineAttribRec(recLineAt);
SetLineAttrib(style2, pixWidth2, mmWidth2, col2);
end;  { - TGraphicBase.SetLineAttrib2 - }

procedure TGraphicBase.SetLineAttribRec(var recLineAt:TypeLineAttrib);
{==================================================================}
{ Change l'attribut de ligne, en donnant un Record, comme celui obtenu
  par GetLineAttribRec. }
begin { - TGraphicBase.SetLineAttribRec - }
with RecLineAt do SetLineAttrib(style2, pixWidth2, mmWidth2, col2);
end;  { - TGraphicBase.SetLineAttribRec - }

procedure TGraphicBase.DrawBox(nPts:integer; x1,y1,x2,y2,x3,y3,x4,y4:Float);
{==========================================================================}
{ nPts = 2  5
  Dessine de nPts-1 lignes de (x1,y1)  (x4,y4),
  avec retour  (x1,y1) si  nPts >= 5.
  Les coordonnes sont ceux utilisateur, dans la fentre 2,
  la couleur, le style et width des lignes sont utilis.
-------------------------------------------------------------------------}
begin { - TGraphicBase.DrawBox - }
if (nPts < 2) then exit;
Line0(x1,y1,x2,y2);
if (nPts < 3) then exit;
LineTo0(x3,y3);
if (nPts < 4) then exit;
LineTo0(x4,y4);
if (nPts < 5) then exit;
LineTo0(x1,y1);
end;  { - TGraphicBase.DrawBox - }

procedure TGraphicBase.BorderWin1( Style1:TStyle; pixWidth1:integer;
                                   mmWidth1:Float; Color1:TColorRef);
{===============================================================================}
{ Sert  mettre un cadre autour de la fentre 1 }
var
  fDoReleaseDC     : boolean;
  OldPen,ThePen    : HPen;
  pixWidthUsed,
  pixXLeft,
  pixXRight,
  pixYTop,
  pixYBottom       : integer;

begin { - TGraphicBase.BorderWin1 - }
fDoReleaseDC:=(hDC0 = 0);

if (hDC0 = 0) then if (GetDC0 = 0) then exit;  { traitement d'erreur ????? }

{ enlve le clipping }
SetGrClipRgn(flaClipNone);

pixWidthUsed:=Trunc(mmWidth1 * vPixXSmm);
if (pixWidthUsed < pixWidth1) then pixWidthUsed:=pixWidth1;

ThePen := CreatePen(Style1, pixWidthUsed, Color1);
OldPen := SelectObject(hDC0, ThePen);

{FrameRgn(hDC0, hRgnWin1, Brush: HBrush; Width, Height: Integer): Bool;}

pixXLeft   :=pixX1Left-1  - (pixWidthUsed div 2);
pixXRight  :=pixX1Left    + pixX1Width+1  + (pixWidthUsed div 2);
pixYTop    :=pixY1Top-1   - pixWidthUsed div 2;
pixYBottom :=pixY1Top     + pixY1Height+1 + (pixWidthUsed div 2);

MoveToEx(hDC0, pixXLeft , pixYTop, nil);
LineTo(hDC0, pixXRight, pixYTop);
LineTo(hDC0, pixXRight, pixYBottom);
LineTo(hDC0, pixXLeft , pixYBottom);
LineTo(hDC0, pixXLeft , pixYTop);

SelectObject(hDC0, OldPen);
DeleteObject(ThePen);

if fDoReleaseDC then ReleaseDC0;
end;  { - TGraphicBase.BorderWin1 - }

procedure TGraphicBase.BorderWin2( Style1:TStyle; pixWidth1:integer;
{================================} mmWidth1:Float; Color1:TColorRef);
{ Sert  mettre un cadre autour de la fentre 2 }
var
  fDoReleaseDC     : boolean;
  OldPen,ThePen    : HPen;
  pixWidthUsed,
  pixXLeft,
  pixXRight,
  pixYTop,
  pixYBottom       : integer;

begin { - TGraphicBase.BorderWin2 - }
fDoReleaseDC:=(hDC0 = 0);

if (hDC0 = 0) then if (GetDC0 = 0) then exit;  { traitement d'erreur ????? }

SetGrClipRgn(flaClipWin2);

pixWidthUsed:=Trunc(mmWidth1 * vPixXSmm);
if (pixWidthUsed < pixWidth1) then pixWidthUsed:=pixWidth1;

ThePen := CreatePen(Style1, pixWidthUsed, Color1);
OldPen := SelectObject(hDC0, ThePen);

pixXLeft   :=pixX2Left  + pixWidthUsed div 2;
pixXRight  :=pixX2Left  + pixX2Width  - (pixWidthUsed) div 2;
pixYTop    :=pixY2Top   + pixWidthUsed div 2;
pixYBottom :=pixY2Top   + pixY2Height - (pixWidthUsed) div 2;

MoveToEx(hDC0, pixXLeft , pixYTop, nil);
LineTo(hDC0, pixXRight, pixYTop);
LineTo(hDC0, pixXRight, pixYBottom);
LineTo(hDC0, pixXLeft , pixYBottom);
LineTo(hDC0, pixXLeft , pixYTop);

SelectObject(hDC0, OldPen);
DeleteObject(ThePen);

if fDoReleaseDC then ReleaseDC0;
end;  { - TGraphicBase.BorderWin2 - }

{ - Mthodes des Marques }
{ ---------------------- }
procedure TGraphicBase.Mark0(nMark:integer);
{==================================================================}
{ Place une marque :  +  x  *  []  #  O  v  }
const
apixXPlus    : array[1.. 4] of integer = (-1, 0, 0, 0);
apixYPlus    : array[1.. 4] of integer = ( 0, 0, 1,-1);
apixXCross   : array[1.. 4] of integer = (-1, 0,-1, 1);
apixYCross   : array[1.. 4] of integer = (-1, 0, 1,-1);
apixXStar    : array[1..10] of integer = (-1, 0, 0, 0, 0, 1,-1, 0,-1, 1);
apixYStar    : array[1..10] of integer = ( 0, 0, 1,-1, 0, 1,-1, 0, 1,-1);
apixXSquare  : array[1.. 4] of integer = ( 1,-1,-1, 1);
apixYSquare  : array[1.. 4] of integer = (-1,-1, 1, 1);
apixXDiamond : array[1.. 4] of integer = ( 0,-1, 0, 1);
apixYDiamond : array[1.. 4] of integer = (-1, 0, 1, 0);

var
  pixLen1,pixLen2,
  nn,nbPts         : integer;
  vPixXMem,vPixYMem: Float;
  vPixLen1,vPixLen2,
  x0,x1,dx,
  y0,y1,dy         : Float;

   procedure DrawMark(pixDX,pixDY,nLen,nbPts:integer; apixDX,apixDY:array of integer);
   {=================================================================================}
   { Se dplacement relatif de pixDX, pixDY pixels, puis dessine la figure
     dfini par la suite de points relatif au point de dpart :
     nLen*apixDX[nn], nLen*apixDY[nn], nn:=0..nbPts-1 }
   var
     pixXOrg,pixYOrg,
     nn               : integer;
   begin { - DrawMark - }
   { Position du centre, au pixel en entier }
   pixXOrg:=pixXPos;
   pixYOrg:=pixYPos;

   { test que position dessinable }
   if (abs(pixXOrg) > 30000) or (abs(pixYOrg) > 30000) then exit;

   { positionnement initial }
   MovePixTo(pixXOrg+pixDX, pixYOrg+pixDY);
   for nn:=0 to nbPts-1 do LinePixTo(pixXOrg+nLen*apixDX[nn], pixYOrg+nLen*apixDY[nn]);
   end;  { - DrawMark - }

begin { - TGraphicBase.Mark0 - }
{==============================}
{ Slectionne le bon style de trait si ncessaire }
if (hPenUsed <> hMarkPenNew) then begin
   {hPenUsed :=hMarkPenNew; fait en fin de procdure }
   SelectObject(hDC0, hMarkPenNew);
   end;

{ pour utiliser LineMMRelTo, avec le bon style de trait, on indique un autre
  Pen que celui utilis rellement }
hPenUsed :=hLinePenNew;

vPixXMem:=vPixXPos;
vPixYMem:=vPixYPos;

{ pour des dplacement en pixels entier }
MMToPix(0.5*mmSizeMark, 0.5*mmSizeMark, vPixLen1, vPixLen2);

pixLen1:=Round(vPixLen1);  pixLen1:=pixLen1;
pixLen2:=Round(vPixLen2);  pixLen2:=pixLen2;

case nMark of
  markPlus     : DrawMark(pixLen1,       0, pixLen1, 4, apixXPlus, apixYPlus);
  markCross    : DrawMark(pixLen1, pixLen1, pixLen1, 4, apixXCross, apixYCross);
  markStar     : DrawMark(pixLen1,       0, pixLen1,10, apixXStar, apixYStar);
  markSquare   : DrawMark(pixLen1, pixLen1, pixLen1, 4, apixXSquare, apixYSquare);
  markDiamond  : DrawMark(pixLen1,       0, pixLen1, 4, apixXDiamond, apixYDiamond);
  markCircle   : begin
                 nbPts:=16;
                 x0:=0.5;
                 y0:=0.0;
                 MoveMMRelTo(x0*mmSizeMark,y0*mmSizeMark);
                 for nn:=1 to nbPts do begin
                    x1:=0.5*cos(nn*2*Pi/nbPts);
                    dx:=x1-x0;
                    x0:=x1;
                    y1:=0.5*sin(nn*2*Pi/nbPts);
                    dy:=y1-y0;
                    y0:=y1;
                    LineMMRelTo(dx*mmSizeMark,dy*mmSizeMark);
                    end;
                 end;
  markTriangle : begin
                 MoveMMRelTo(0,-0.5*mmSizeMark);

                 LineMMRelTo(-sqrt(3)*0.25*mmSizeMark, 0.75*mmSizeMark);
                 LineMMRelTo( sqrt(3)*0.50*mmSizeMark, 0);
                 LineMMRelTo(-sqrt(3)*0.25*mmSizeMark,-0.75*mmSizeMark);
                 end;
  else           DrawMark(pixLen1,       0, pixLen1, 4, apixXPlus, apixYPlus);
  end; { - case nMark of - }

{ revient au point de dpart }
MovePixTo(vPixXMem, vPixYMem);

{ pen rellement utilis }
hPenUsed :=hMarkPenNew;
end;  { - TGraphicBase.Mark0 - }

procedure TGraphicBase.MarkTo(X1,Y1:Float; nMark:integer);
{==================================================================}
{ Se positionne avant de placer une marque }
begin { - TGraphicBase.MarkTo - }
MoveTo0(X1,Y1);
Mark0(nMark);
end;  { - TGraphicBase.MarkTo - }

procedure TGraphicBase.PolyMark(nbData,nMark:integer; var aX,aY:array of Float);
{==================================================================}
var nn  : integer;
begin { - TGraphicBase.PolyMark - }
for nn:=0 to nbData-1 do MarkTo(aX[nn], aY[nn], nMark);
end;  { - TGraphicBase.PolyMark - }

procedure TGraphicBase.GetMarkAttribRec(var recMarkAt:TypeMarkAttrib);
{====================================================================}
{ Lit l'attribut de Marque actuelle }
begin { - TGraphicBase.GetMarkAttribRec - }
with RecMarkAt do begin
   style2     :=styleMark;
   pixWidth2  :=pixWidthMark;
   mmWidth2   :=mmWidthMark;
   col2       :=colMark;
   mmSize2    :=mmSizeMark;
   end; { - with RecMarkAt do begin - }
end;  { - TGraphicBase.GetMarkAttribRec - }

procedure TGraphicBase.SetMarkAttrib(
{===================================}
   style2:TStyle;  pixWidth2:integer; mmWidth2:Float;
   col2:TColorRef; mmSize2:Float);
{ Change l'attribut de Marque }
var
  hOldPen,
  hNewPen          : HPen;
  fOK              : boolean;

begin { - TGraphicBase.SetMarkAttrib - }
styleMark    :=style2;
pixWidthMark :=pixWidth2;
mmWidthMark  :=mmWidth2;
colMark      :=col2;
mmSizeMark   :=mmSize2;

{ Si le device contexte est ouvert, il faut changer l'attribut de ligne }
if (hDC0 <> 0) then begin
   pixWidthMarkUsed:=Trunc(mmWidthMark * vPixXSmm);
   if (pixWidthMarkUsed < pixWidthMark) then pixWidthMarkUsed:=pixWidthMark;

   { cre un nouveau style de marque, et efface l'ancien }
   hNewPen :=CreatePen(styleMark, pixWidthMarkUsed, colMark);

   if (hPenUsed = hMarkPenNew) then begin
      { le style de ligne utilis actuellement est celui que l'on
        va changer. Donc il faut le dslectionner }
      hOldPen:=SelectObject(hDC0, hNewPen); hOldPen:=hOldPen;
      { ici hOldPen = hMarkPenNew }
      end;

   { efface l'ancien }
   fOK:=DeleteObject(hMarkPenNew);  fOK:=fOK;

{   if not fOK then error...;}

   { mmorise le nouveau style utilis }
   hMarkPenNew:=hNewPen;
   end; { - if (hDC0 <> 0) then begin - }
end;  { - TGraphicBase.SetMarkAttrib - }

procedure TGraphicBase.SetMarkAttribRec(var recMarkAt:TypeMarkAttrib);
{==================================================================}
{ Change l'attribut de Mark, en donnant un Record, comme celui obtenu
  par GetLineAttribRec. }
begin { - TGraphicBase.SetMarkAttribRec - }
with recMarkAt do SetMarkAttrib(style2, pixWidth2, mmWidth2, colMark, mmSizeMark);
end;  { - TGraphicBase.SetMarkAttribRec - }

{ - Mthodes des Textes }
{ ---------------------- }

procedure TGraphicBase.WriteMove(rX,rY:Float; strS:string; flags:word);
{=====================================================================}
{ Dplacement relatif avant d'afficher le string, et indique dans flags
  le type de dplacement dsir aprs l'affichage.
  Flags <- 1 => en fin de string,
  Flags <- 2 => sous le string.
  Flags <- 4 => criture verticale }
var
  fDoReleaseDC     : boolean;
  pixMem,
  pixDX,pixDY      : integer;
  lExtend          : longint;
  lfEscapementMem  :integer;
  colOld           : TColorRef;
  fVert            : boolean;  { criture verticale }
  fontSymbol       : TLogFont;
  rSize            : Size;

   procedure ChangeFontSize(nExposant:integer);
   {==========================================}
   { change la taille des fonts en fonction des exposants et indices }
   begin { - ChangeFontSize - }
   if (nExposant <> 0) then begin
      fontText.lfHeight :=Round(-0.8*sizeText * vPixYSmm * 25.4 / 72);
      fontSymbol.lfHeight :=Round(-0.8*sizeText * vPixYSmm * 25.4 / 72);
      end
   else begin
      fontText.lfHeight :=Round(-sizeText * vPixYSmm * 25.4 / 72);
      fontSymbol.lfHeight :=Round(-sizeText * vPixYSmm * 25.4 / 72);
      end;

   SelectFont(hDC0, fontText);  // from wFontManager
   end;  { - ChangeFontSize - }

   function TextOutWithSymbol(fTxt:boolean):longint;
   {================================================}
   { fTxt = TRUE => affiche le texte,
            sinon ne fait que de mesurer le texte Extend.
     Retourne la taille en pixels prise par le string  l'affichage. }
   var
     pixXP0,
     pixYP0,
     pixXP,
     pixYP,
     nPos1,
     nBeg,
     nLen
                   : integer;
     nExposant
     { niveau dans la ligne.
        0 = normal
        1 = exposant
       -1 = indice }
     ,pixExpInd    { dcalage en pixels pour indices et exposant }
                   : integer;
     rSize            : Size;

   begin { - TextOutWithSymbol - }
   pixXP0 :=pixXPos - Round(rX*pixDX);
   pixYP0 :=pixYPos - Round(rY*pixDY);
   pixXP  :=pixXP0;
   pixYP  :=pixYP0;
   nBeg :=1;
   nLen :=length(strS);
   nExposant:=0;

   { dcalage des exposants et indices }
   GetTextExtentPoint32(hDC0, @strS[nBeg], nLen, rSize);
   pixExpInd :=rSize.cy div 2;

   { boucle, pour affichage, en cherchant les caractres spciaux }
   repeat
      nPos1:=Pos('\', Copy(strS,nBeg, 250));

      if (nPos1 = 0) then begin
         { plus de caractres spciaux }
{         SelectFont(hDC0, fontText); dj, par dfaut }
         nLen:=length(strS)+1 - nBeg;
         if fTxt then TextOut(hDC0, pixXP, pixYP, @strS[nBeg], nLen);
         fontText.lfHeight :=Round(-sizeText * vPixYSmm * 25.4 / 72);
         fontSymbol.lfHeight :=Round(-sizeText * vPixYSmm * 25.4 / 72);

         { retourne la dimension du texte affich }
         GetTextExtentPoint32(hDC0, @strS[nBeg], nLen, rSize);
         if fVert then begin
            pixYP:=pixYP - rSize.cx;
            pixXP:=pixXP + rSize.cy;
            TextOutWithSymbol:=MAKELong(pixYP0 - pixYP, pixXP - pixXP0);
            end
         else begin
             pixXP:=pixXP + rSize.cx;
             pixYP:=pixYP - rSize.cy;
             TextOutWithSymbol:=MAKELong(pixXP - pixXP0, pixYP0 - pixYP);
             end;
         exit;
         end;

      nLen:=nPos1 - 1;
      if (nLen > 0) then begin
         { imprime le texte normal }
{         SelectFont(hDC0, fontText);  dj par dfaut }
         if fTxt then TextOut(hDC0, pixXP, pixYP, @strS[nBeg], nLen);
         GetTextExtentPoint32(hDC0, @strS[nBeg], nLen, rSize);
         if fVert then pixYP:=pixYP - rSize.cx
         else          pixXP:=pixXP + rSize.cx;
         end;

      { Suite du texte  }
      nBeg:=nBeg + nLen+1;  { strS[nBeg-1] = '\'  }

      if (strS[nBeg] = 'g') and (strS[nBeg+1] = '(') then begin
         { dbut du texte avec symboles }
         nBeg:=nBeg + 2;  { saute le '\g(' }

         { imprime le texte de caractres spciaux }
         nPos1:=Pos(')', Copy(strS,nBeg,250));
         if (nPos1 > 0) then nLen:=nPos1-1;

         if (nLen > 0) then begin
            SelectFont(hDC0, fontSymbol);  // from wFontManager
            if fTxt then TextOut(hDC0, pixXP, pixYP, @strS[nBeg], nLen);
            GetTextExtentPoint32(hDC0, @strS[nBeg], nLen, rSize);
            if fVert then pixYP:=pixYP - rSize.cx
            else          pixXP:=pixXP + rSize.cx;

            { reslectionne la font par dfaut }
            SelectFont(hDC0, fontText);  // from wFontManager
            end;

         { Suite du texte  }
         nBeg:=nBeg + nLen+1;  { strS[nBeg-1] = ')' }
         end
      else if (strS[nBeg] = '+') then begin
         { exposant }
         inc(nExposant);
         nBeg:=nBeg + 1;  { saute le '\+' }
         if fVert then pixXP:=pixXP - pixExpInd
         else          pixYP:=pixYP - pixExpInd;
         if (nExposant > 0) then begin
            { Sinon l'exposant est trop haut }
            if fVert then pixXP:=pixXP + pixExpInd div 2
            else          pixYP:=pixYP + pixExpInd div 2;
            end;

         ChangeFontSize(nExposant);
         end
      else if (strS[nBeg] = '-') then begin
         { indice }
         if (nExposant > 0) then begin
            { ramne l'exposant au niveau normal }
            if fVert then pixXP:=pixXP - pixExpInd div 2
            else          pixYP:=pixYP - pixExpInd div 2;
            end;

         dec(nExposant);
         nBeg:=nBeg + 1;  { saute le '\-' }
         if fVert then pixXP:=pixXP + pixExpInd
         else          pixYP:=pixYP + pixExpInd;

         ChangeFontSize(nExposant);
         end
      else begin { par une commande spciale }
          { strS[nBeg] = caractre qui suit le '\' }
{         SelectFont(hDC0, fontText); dj, par dfaut }
         { Si deux  '\'  se suivent immdiatement, ignore le premier. }
         if (strS[nBeg] = '\') then inc(nBeg);

         { affiche le '\' }
         if fTxt then TextOut(hDC0, pixXP, pixYP, @strS[nBeg-1], 1);
         GetTextExtentPoint32(hDC0, @strS[nBeg], nLen, rSize);
         if fVert then pixYP:=pixYP - rSize.cx
         else          pixXP:=pixXP + rSize.cx;
         end;
      until FALSE; { for ever, on sort avec un exit; }
   end;  { - TextOutWithSymbol - }

begin { - TGraphicBase.WriteMove - }
{==================================}
{ mmorise l'angle d'criture relativement  l'horizontale }
lfEscapementMem:=fontText.lfEscapement;

{ pour dfinir une font de symbole mathmatiques. }
with fontSymbol do begin
    lfHeight           :=fontText.lfHeight;
    lfWidth            :=0;
    lfEscapement       :=0; { en 1/10 de degrs }
    lfOrientation      :=0;   { ignor }
    lfWeight           :=FW_BOLD;  {FW_DONTCARE;}
    lfItalic           :=0;
    lfUnderline        :=0;
    lfStrikeOut        :=0;
    lfCharSet          :=ANSI_CHARSET;
    lfOutPrecision     :=OUT_DEFAULT_PRECIS;
    lfClipPrecision    :=CLIP_DEFAULT_PRECIS;
    lfQuality          :=DEFAULT_QUALITY;
    lfPitchAndFamily   :=DEFAULT_PITCH or FF_SWISS;
    StrPCopy(@lfFaceName, 'Symbol');
   end; { - with fontSymbol do begin - }

fDoReleaseDC:=(hDC0 = 0);
if (hDC0 = 0) then begin
   if (GetDC0 = 0) then exit; {error}
   {MoveToEx(hDC0, pixXPos, pixYPos, nil);}
   end;

fVert:=((Flags and 4) <> 0);
if fVert then begin
   { pour criture verticale }
   fontText.lfEscapement:=900;
   fontSymbol.lfEscapement:=900;
   SelectFont(hDC0, fontText);  // from wFontManager
   end;

if (strS <> '') then begin
   { taille en pixels du string }
   pixDX:=0;
   pixDY:=0;
//   GetTextExtentPoint32(hDC0, @strS[1], length(strS), rSize);
   lExtend:=TextOutWithSymbol(FALSE);
   rSize.cx:=LOWORD(lExtend);
   rSize.cy:=HIWORD(lExtend);
   end
else begin
  GetTextExtentPoint32(hDC0, 'W', 1, rSize);
  end;

pixDX:=rSize.cx;
pixDY:=rSize.cy;

if (strS = '') then pixDX:=0;

if ((Flags and 4) <> 0) then begin
   { criture verticale }
   pixMem:= pixDX;
   pixDX := pixDY;
   pixDY :=-pixMem;
   end;

{ positionnement avant l'affichage }
SetBkMode(hDC0, TRANSPARENT);
colOld:=SetTextColor(hDC0, colText);
SetGrClipRgn(flaClipWin1);

TextOutWithSymbol(TRUE);

{TextOut(hDC0, pixXPos - Round(rX*pixDX), pixYPos - Round(rY*pixDY), @strS[1], length(strS));}
SetTextColor(hDC0, colOld);

{ suivant la demande, se place aprs le string, ou en dessous, ou reste en place }
if ((Flags and 1) = 1) then begin
   pixXPos  :=pixXPos  + pixDX;
   vPixXPos :=vPixXPos + pixDX;
   end;

if ((Flags and 2) = 2) then begin
   pixYPos  :=pixYPos  + pixDY;
   vPixYPos :=vPixYPos + pixDY;
   end;

{ Se met effectivement  l'endroit dsir aprs le write... }
MoveToEx(hDC0, pixXPos, pixYPos, nil);

if ((Flags and 4) <> 0) then begin
   { Se remet dans l'ancien mode d'criture }
   fontText.lfEscapement:=lfEscapementMem;
   SelectFont(hDC0, fontText);  // from wFontManager
   end;

if (fontText.lfEscapement <> lfEscapementMem )then begin
   fontText.lfEscapement:=lfEscapementMem;
   SelectFont(hDC0, fontText);  // from wFontManager
   end;

if fDoReleaseDC then ReleaseDC0;
end;  { - TGraphicBase.WriteMove - }

procedure TGraphicBase.Write0(strS:string);
{========================================}
begin { - TGraphicBase.Write0 - }
WriteMove(0,0, strS, 1);
end;  { - TGraphicBase.Write0 - }

{ Avec retour sous la position de dpart }
procedure TGraphicBase.Writeln0(strS:string);
{==========================================}
begin { - TGraphicBase.Writeln0 - }
WriteMove(0, 0, strS, 2);
end;  { - TGraphicBase.Writeln0 - }

procedure TGraphicBase.WriteMoveRel(rX,rY:Float; strS:string);
{============================================================}
begin { - TGraphicBase.WriteMoveRel - }
WriteMove(rX, rY, strS, 0);
end;  { - TGraphicBase.WriteMoveRel - }

procedure TGraphicBase.WritelnMoveRel(rX,rY:Float; strS:string);
{==============================================================}
begin { - TGraphicBase.WriteMoveRel - }
WriteMove(rX, rY, strS, 2);
end;  { - TGraphicBase.WriteMoveRel - }

procedure TGraphicBase.WriteVert(strS:string);
{==========================================}
{ Ecriture verticale,
  Avec continuation en fin de dernire position. }
begin { - TGraphicBase.WriteVert - }
WriteMove(0, 0, strS, 2 + 4);
end;  { - TGraphicBase.WriteVert - }

procedure TGraphicBase.WritelnVert(strS:string);
{==========================================}
{ Avec retour sous la position de dpart, criture verticale }
begin { - TGraphicBase.WritelnVert - }
WriteMove(0, 0, strS, 1 + 4);
end;  { - TGraphicBase.WritelnVert - }

procedure TGraphicBase.WriteVertMoveRel(rX,rY:Float; strS:string);
{================================================================}
begin { - TGraphicBase.WriteVertMoveRel - }
WriteMove(rX, rY, strS, 0 + 4);
end;  { - TGraphicBase.WriteVertMoveRel - }

procedure TGraphicBase.WritelnVertMoveRel(rX,rY:Float; strS:string);
{==================================================================}
begin { - TGraphicBase.WriteVertMoveRel - }
WriteMove(rX, rY, strS, 2 + 4);
end;  { - TGraphicBase.WriteVertMoveRel - }

procedure TGraphicBase.GetPixTextSize(strS:string; var pixX,pixY:integer);
{=======================================================================}
var
  fDoReleaseDC     : boolean;
  rSize            : Size;

begin { - TGraphicBase.GetPixTextSize - }
fDoReleaseDC:=(hDC0 = 0);
if (hDC0 = 0) then begin
   if (GetDC0 = 0) then exit; {error}
   {MoveToEx(hDC0, pixXPos, pixYPos, nil);}
   end;

{ taille en pixels du string }
GetTextExtentPoint32(hDC0, @strS[1], length(strS), rSize);

if (rSize.cx > 0) then begin
   pixX:=rSize.cx;
   pixY:=rSize.cy;
   end
else begin
   DBW(1, 'Warnig, in WriteMove, no text on printer');
   DBW(1, 'size x, y =' + IntToStr(rSize.cx,9) + IntToStr(rSize.cy,9));
   pixX:=0;
   pixY:=0;
   {?????????????}
   end;

if fDoReleaseDC then ReleaseDC0;
end;  { - TGraphicBase.GetPixTextSize - }

procedure TGraphicBase.GetMMTextSize(strS:string; var mmX,mmY:Float);
{==================================================================}
var
  pixX,pixY        : integer;
begin { - TGraphicBase.GetMMTextSize - }
GetPixTextSize(strS, pixX, pixY);
mmX:=pixX / vPixXSmm;
mmY:=pixY / vPixYSmm;
end;  { - TGraphicBase.GetMMTextSize - }

procedure TGraphicBase.SaveTextAttribRec(var recTextAt:TypeTextAttrib);
{==================================================================}
begin { - TGraphicBase.SaveTextAttribRec - }
end;  { - TGraphicBase.SaveTextAttribRec - }

procedure TGraphicBase.GetTextAttribRec( var recTextAt:TypeTextAttrib);
{==================================================================}
begin { - TGraphicBase.GetTextAttribRec - }
end;  { - TGraphicBase.GetTextAttribRec - }



{ *** UNIT wGrBase *** }
end.
